"""
The main python toolkit application.
"""
#---Logging---------------------------------------------------------------------
import logging
log = logging.getLogger('App')

#---Imports---------------------------------------------------------------------
import __main__         #access to the main namespace
import os.path          #path operations
import sys              #access sys arguments
import imp              #package checking
import wx               #gui libary

#import the aui libary and store under the main app level for all tools to use
#this makes it easy to switch to pyaui if necessary
import wx.aui as aui

import ptk_lib
from ptk_lib.message_bus.mb_node import MBNode
from ptk_lib.message_bus.mb_socket_client import MBSocketClient
from ptk_lib.message_bus.wx_message_bus import wxMessageBus
from ptk_lib.misc import USERDIR, TOOLDIR

from ptk_lib.icons import splash
from ptk_lib import tool_manager       #plugins/tools management

#Core tools
from ptk_lib.core_tools import taskicon         #taskbar/settings icon tool
from ptk_lib.core_tools import engine_manager
from ptk_lib.core_tools import fileio           #file import and export system

from ptk_lib.core_tools import console          #main console window
from ptk_lib.core_tools import editor           #editor tool

from ptk_lib.core_tools import views            #GUI views of python objects
from ptk_lib.core_tools import pathman          #python path management tool
from ptk_lib.core_tools import nsbrowser        #namespace browser
from ptk_lib.core_tools import inspector        #object inspection tool
from ptk_lib.core_tools import cmdhistory       #command history viewer

from ptk_lib.core_tools import numpy_pack       #numpy module extensions

#---main application class------------------------------------------------------
class PTKApp(wx.App):
    """ The main ptk application class """
    def __init__(self,name, files):
        """Create the ptk app"""
        wx.App.__init__(self,redirect=False,clearSigInt=False)

        #set application name and check this is not a second instance-----------
        self.SetAppName(name)
        self.imon = wx.SingleInstanceChecker(name+'-'+wx.GetUserId(), USERDIR) 
        #check if another is running?
        if self.imon.IsAnotherRunning():
            log.info('App is a second instance')
            self.SecondInstance(files)
            #program stopped
        #-----------------------------------------------------------------------

        #store directory and app name
        self.ptkdir = ptk_lib.__file__.rpartition(os.sep)[0] +os.sep
        self.name = name

        #---Show a splash screen -----------------------------------------------
        bmp = splash.splash.GetBitmap()
        splashscreen = wx.SplashScreen(bmp, wx.SPLASH_CENTRE_ON_SCREEN| wx.SPLASH_NO_TIMEOUT, 10000, None, -1)
        splashscreen.Show()

        #---Start the Message Bus ----------------------------------------------
        log.info('Starting MessageBus')
        self.msg_bus = wxMessageBus(name='PTK', timeout=30)
        try:
            #load the messenger port from settings
            cfg = self.GetConfig()
            cfg.SetPath("app//")
            port = cfg.ReadInt("message_bus_port",6666)
            self.msg_bus.start_server(port=port, allow_ext=False)
        except:
            pass
        self.msg_node = MBNode(self.msg_bus, 'App')
        self.msg_node.set_handler('NewInstance', self.msg_new_instance) 

        #initialise tool plugin system -----------------------------------------   
        log.info('Starting Toolmanager')
        self.toolmgr = tool_manager.ToolManager(TOOLDIR)
       
        #Start core tools ------------------------------------------------------
        log.info('Starting core tools')

        #top level tools - need to be started in the correct order
        self.toolmgr.start_tool('TaskIcon')         #first for console/editor menu items
        self.toolmgr.start_tool('Console')          #before enginemanager for toolbars
        self.toolmgr.start_tool('FileIO')           #before editor for editor importer
        self.toolmgr.start_tool('Editor')           #after file to add importer
        self.toolmgr.start_tool('EngineManager')    #after console and editor to add toolbars

        #other tools doesn't really matter about the order.
        self.toolmgr.start_tool('Views')
        self.toolmgr.start_tool('PathManager')
        self.toolmgr.start_tool('NSBrowser')
        self.toolmgr.start_tool('CmdHistory')
        self.toolmgr.start_tool('Inspector')

        #module packs
        try:
            imp.find_module('numpy')
            self.toolmgr.start_tool('NumPy Pack')
        except:
            log.info('Numpy not found - skipping loading tool - NumPy Pack')
        log.info('Done loading core tools')

        #start the user tools -------------------------------------------------
        log.info('Starting user tools')
        self.toolmgr.load_settings()     #load other tools
        log.info('Done loading user tools')

        #Publish app.init message ---------------------------------------------
        log.info('Application init done - publishing App.Init message')    
        self.msg_node.publish_msg('App.Init',())

        #process any input arguments ------------------------------------------
        wx.CallAfter(self.ProcessInputArgs, files)
    
        #close splash screen---------------------------------------------------
        wx.CallAfter(splashscreen.Close)

    #---------------------------------------------------------------------------    
    def ShowTips(self, override=False):
        config = self.GetConfig()
        config.SetPath("App//")
        showtip = config.ReadBool("show_tips",True)
        tip_n   = config.ReadInt("tip_n",0)

        if showtip or override:
            tp = wx.CreateFileTipProvider( self.ptkdir+"docs/tips.txt", tip_n)
            showtip = wx.ShowTip(None, tp, showtip)
            tip_n = tp.GetCurrentTip()
            config.WriteBool("show_tips", showtip)
            config.WriteInt("tip_n", tip_n)
            config.Flush()

    def GetConfig(self):
        """
        Return the application wxConfig object
        """
        cfg = wx.FileConfig(localFilename=USERDIR+'options')
        return cfg

    def SecondInstance(self, files):
        """Called when this app is a second instance"""
        #load the messenger port from settings
        cfg = self.GetConfig()
        cfg.SetPath("App//")
        port = cfg.ReadInt("message_bus_port",6666)

        #create a client messenger to communicate with running instance
        msg_client = MBSocketClient(name='PTK second instance',
                                    address=('localhost',port))
        msg_client.connect()
            
        #send a message with the files to open
        #convert to absolute paths
        data = []
        for file in files:
            data.append(os.path.abspath(file))

        msg_client.send_msg('App', 'NewInstance', (data,))
        #now close and exit
        msg_client.close()
        log.info('Closing second instance')
        logging.shutdown()
        exit()

    def ProcessInputArgs(self,files=[]):
        """
        Process the system arguments from this instance or a second instance 
        sent via messenger
        """
        log.info('Processing input filepaths: '+str(files))
        #if no files to open i.e. ==[] show the console window
        if files ==[]:
            self.msg_node.send_msg('Console', 'Show',())
            wx.CallAfter(self.ShowTips)
        else:
            #open the files
            self.msg_node.send_msg('Editor','Open',files)

    def VetoExit(self):
        """
        Prevent an imminent application exit
        """
        self.exit_veto = True

    def Exit(self):
        """
        Exit the application
        """
        log.info('Application exit request')

        #Send exitcheck to ensure it is ok to exit - listeners call app.VetoExit
        # to prevent application exit.
        log.info('Preforming exit check') 
        self.exit_veto = False
        self.msg_node.publish_msg('App.ExitCheck',())
        wx.YieldIfNeeded()
        if self.exit_veto is True:
            log.info('Exit vetoed')
            return
      
        #send application exit message - any listeners can do exit tasks
        log.info('Continuing with exit, sending exit message')
        self.msg_node.publish_msg('App.Exit',())
        wx.YieldIfNeeded()
        
        #shutdown toolmanager
        self.toolmgr.shutdown()

        #shutdown messenger
        self.msg_bus.shutdown()

        #and exit
        log.info('Actually exiting now')
        self.imon = None
        logging.shutdown()
        wx.App.Exit(self)

    #---message handlers--------------------------------------------------------
    def msg_new_instance(self,msg):
        """
        Called with the input arguments to a second instance - the sending
        instance then closes.
        """
        files, = msg.get_data()
        log.info('Second instance started, args were: '+str(files))
        self.ProcessInputArgs(files)
