"""
This is a frame class to easily add some aui related functionally:
    - add toolbars complete with a menu checkitem   
    - save and load layouts complete with menu checkitems
"""

import pickle
import wx

import wx.aui as aui

#---Layout class----------------------------------------------------------------
class Layout():
    def __init__(self,parent,menu,
                    name="",pos=(-1,-1),size=(800,600),max=False,layout="",
                    data=None):
        """
        Create a new aui window (=parent) layout .
        Adds a checkitem in menu if not None.

        Create from values:
            name    = layout name string,
            pos     = frame position tuple,
            size    = frame size tuple,
            max     = frame maximised boolean,
            layout  = aui layout string 

        Create from a pickled string generated from GetSaveString, if data is
        not None
        """
        #references to the parent frame and menu to use
        self.parent = parent
        self.menu = menu 
        self.menuitem = None

        if data==None:
            #store info
            self.name = name
            self.pos  = pos
            self.size = size
            self.max  = max
            self.layout = str(layout)
        else:
            self.LoadString(data)

        #Add a menu item
        if menu!=None:
            self.AddMenuItem(menu)

    def Apply(self):
        """Apply these settings to the parent frame"""
        #maximise window?
        if self.max is True: #if it is set the size to default and then maximise
            self.parent.SetPosition((-1,-1))
            self.parent.SetSize((800,600))
            self.parent.Maximize(True)
        else:
            #set size and position
            self.parent.SetSize(self.size)
            self.parent.SetPosition(self.pos)
            #self.parent.Maximize(False)

        #load aui layout
        if self.layout!="":
            self.parent.auimgr.LoadPerspective(self.layout)
        #update aui manager
        self.parent.auimgr.Update()

        
    def SaveCurrentState(self):
        """Saves the current state of the console frame to this layout"""
        #maximised,size and position
        self.max = self.parent.IsMaximized()
        self.size = self.parent.GetSizeTuple()
        self.pos = self.parent.GetScreenPositionTuple()

        #aui layout
        self.layout = self.parent.auimgr.SavePerspective()

    def AddMenuItem(self,menu):
        """Add a menu item to menu to load this layout"""
        id = wx.NewId()  
        self.menuitem = wx.MenuItem(menu, id ,self.name,"Load this layout")
        menu.AppendItem(self.menuitem)
        self.parent.Bind(wx.EVT_MENU,lambda event: self.Apply(), self.menuitem)

    def Rename(self,new):
        """Renames the layout and alters the asociated menu item"""
        #make sure name is a string
        self.name = new
        #find and change the menu item
        if self.menuitem is not None:
            self.menuitem.SetItemLabel(self.name)   

    def RemoveMenuItem(self):
        """Removes this layouts menu """
        if self.menuitem is not None:
            self.menu.RemoveItem(self.menuitem) 
            self.menuitem.Destroy()
            self.menuitem = None

    def GetSaveString(self):
        """Returns a save string for writing to the config"""
        data = (self.name,self.pos,self.size,self.max,self.layout)
        s = pickle.dumps(data)
        return s

    def LoadString(self,data):
        """Load settings from a save string"""
        self.name,self.pos,self.size,self.max,self.layout  = pickle.loads(str(data)) 

#---Mixin class-----------------------------------------------------------------
class AUIFrame(wx.Frame):
    def __init__(self,parent,id,title,size,pos):
        """
        Add aui with support for
            - add toolbars complete with a menu checkitem   
            - save and load layouts complete with menu checkitems
        """
        wx.Frame.__init__(self, parent, id, title, size=size,pos=pos)

        #initalise an aui manager
        self.auimgr = aui.AuiManager()
        #not in wxpython version >2.8?        
        #self.auimgr.SetFlags(aui.AUI_MGR_DEFAULT |aui.AUI_MGR_ALLOW_ACTIVE_PANE)
        self.auimgr.SetManagedWindow(self)

        #list of layout objects
        self._layouts = []
        self.path = None #save path 
        self.laymenu = None #menu for checkitems

    def Show(self,show=True):
        self.auimgr.Update()
        wx.Frame.Show(self,show)

    #---Layouts Interface-------------------------------------------------------
    def AddLayout(self,layout=None,name="New layout"):
        """
        Save a new layout:
            If layout is None a new layout is created using the name and the 
            current settings
            If menu is given a checkitem will be added to the menu
        """
        if layout is None:
            layout = Layout(self,name=name,menu=self.laymenu)
            layout.SaveCurrentState()
        self._layouts.append(layout)

    def RemoveLayout(self,n):
        """Remove the nth saved layout (including from any menus)"""
        layout = self._layouts.pop(n)
        layout.RemoveMenuItem()
        del layout

    def RenameLayout(self,n,newname):
        """Rename the nth saved layout (including from any menus)"""
        #make sure name is a string
        newname = str(newname)
        layout = self._layouts[n]
        layout.Rename(newname)

    def SetSavePath(self,path):
        """Set the save path in the wxconfig object"""
        self.path = path

    def SetLayoutsMenu(self,menu):
        """Set the menu to add layout checkitems to"""
        self.laymenu = menu
        id =wx.NewId()
        self.laymenu.Append(id,'Save Layout','Save the window layout')
        self.Bind(wx.EVT_MENU, self.OnSaveLayout, id=id)
        id =wx.NewId()
        self.laymenu.Append(id,'Rename Layout','Rename a saved window layout')
        self.Bind(wx.EVT_MENU, self.OnRenameLayout, id=id)
        id =wx.NewId()
        self.laymenu.Append(id,'Delete Layout','Delete a saved window layout')
        self.Bind(wx.EVT_MENU, self.OnDeleteLayout, id=id)
        self.laymenu.AppendSeparator()

    def SaveLayouts(self):
        """
        Save the layouts to the path (set via SetSavePath) in the wxconfig 
        object
        """
        #check save/load path is set
        if self.path is None:
            return
        #Get the config object
        cfg = wx.GetApp().GetConfig()

        #save the current layout
        cur = Layout(self,menu=None,data=None,name='last')
        cur.SaveCurrentState()
        s = cur.GetSaveString()
        cfg.Write(self.path+"//current_layout",s)

        #delete any old saved layouts 
        cfg.DeleteGroup(self.path+"//layouts")
        #save the other layouts
        n=0
        for layout in self._layouts:
            s = layout.GetSaveString()
            cfg.Write(self.path+"//layouts//saved"+str(n),s)
            n = n+1
        cfg.Flush()


    def LoadLayouts(self):
        """
        Loads the layouts from the path (set via SetSavePath) in the wxconfig 
        object. Adds checkitems into menu set via SetLayoutsMenu
        """
        #check save/load path is set
        if self.path is None:
            return
        #Get the config object
        cfg = wx.GetApp().GetConfig()
        
        #load the saved layouts
        cfg.SetPath(self.path+"//layouts//")
        num = cfg.GetNumberOfEntries()
        for n in range(0,num):
            s = cfg.Read("saved"+str(n),"")
            if s!="":
                layout = Layout(self,menu=self.laymenu,data=s)
                self.AddLayout(layout)

        #load the current layout
        cfg.SetPath("")
        s = cfg.Read(self.path+"//current_layout","")
        if s!="":
            cur = Layout(self,menu=None,data=s)
            cur.Apply() #apply this layout


    #---Toolbars interfaces-----------------------------------------------------
    def SetToolbarsMenu(self,menu):
        """Set the toolbars menu to use"""
        self.toolsmenu = menu

    def AddToolbar(self,tb,tbname,pane=None,helpstring=None):
        """
        Add a toolbar to the console frame complete with entry in the menu set
        via SetToolbarsMenu)

        Pass a toolbar and string for the name and an aui pane info with the 
        desired settings (or None for default settings), helpstring is the menu 
        item tip, use none for defualt.
        """
        #defualt pane settings
        if pane is None:
            pane = (aui.AuiPaneInfo().Name(tbname)
                    .Caption(tbname).ToolbarPane().CloseButton(True)
                    .CaptionVisible(False).DestroyOnClose(False).Top()
                    .LeftDockable(False).RightDockable(False))
        #check name and DestroyOnClose
        pane.Name(tbname)
        pane.DestroyOnClose(False)

        #add to the window using aui manager
        self.auimgr.AddPane(tb,pane)
        self.auimgr.Update()

        #check help string
        if helpstring is None:
            helpstring = 'Show/Hide the '+tbname+' toolbar'

        #add a tick item to the view>toolbars menu
        id = wx.NewId()
        item = wx.MenuItem(self.toolsmenu, id ,tbname,helpstring,wx.ITEM_CHECK)
        self.toolsmenu.AppendItem(item)
        item.Check(True) #tick it

        #Bind events to handle the tick/unticking and show/hiding
        self.Bind(wx.EVT_MENU,self.OnToolsMenu,item)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateToolsMenu,item)

    #---event handlers----------------------------------------------------------
    def OnToolsMenu(self,event):
        """show hide a menubar"""
        id = event.GetId()
        item = self.toolsmenu.FindItemById(id)
        name = item.GetLabel()
        pane = self.auimgr.GetPane(name)
        if event.IsChecked() is True:
            pane.Show()
            self.auimgr.Update()
        else:
            pane.Hide()
            self.auimgr.Update()

    def OnUpdateToolsMenu(self,event):
        """check the tick state of the menu items"""
        #check the state of each toolbar in turn
        id = event.GetId()
        item = self.toolsmenu.FindItemById(id)
        name = item.GetLabel()
        pane = self.auimgr.GetPane(name)
        event.Check( pane.IsShown() )

    def OnSaveLayout(self,event):
        """Handles saving of an aui layout"""
        dlg = wx.TextEntryDialog(self, "Enter a name for the new layout:", "Python toolkit")
        dlg.SetValue(("New layout %d")%(len(self._layouts)+1))

        if dlg.ShowModal() == wx.ID_OK:
            name = dlg.GetValue()
            self.AddLayout(name=name)
    
    def OnRenameLayout(self,event):
        """Handles renaming of an aui layout"""
        #get layout to change
        names = []
        for x in self._layouts:
            names.append(x.name)
        dlg = wx.SingleChoiceDialog(
                self, 'Select layout to rename:', 'Rename layout',
                names, wx.CHOICEDLG_STYLE)
        if dlg.ShowModal() != wx.ID_OK:
            dlg.Destroy()
            return
        
        n = dlg.GetSelection()
        dlg.Destroy()

        #get new name
        dlg = wx.TextEntryDialog(self, 'Enter new name:','Rename layout', 
                    self._layouts[n].name)
        if dlg.ShowModal() == wx.ID_OK:
            new = dlg.GetValue()
            self.RenameLayout(n,new)
        dlg.Destroy()

    def OnDeleteLayout(self,event):
        """Handles deleting an aui layout"""
        #get layout to delete
        names = []
        for x in self._layouts:
            names.append(x.name)
        dlg = wx.SingleChoiceDialog(
                self, 'Select layout to remove:', 'Remove layout',
                names, wx.CHOICEDLG_STYLE)
        if dlg.ShowModal() == wx.ID_OK:
            n = dlg.GetSelection()
            self.RemoveLayout(n)
        dlg.Destroy()
