"""
This contains the main console window:

ConsoleFrame - The main matlab style window.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('ConsoleFrame')
#log.setLevel(logging.DEBUG)

#---Imports---------------------------------------------------------------------
import wx
import wx.aui as aui

from ptk_lib import VERSION
from ptk_lib.controls import dialogs
from ptk_lib.icons import ptk_icons, common16
from ptk_lib.controls import auiframe
from ptk_lib.misc import open_help

from ptk_lib.core_tools.fileio import FileDrop
from ptk_lib.core_tools.taskicon import PTKInfoDialog

from console_stc import ConsoleSTC
import console_icons

#---Frame class ----------------------------------------------------------------
class ConsoleFrame(auiframe.AUIFrame):
    def __init__(self, tool):

        #reference to parent tool
        self.tool = tool

        #create the frame
        auiframe.AUIFrame.__init__( self, None, -1, "PTK Console", 
                                    size=(800, 600),pos=(-1,-1) )

        #set the window icon
        ib = wx.IconBundle()
        ib.AddIcon(console_icons.console16.GetIcon())
        ib.AddIcon(console_icons.console32.GetIcon())
        ib.AddIcon(console_icons.console48.GetIcon())
        self.SetIcons(ib)

        #create the status bar
        self.CreateStatusBar()
        self.SetStatusText("Python toolkit v"+VERSION)

        #create a droptarget
        self.dt = FileDrop()
        self.SetDropTarget(self.dt)

        #create the menu
        log.debug('Creating menu')
        self.menubar = ConsoleMenu(self, self.tool)
        self.SetMenuBar(self.menubar)
        self.SetToolbarsMenu(self.menubar.toolbars_menu)

        #create the console pane
        log.debug('Creating ConsoleBook')

        style = wx.aui.AUI_NB_DEFAULT_STYLE | wx.aui.AUI_NB_BOTTOM
        self.conbook = wx.aui.AuiNotebook(self,-1,style=style)
        self._pages = {} #dict of {engname:stc} for finding page index

        self.conbook.Bind(wx.aui.EVT_AUINOTEBOOK_PAGE_CLOSE,self.OnConsoleClose)
        self.conbook.Bind(wx.aui.EVT_AUINOTEBOOK_PAGE_CHANGED,self.OnConsoleChange)

        pane = aui.AuiPaneInfo()
        name='Console'
        pane.Name(name) #id name
        pane.Caption(name) # caption
        pane.Center() #position
        pane.CloseButton(False) #close button
        pane.MaximizeButton()
        pane.MinimizeButton()
        pane.Floatable(False)
        pane.Movable(True)
        pane.BestSize( (400,600) )
        pane.DestroyOnClose(False)
        self.auimgr.AddPane(self.conbook,pane)

        ##create the main toolbar
        log.debug('Creating the toolbar')
        self.maintools = MainTools(self, self.tool)
        pane = (aui.AuiPaneInfo().Name('Main toolbar')
                    .Caption('Main toolbar').ToolbarPane().CloseButton(True)
                    .CaptionVisible(False)
                    .DestroyOnClose(False).Top().Row(0).LeftDockable(False)
                    .RightDockable(False))
        self.AddToolbar(self.maintools,'Main toolbar',pane,
                        helpstring = 'Show/hide the Main toolbar')


        #load the frame settings adding layout checkitems to the view menu
        self.SetSavePath('Console')
        self.SetLayoutsMenu(self.menubar.view_menu)

        #bind close events
        self.Bind(wx.EVT_CLOSE,self.OnClose)

        #update aui manager
        self.auimgr.Update()

        #subscribe to engine messages
        self.tool.msg_node.subscribe('Engine.Busy', self.msg_eng_busy)
        self.tool.msg_node.subscribe('Engine.Done', self.msg_eng_done)

        self.tool.msg_node.subscribe('Engine.Debug.Paused', self.msg_debug_paused)
        self.tool.msg_node.subscribe('Engine.Debug.Resumed', self.msg_debug_resumed)

        self.tool.msg_node.subscribe('Engine.Switched', self.msg_eng_switched)

    #---interface methods-------------------------------------------------------
    def GetMenu(self,menu):
        """
        Return a reference to the menu/menubar:
        menu = 'menubar','file','edit','view','toolbars','tools','help'
        """
        menu = menu.lower()
        if menu=='menubar':
            return self.menubar
        if menu=='file':
            return self.menubar.file_menu  #file menu
        if menu=='edit':
            return self.menubar.edit_menu  #edit menu
        if menu=='view':
            return self.menubar.view_menu  #view menu
        if menu=='toolbars':
            return self.menubar.toolbars_menu  #toolbar submenu
        if menu=='tools':
            return self.menubar.tool_menu  #tool menu
        if menu=='help':
            return self.menubar.help_menu  #help menu

    def AddConsole(self, engname, engtype):

        if self._pages.has_key(engname):
            #find old stc with that name and remove it
            old = self._pages.pop(engname)
            n = self.conbook.GetPageIndex(old)
            self.conbook.RemovePage(n)
            old.Destroy()

        #create console        
        console = ConsoleSTC(self.conbook, engname, self.tool.cmdhistory)

        self.conbook.AddPage(console, engname, False, bitmap=console_icons.console16.GetBitmap())
        self.conbook.SetSelection(self.conbook.GetPageIndex(console))
        self._pages[ engname ] = console

    def GetConsole(self, engname):
        """get the console, returns None if no console"""
        return self._pages.get(engname,None)

    def GetAllConsoles(self):
        return self._pages.values()

    def GetCurrentConsole(self):
        """Get the console controlling the current engine"""
        engmgr = wx.GetApp().toolmgr.get_tool('EngineManager')
        cureng = engmgr.get_current_engname()
        if cureng is None:
            return None
        return self._pages.get(cureng, None)

    #---message handlers ------------------------------------------------------
    def msg_eng_busy(self, msg):
        engname = msg.data[0]
        engmgr = wx.GetApp().toolmgr.get_tool('EngineManager')
        cureng = engmgr.get_current_engname()
        if engname == cureng:
            self.StatusBar.SetStatusText('Busy...')
            self.StatusBar.Update()

    def msg_eng_done(self, msg):
        engname = msg.data[0]
        engmgr = wx.GetApp().toolmgr.get_tool('EngineManager')
        cureng = engmgr.get_current_engname()
        if (engname == cureng):
            self.StatusBar.SetStatusText('Done.')
            self.StatusBar.Update()

    def msg_debug_paused(self, msg):
        engname = msg.data[0]
        engmgr = wx.GetApp().toolmgr.get_tool('EngineManager')
        cureng = engmgr.get_current_engname()
        if engname == cureng:
            self.StatusBar.SetStatusText('Paused')
            self.StatusBar.Update()

    def msg_debug_resumed(self, msg):
        engname = msg.data[0]
        engmgr = wx.GetApp().toolmgr.get_tool('EngineManager')
        cureng = engmgr.get_current_engname()
        if engname == cureng:
            self.StatusBar.SetStatusText('Busy (debugging)...')
            self.StatusBar.Update()

    def msg_eng_switched(self, msg):
        engmgr = wx.GetApp().toolmgr.get_tool('EngineManager')
        eng = engmgr.get_current_engine()
        #enable/disable import menu item.
        if eng is None:
            import_item = self.menubar.file_menu.FindItemById(self.menubar.ID_IMPORT)
            import_item.Enable(False)
            return
        else:
            import_item = self.menubar.file_menu.FindItemById(self.menubar.ID_IMPORT)
            import_item.Enable(True)

        busy, debug, profile = eng.get_state() 
        if busy and debug:
            string = 'Busy (debugging)...'
        elif busy and profile:
            string = 'Busy (profiling)...'
        elif busy:
            string = 'Busy...'
        else:
            string = 'Ready' 
        self.StatusBar.SetStatusText(string)
        self.StatusBar.Update()

    #---Event handlers----------------------------------------------------------
    def OnClose(self,event):
        """
        Window close event handler
        """
        self.Hide()
    
    def OnConsoleClose(self, event):
        #confirm console/engine close
        num  = self.conbook.GetSelection()
        console = self.conbook.GetPage(num)
        name = console.engname
        
        #check if the console is connected to an engine
        if console._connected is False:
            self._pages.pop( name, None )
            event.Skip()
            return

        name = console.engname
        res = dialogs.ConfirmDialog('Close Engine: '+name+' ?','Close Engine?')

        #close engine and remove from enginemanager
        if res is False:
            event.Veto()
        else:
            #close engine
            engmgr = wx.GetApp().toolmgr.get_tool('EngineManager')
            engmgr.close_engine(name)
            self._pages.pop( name, None )
            event.Skip()

    def OnConsoleChange(self, event):
        num  = self.conbook.GetSelection()  
        console = self.conbook.GetPage(num)

        #check if the console is connected to an engine
        if console._connected is False:
            event.Skip()
            return
        try:
            engmgr = wx.GetApp().toolmgr.get_tool('EngineManager')
            engmgr.set_current_engname(console.engname)
        except:
            pass

#---main toolbar----------------------------------------------------------------
class MainTools(wx.ToolBar):
    def __init__(self,parent, tool):

        wx.ToolBar.__init__(self,parent,-1,wx.DefaultPosition, 
                    wx.DefaultSize,wx.TB_FLAT | wx.TB_NODIVIDER)
        self.parent = parent
        self.tool = tool

        #set the icon size
        size=wx.Size(16, 16)
        self.SetToolBitmapSize(size)
        #load some icons
        new_bmp     = common16.document_new.GetBitmap()
        open_bmp    = common16.document_open.GetBitmap()
        cut_bmp     = common16.edit_cut.GetBitmap()
        copy_bmp    = common16.edit_copy.GetBitmap()
        paste_bmp   = common16.edit_paste.GetBitmap()
        bmp_clear    = common16.edit_clear.GetBitmap() 
        help_bmp    = common16.help_browser.GetBitmap()

        #new
        self.AddLabelTool( wx.ID_NEW, "New", new_bmp,shortHelp='New',
                            longHelp='Create a new file in the editor')
        self.Bind(wx.EVT_TOOL, self.OnNew, id=wx.ID_NEW)
        #open
        self.AddLabelTool( wx.ID_OPEN, "Open", open_bmp,shortHelp='Open',
                            longHelp='Open an existing file in the editor')
        self.Bind(wx.EVT_TOOL, self.OnOpen, id=wx.ID_OPEN)
        self.AddSeparator()
        #-----------------------------------------------------------------------
        #cut
        self.AddLabelTool( wx.ID_CUT, "Cut", cut_bmp,shortHelp='Cut',
                            longHelp='Cut selection to the clipboard')
        self.Bind(wx.EVT_TOOL, self.OnCut, id=wx.ID_CUT)
        #copy
        self.AddLabelTool( wx.ID_COPY, "Copy", copy_bmp,shortHelp='Copy',
                            longHelp='Copy selection to the clipboard')
        self.Bind(wx.EVT_TOOL, self.OnCopy, id=wx.ID_COPY)
        #paste
        self.AddLabelTool( wx.ID_PASTE, "Paste", paste_bmp,shortHelp='Paste',
                            longHelp='Paste from clipboard')
        self.Bind(wx.EVT_TOOL, self.OnPaste, id=wx.ID_PASTE)
        self.AddSeparator()
        
        #Clear console
        self.AddLabelTool( wx.ID_CLEAR, "Clear", bmp_clear, shortHelp='Clear console',
                        longHelp='Clear the current engines console')
        self.Bind(wx.EVT_TOOL, self.OnClear, id=wx.ID_CLEAR)
        self.AddSeparator()

        #help
        self.AddLabelTool( wx.ID_HELP, "Help", help_bmp,shortHelp='Help',longHelp='Open help documentation')
        self.Bind(wx.EVT_TOOL, self.OnHelp, id=wx.ID_HELP)
        #-----------------------------------------------------------------------
        self.Realize()


    #---Event handlers----------------------------------------------------------
    def OnNew(self,event):
        """New file (menu and toolbar) handler"""
        self.tool.msg_node.send_msg('Editor','New',())

    def OnOpen(self,event):
        """Open file (menu and toolbar) handler"""
        self.tool.msg_node.send_msg('Editor','Open',())
    
    def OnCut(self,event):
        """Calls the cut method of the active pane (if it exists)"""
        active= self.FindFocus()
        if active is not None: 
            try:
                active.Cut()
            except AttributeError:
                pass

    def OnCopy(self,event):
        """Calls the copy method of the active pane (if it exists)"""
        active= self.FindFocus()
        if active is not None: 
            try:
                active.Copy()
            except AttributeError:
                pass

    def OnPaste(self,event):
        """Calls the paste method of the active pane (if it exists)"""
        active= self.FindFocus()
        if active is not None: 
            try:
                active.Paste()
            except AttributeError:
                pass

    def OnClear(self,event): 
        console = self.parent.GetCurrentConsole()
        if console is None:
            return
        console.clear_console()

    def OnHelp(self,event):
        """Open the help browser"""
        open_help()

#----Menu ----------------------------------------------------------------------
class ConsoleMenu(wx.MenuBar):
    def __init__(self,parent, tool):
        wx.MenuBar.__init__(self)

        #parent tool
        self.tool = tool

        #parent frame
        self.parent = parent

        self.file_menu       = wx.Menu()     #file menu
        self.edit_menu       = wx.Menu()     #edit menu
        self.view_menu       = wx.Menu()     #view menu
        self.toolbars_menu   = wx.Menu()     #toolbar submenu
        self.tool_menu       = wx.Menu()     #tool menu
        self.help_menu       = wx.Menu()     #help menu

        ##add the menus to the menu bar
        self.Append(self.file_menu, "File")
        self.Append(self.edit_menu, "Edit")
        self.Append(self.view_menu, "View")
        self.Append(self.tool_menu, "Tools")
        self.Append(self.help_menu, "Help")

        ##file menu
        self.file_menu.Append(wx.ID_NEW ,'New','Create a new file in the editor')
        self.file_menu.Append(wx.ID_OPEN,'Open','Open an existing file in the editor')
        self.file_menu.AppendSeparator()
        self.ID_IMPORT = wx.NewId()
        self.file_menu.Append(self.ID_IMPORT,'Import...','Import python data from file')
        self.file_menu.AppendSeparator()
        self.file_menu.Append(wx.ID_CLOSE, 'Hide Console','Hides the console window to the taskbar')
        self.file_menu.Append(wx.ID_EXIT, 'Exit','Exit PTK')
        #bindings
        self.parent.Bind(wx.EVT_MENU, self.OnNew, id=wx.ID_NEW)
        self.parent.Bind(wx.EVT_MENU, self.OnOpen, id=wx.ID_OPEN)
        self.parent.Bind(wx.EVT_MENU, self.OnImport, id=self.ID_IMPORT)
        self.parent.Bind(wx.EVT_MENU, self.OnMenuClose, id=wx.ID_CLOSE)
        self.parent.Bind(wx.EVT_MENU, self.OnExit, id=wx.ID_EXIT)

        ##edit menu
        self.edit_menu.Append(wx.ID_CUT,'Cut','Cut selection to clipboard')
        self.edit_menu.Append(wx.ID_COPY,'Copy','Copy selection to clipboard')
        self.edit_menu.Append(wx.ID_PASTE,'Paste','Paste from clipboard')
        self.edit_menu.AppendSeparator()
        self.edit_menu.Append(wx.ID_PREFERENCES, 'Preferences', 'Edit the program preferences') 
        #bindings
        self.parent.Bind(wx.EVT_MENU, self.OnCut, id=wx.ID_CUT)
        self.parent.Bind(wx.EVT_MENU, self.OnCopy, id=wx.ID_COPY)
        self.parent.Bind(wx.EVT_MENU, self.OnPaste, id=wx.ID_PASTE)
        self.parent.Bind(wx.EVT_MENU, self.OnPreferences, id=wx.ID_PREFERENCES)

        ##view menu
        self.view_menu.AppendSubMenu(self.toolbars_menu , 'Toolbars...','Show Toolbars')
        self.view_menu.AppendSeparator()
        #layouts added in auimixin

        ##tools menu       
        #tools add themselves here

        ##Help menu
        self.help_menu.Append(wx.ID_HELP, 'Help', 'Open the python documentation...') 
        tipid=wx.NewId()
        self.help_menu.Append(tipid, 'Show tips', 'Show tips') 
        self.help_menu.Append(wx.ID_ABOUT, 'About...', 'About this program...') 

        #bindings
        self.parent.Bind(wx.EVT_MENU, self.OnHelp, id=wx.ID_HELP)
        self.parent.Bind(wx.EVT_MENU, self.OnTip, id=tipid)
        self.parent.Bind(wx.EVT_MENU, self.OnAbout, id=wx.ID_ABOUT)

    #---Event handlers----------------------------------------------------------
    #file menu
    def OnNew(self,event):
        """New file (menu and toolbar) handler"""
        self.tool.msg_node.send_msg('Editor','New',())

    def OnOpen(self,event):
        """Open file (menu and toolbar) handler"""
        self.tool.msg_node.send_msg('Editor','Open',())

    def OnImport(self,event):
        """Import python data  from file"""
        self.tool.msg_node.send_msg('FileIO','Import',('',))

    def OnMenuClose(self,event):
        self.parent.Hide()

    def OnExit(self, event):
        """Called when exit menu item selected"""
        app = wx.GetApp()
        app.Exit()

    #edit menu
    def OnCut(self,event):
        """Calls the cut method of the active pane (if it exists)"""
        active= self.FindFocus()
        if active is not None: 
            try:
                active.Cut()
            except AttributeError:
                pass

    def OnCopy(self,event):
        """Calls the copy method of the active pane (if it exists)"""
        active= self.FindFocus()
        if active is not None: 
            try:
                active.Copy()
            except AttributeError:
                pass

    def OnPaste(self,event):
        """Calls the paste method of the active pane (if it exists)"""
        active= self.FindFocus()
        if active is not None: 
            try:
                active.Paste()
            except AttributeError:
                pass

    def OnPreferences(self,event):
        """Show the settings dialog"""
        self.tool.msg_node.send_msg('TaskIcon','Settings.Show',())

    #view menu
    #
    # done in auimixin
    #

    #tools menu
    #
    # added by tools
    #

    #help menu
    def OnHelp(self,event):
        """Open the help browser"""
        open_help()  

    def OnTip(self,event):
        """Open the help tips"""
        app=wx.GetApp()
        app.ShowTips(override=True)

    def OnAbout(self,event):
        """Opens the about box"""
        PTKInfoDialog()


