"""
The console use a styled text control for input/output.


It implements the engine module console interface along with extras such as 
autocompletes and calltips.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('ConsoleSTC')
log.setLevel(logging.DEBUG)

#---Imports---------------------------------------------------------------------
import keyword                  
import wx
from wx import stc

from ptk_lib.engine.console import ConsoleInterface
from ptk_lib.core_tools.fileio import FileDrop

import console_tasks
from calltip import CallTip

#cursor navigation keys
NAVKEYS = ( wx.WXK_END  , wx.WXK_HOME,
            wx.WXK_LEFT , wx.WXK_NUMPAD_LEFT,
            wx.WXK_RIGHT, wx.WXK_NUMPAD_RIGHT,
            wx.WXK_UP   , wx.WXK_NUMPAD_UP,
            wx.WXK_DOWN , wx.WXK_NUMPAD_DOWN,
            wx.WXK_PAGEDOWN, wx.WXK_PAGEUP )

#-------------------------------------------------------------------------------
class ConsoleSTC(ConsoleInterface,stc.StyledTextCtrl):
    """ A console control based upon the wx.StyledTextCtrl """
    def __init__(self, parent, engname, cmdhistory):
        """
        Create a console for the engine.

        parent      -   wxWindow parent. 
        engname     -   the engine name to control.
        cmdhistory  -   the CmdHistory object given providing the history 
                        (see history.py for details)
        """
        #-----------------------------------------------------------------------
        #initalise the stc
        #-----------------------------------------------------------------------
        style=wx.CLIP_CHILDREN | wx.SUNKEN_BORDER
        stc.StyledTextCtrl.__init__(self, parent, id=-1, style=style)

        self._initSTC()         #setup stc styles
        self._initEvents()      #bind event handlers

        #-----------------------------------------------------------------------
        # Internals for console
        #-----------------------------------------------------------------------
        #engine is waiting for user input.
        self._reading  = False     

        #engine is in debugging mode.
        self._debugging = False

        #prompt position info used to get the entered line.
        self._promptpos  = 0
        self._promptline = 1

        #Command history - cmdhistory is an object which provides a command 
        #history see history.py for details
        self._previous  = ''        #the last user command sent is stored here
        self._hist = cmdhistory

        #options
        self._showlinenumbers = False
        self._showautocomps = True
        self._showcalltips = True

        #create a calltip window
        self._calltip =  CallTip(self, "", pos=(-1,-1))

        #-----------------------------------------------------------------------
        # Setup console object
        #-----------------------------------------------------------------------
        app = wx.GetApp()
        #store the EngineInterface for convienence
        engmgr = app.toolmgr.get_tool('EngineManager')
        self.eng = engmgr.get_engine(engname)   
        ConsoleInterface.__init__(self, app.msg_bus, engname)

        #register console engtasks with the engine
        self.eng.register_task( console_tasks.get_auto_comps)
        self.eng.register_task( console_tasks.get_call_tip)
        
    #---Initialisation tasks----------------------------------------------------
    def _initSTC(self):
        
        # Set the fonts
        self.faces = { 
              'times': 'Courier New',
              'mono' : 'Courier New',
              'helv' : 'Courier New',
              'other': 'Courier New',
              'size' : 10,
              'backcol'   : '#FFFFFF',
              'calltipbg' : '#FFFFB8',
              'calltipfg' : '#404040'
            }

        self.SetSelForeground(True, wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT))
        self.SetSelBackground(True, wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT))
        
        self.SetKeyWords(0, " ".join(keyword.kwlist))
        self.SetLexer(stc.STC_LEX_PYTHON) #need to print lines longer than 64k!!!

        #whitespaces and tabs
        self.SetViewWhiteSpace(False)
        self.SetViewEOL(False);
        self.SetEOLMode(stc.STC_EOL_LF)

        self.SetTabWidth(4)
        self.SetUseTabs(False)
        self.SetWrapMode(True);
        self.SetEndAtLastLine(False);

        #autocomplete
        self.AutoCompSetDropRestOfWord(True); #delete word after autocomplete point
        self.AutoCompSetIgnoreCase(True);     #ignore case
        self.AutoCompSetAutoHide(False);      #dont show list if no matches
        self.AutoCompStops(" .,;:([)]}\'\"\\<>%^&+-=*/|`"); #chars to stop autocompletes
        self.SetWordChars(" ,;:([)]}\'\"\\<>%^&+-=*/|`"); #set chars that seperate words

        self.SetBufferedDraw(True)

        #calltip for doc string
        self.CallTipSetBackground(self.faces['calltipbg'])
        self.CallTipSetForeground(self.faces['calltipfg'])

        #disable undo/redo
        self.SetUndoCollection(False);

        #create a droptarget
        self.dt = FileDrop()
        self.SetDropTarget(self.dt)  

        #apply the options
        self.LoadOptions()

    def _initEvents(self):
        """Initialise the event handlers"""
        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)#when losing window focus
        
        #update the stc when user is typing.
        self.Bind(wx.stc.EVT_STC_UPDATEUI, self.OnUpdateUI)

        # Assign handler for the context menu and edit events
        self.Bind(wx.EVT_CONTEXT_MENU, self.OnContextMenu)
        self.Bind(wx.EVT_UPDATE_UI, self.OnCMenuUpdateUI,)
        self.Bind(wx.EVT_MENU, self.OnCut, id=wx.ID_CUT)
        self.Bind(wx.EVT_MENU, self.OnCopy, id=wx.ID_COPY)
        self.Bind(wx.EVT_MENU, self.OnPaste, id=wx.ID_PASTE)
        self.Bind(wx.EVT_MENU, self.OnSelectAll, id=wx.ID_SELECTALL)
        self.Bind(wx.EVT_MENU, self.OnClear, id=wx.ID_CLEAR)

        self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy, self)

    #---------------------------------------------------------------------------
    # console methods
    #---------------------------------------------------------------------------
    def prompt(self, more=False):
        """
        Engine prompts for more user commands.
        """
        #check engine is still connected
        if self._connected is False:
            log.warning('Attempt to prompt when engine is disconnected!')
            return

        #disable the prompt.
        if more is None and self._busy is False:
            self._busy = True
            self.SetSelection(self._promptpos - len(self._ps1), self.GetLength())
            self.ReplaceSelection('')
            return 
        elif more is None:
            #not prompting anyway
            return

        #get the prompt
        if more is True:
            prompt = self._ps2
        else:
            prompt = self._ps1

        #write the prompt
        self.SetCurrentPos(self.GetLength()) #set cursor to end position
        self.AddText(prompt)            #add the prompt
        self.EnsureCaretVisible()

        #If this is not a continuation then we want to update the prompt 
        #position to prevent previous commands/output from being edited and we
        #want to add the previous command to the command history
        if more is False:
            #update prompt position
            self._promptline = self.GetCurrentLine()
            self._promptpos = self.GetCurrentPos()
            #add previous command to the history 
            #(as more is false this must have been executed or caused an error)
            self._hist.AddCommand(self._previous[:-1]) # remove trailling \n
            #TODO: engine publish msg about the command it just ran.
            # history adds automatically.

        #set state flags
        self._busy = False        #can't be busy as engine wants a user cmd
        self._debugging = False   #or debugging
        self._reading = False     #or reading


    def prompt_stdin(self,more=False):
        """
        Engine wants to read from the console std_in.
        """
        #check engine is still connected
        if self._connected is False:
            log.warning('Attempt to read_stdin when engine is disconnected!')
            return
    
        #disable the prompt.
        if more is None and self._reading is True:
            self._reading = False
            self.SetSelection(self._promptpos - len(self._ps3), self.GetLength())
            self.ReplaceSelection('')
            return 
        elif more is None:
            #not prompting anyway
            return

        #get the prompt
        if more is True:
            prompt = self._ps2
        else:
            prompt = self._ps3

        #write the prompt
        self.SetCurrentPos(self.GetLength()) #set cursor to end position
        self.AddText(prompt)            #add the prompt
        self.EnsureCaretVisible()

        #we only need to update the prompt position if this is a new read, this
        #is to prevent previous commands/output from being edited. 
        #If this is a continuation we can allow the previous input to be edited.
        if more is False:
            #update prompt position
            self._promptline = self.GetCurrentLine()
            self._promptpos = self.GetCurrentPos()
        #put the console into reading mode
        self._reading = True

    def prompt_debug(self, more=False):
        """
        Called when the debugger pauses, this method prompts for user 
        debugger commands.
        """
        #check engine is still connected
        if self._connected is False:
            log.warning('Attempt to do paused action when engine is disconnected!')
            return

        #disable the debugging prompt.
        if more is None and self._debugging is True:
            self._debugging = False
            self.SetSelection( self._promptpos - len(self._ps4), self.GetLength())
            self.ReplaceSelection('')
            return
        elif more is None:
            #not prompting anyway
            return

        #already showing debug prompt do nothing.
        if self._debugging is True:
            return

        #get the prompt
        if more is True:
            prompt = self._ps2
        else:
            prompt = self._ps4

        #write the prompt
        self.SetCurrentPos(self.GetLength()) #set cursor to end position
        self.AddText(prompt)            #add the prompt
        self.EnsureCaretVisible()

        #we only need to update the prompt position if this is a new read, this
        #is to prevent previous commands/output from being edited. 
        #If this is a continuation we can allow the previous input to be edited.
        if more is False:
            #update prompt position
            self._promptline = self.GetCurrentLine()
            self._promptpos = self.GetCurrentPos()
        self._debugging = True

    def write_stdout(self,string):
        """
        Engine writes to the consoles std_out.
        """
        if self.AutoCompActive():
            self.AutoCompCancel()

        #if we are prompting or reading user input add the string before the 
        #current prompt
        if (self._busy is False) or (self._reading is True) or (self._debugging is True):
            #add the line before the current prompt 
            pos = self.PositionFromLine(self._promptline)
            #check that we are not at the beginning
            if pos==0:
                self.SetCurrentPos(pos)
                self.AddText('\n')
                self._promptpos +=1
                self._promptline+=1

            #add the text before the newline
            self.SetCurrentPos(pos-1)
            self.AddText(string)

            #update the prompt position
            slen = len(string)
            slines = string.count('\n')
            self._promptline =self._promptline+slines
            self._promptpos = self._promptpos+slen
            self.SetAnchor(self.GetLength())
            self.SetCurrentPos(self.GetLength()) #set cursor to end position
            self.EnsureCaretVisible()
            return

        #just add the text at the end.
        self.SetAnchor(self.GetLength())
        self.SetCurrentPos(self.GetLength()) #set cursor to end position
        self.AddText(string)
        self.EnsureCaretVisible()

    def write_stderr(self,string):
        """
        Engine writes to the consoles std_err.
        """
        if self.AutoCompActive():
            self.AutoCompCancel()

        #if we are prompting or reading user input add the string before the 
        #current prompt
        if (self._busy is False) or (self._reading is True) or (self._debugging is True):
            #add the line before the current prompt 
            pos = self.PositionFromLine(self._promptline)
            #check that we are not at the beginning
            if pos==0:
                self.SetCurrentPos(pos)
                self.AddText('\n')
                self._promptpos +=1
                self._promptline+=1

            #add the text before the newline
            self.SetCurrentPos(pos-1)
            self.AddText(string)

            #update the prompt position
            slen = len(string)
            slines = string.count('\n')
            self._promptline =self._promptline+slines
            self._promptpos = self._promptpos+slen
            self.SetAnchor(self.GetLength())
            self.SetCurrentPos(self.GetLength()) #set cursor to end position
            self.EnsureCaretVisible()
            return

        #just add the text at the end.
        self.SetAnchor(self.GetLength())
        self.SetCurrentPos(self.GetLength()) #set cursor to end position
        self.AddText(string)
        self.EnsureCaretVisible()


    def write_debug(self,string):
        """
        Engine writes a debugger message.
        """
        if self.AutoCompActive():
            self.AutoCompCancel()

        if string.endswith('\n') is False:
            string = string+'\n'

        #if we are prompting or reading user input add the string before the 
        #current prompt
        if (self._busy is False) or (self._reading is True) or (self._debugging is True):
            #add the line before the current prompt 
            pos = self.PositionFromLine(self._promptline)
            #check that we are not at the beginning
            if pos==0:
                self.SetCurrentPos(pos)
                self.AddText('\n')
                self._promptpos +=1
                self._promptline+=1

            #add the text before the newline
            self.SetCurrentPos(pos-1)
            self.AddText(string)

            #update the prompt position
            slen = len(string)
            slines = string.count('\n')
            self._promptline =self._promptline+slines
            self._promptpos = self._promptpos+slen
            self.SetAnchor(self.GetLength())
            self.SetCurrentPos(self.GetLength()) #set cursor to end position
            self.EnsureCaretVisible()
            return

        #just add the text at the end.
        self.SetAnchor(self.GetLength())
        self.SetCurrentPos(self.GetLength()) #set cursor to end position
        line,pos = self.GetCurLine()
        if line!='':    #add a newline before if necessary 
            self.AddText('\n')
        self.AddText(string)
        self.EnsureCaretVisible()

    def clear_console(self):
        """
        Clear the console
        """
        if self.AutoCompActive():
            self.AutoCompCancel()
        if self.CallTipActive():
            self.CallTipCancel()

        #prompting so clear all but the current line.
        if (self._busy is False) or (self._reading is True) or (self._debugging is True):
            #get the current prompt
            startpos = self.PositionFromLine(self._promptline)
            prompt = self.GetTextRange(startpos,self._promptpos)
            #get the current line
            endpos = self.GetLength()
            line = self.GetTextRange(self._promptpos,endpos)
            #clear the control and add the prompt and line
            self.ClearAll()
            self.AddText(prompt)
            self._promptline = self.GetCurrentLine()
            self._promptpos = self.GetCurrentPos()
            self.AddText(line)
        else:
            self.ClearAll()

    def exec_source(self,source):
        """
        Execute the source as if entered at the console
        """
        #check engine is still connected
        if self._connected is False:
            log.warning('Attempt to execute source in console when engine is disconnected!')
            return

        #busy  cannnot exectute currently.
        if (self._busy is True):
            log.warning('Attempt to execute source in console when engine is busy!')
            return False

        endpos = self.GetTextLength()
        self.SetSelection(self._promptpos,endpos)
        text = self._AddPrompts(source, self._ps2)
        self.ReplaceSelection(text+'\n\n')
        self.PushLine()

    #---------------------------------------------------------------------------
    # Options
    #---------------------------------------------------------------------------
    def LoadOptions(self):
        """Load and apply the global console options settings"""
        cfg = wx.GetApp().GetConfig()
        cfg.SetPath("Console//")

        #line numbers
        flag = cfg.ReadBool("show_linenumbers",False)
        self.SetShowLineNumbers(flag)

        #show autocomps
        flag = cfg.ReadBool("show_autocomps",True)
        self.SetShowAutoComps(flag)

        #show calltips
        flag = cfg.ReadBool("show_calltips",True)
        self.SetShowCalltips(flag)

        #use syntax highlighting
        flag = cfg.ReadBool("use_syntax_highlight",True)
        self.SetStyles(flag)

    def SetStyles(self,flag):
        faces = self.faces
        # Global default styles for all languages
        self.StyleSetSpec(wx.stc.STC_STYLE_DEFAULT,     "face:%(helv)s,size:%(size)d" % faces)
        self.StyleClearAll()  # Reset all to be like the default

        if flag is True:

            # Global default styles for all languages
            self.StyleSetSpec(wx.stc.STC_STYLE_DEFAULT,     "face:%(helv)s,size:%(size)d" % faces)
            self.StyleSetSpec(wx.stc.STC_STYLE_LINENUMBER,  "back:#C0C0C0,face:%(helv)s,size:%(size)d" % faces)
            self.StyleSetSpec(wx.stc.STC_STYLE_CONTROLCHAR, "face:%(other)s" % faces)
            self.StyleSetSpec(wx.stc.STC_STYLE_BRACELIGHT,  "fore:#FFFFFF,back:#0000FF,bold")
            self.StyleSetSpec(wx.stc.STC_STYLE_BRACEBAD,    "fore:#000000,back:#FF0000,bold")

            # Python styles
            # Default 
            self.StyleSetSpec(wx.stc.STC_P_DEFAULT, "fore:#000000,face:%(helv)s,size:%(size)d" % faces)
            # Comments
            self.StyleSetSpec(wx.stc.STC_P_COMMENTLINE, "fore:#007F00,face:%(other)s,size:%(size)d" % faces)
            # Number
            self.StyleSetSpec(wx.stc.STC_P_NUMBER, "fore:#007F7F,size:%(size)d" % faces)
            # String
            self.StyleSetSpec(wx.stc.STC_P_STRING, "fore:#7F007F,face:%(helv)s,size:%(size)d" % faces)
            # Single quoted string
            self.StyleSetSpec(wx.stc.STC_P_CHARACTER, "fore:#7F007F,face:%(helv)s,size:%(size)d" % faces)
            # Keyword
            self.StyleSetSpec(wx.stc.STC_P_WORD, "fore:#00007F,bold,size:%(size)d" % faces)
            # Triple quotes
            self.StyleSetSpec(wx.stc.STC_P_TRIPLE, "fore:#7F0000,size:%(size)d" % faces)
            # Triple double quotes
            self.StyleSetSpec(wx.stc.STC_P_TRIPLEDOUBLE, "fore:#7F0000,size:%(size)d" % faces)
            # Class name definition
            self.StyleSetSpec(wx.stc.STC_P_CLASSNAME, "fore:#0000FF,bold,underline,size:%(size)d" % faces)
            # Function or method name definition
            self.StyleSetSpec(wx.stc.STC_P_DEFNAME, "fore:#007F7F,bold,size:%(size)d" % faces)
            # Operators
            self.StyleSetSpec(wx.stc.STC_P_OPERATOR, "size:%(size)d" % faces)
            # Identifiers
            self.StyleSetSpec(wx.stc.STC_P_IDENTIFIER, "fore:#000000,face:%(helv)s,size:%(size)d" % faces)
            # Comment-blocks
            self.StyleSetSpec(wx.stc.STC_P_COMMENTBLOCK, "fore:#CC0000,size:%(size)d" % faces)
            # End of line where string is not closed
            self.StyleSetSpec(wx.stc.STC_P_STRINGEOL, "fore:#000000,face:%(mono)s,back:#E0C0E0,eol,size:%(size)d" % faces)

        self.SetCaretForeground("BLUE")

    def SetShowLineNumbers(self,flag):
        """Show line numbers"""
        self._showlinenumbers = flag
        if flag:
            self.SetMarginType(1, wx.stc.STC_MARGIN_NUMBER)
            self.SetMarginWidth(1, 40)
        else:
            #no line numbers so set margin width
            self.SetMarginWidth(1, 0)
            self.SetMarginType(1, 0)

    def SetShowAutoComps(self,flag):
        """Automattically show autocompletes"""
        self._showautocomps = flag

    def SetShowCalltips(self,flag):
        """Automatically show calltips"""
        self._showcalltips = flag
    #---------------------------------------------------------------------------
    #  User entry methods
    #---------------------------------------------------------------------------
    def PushLine(self):
        """ Push an entered line to the engine"""
        if self._connected is False:
            log.warning('Attempt to push line to engine when disconnected!')
            return

        #reset history search/position when the line is changed.
        self._hist.SetSearchString()

        #get the line from the stc
        endpos = self.GetLength()
        line = self.GetTextRange(self._promptpos,endpos)

        #1) console is reading from StdIn
        if self._reading:
            #unset flag
            self._reading = False
            #strip continuation prompts
            line = self._StripPrompts(line,self._ps2)
            # In case the command is unicode try encoding it
            if type(line) == unicode:
                try:
                    line = line.encode(wx.GetDefaultPyEncoding())
                except UnicodeEncodeError:
                    pass # otherwise leave it alone

            #send the line to the engine
            self.push(line)
            return

        #2) console is debugging
        if self._debugging:
            #unset the flag
            self._debugging = False
            #strip continuation prompts
            line = self._StripPrompts(line,self._ps2)
            # In case the command is unicode try encoding it
            if type(line) == unicode:
                try:
                    line = line.encode(wx.GetDefaultPyEncoding())
                except UnicodeEncodeError:
                    pass # otherwise leave it alone

            #send the line to the engine
            self.push(line)
            return

        #3) Prompting for command
        else:
            self._busy = True #this will be unset by the next call to Prompt
            #strip any continuation prompts from the line
            line = self._StripPrompts(line,self._ps2)

            #store the line, incase we want to add it to the command history.
            #this is done in the next call to Prompt that is not a continuation.
            self._previous = line

            #send line to engine - the engine will process this line and send a 
            #message to: 
            #   prompt if a new command is needed.
            #   read_stdin if input from the user is needed.
            #   write_stdout to display output
            #   write_stderr to display errors
            self.push(line)
            return

    def _AddPrompts(self,line,prompt):
        """Adds continuation prompts to a multiline command"""
        line=line.replace('\n','\n'+prompt)
        return line

    def _StripPrompts(self,line,prompt):
        """Remove continuation prompts from a multiline command"""
        line = line.replace('\n'+prompt, '\n')
        return line    
    #---------------------------------------------------------------------------
    # Interface methods
    #---------------------------------------------------------------------------
    def Cut(self):
        """Remove selection and place it on the clipboard."""
        if self.CanEdit() is True:
            if self.AutoCompActive():
                self.AutoCompCancel()
            if self.CallTipActive():
                self.CallTipCancel()
            stc.StyledTextCtrl.Cut(self)
            #reset history search/position when the line is changed.
            self._hist.SetSearchString()
        else:
            #line = self.GetSelectedText()
            #line = self._StripPrompts(line,self._ps1)
            #line = self._StripPrompts(line,self._ps2)
            #line = self._StripPrompts(line,self._ps3)
            #data = wx.TextDataObject(line)
            #print line,data
            #self._clip(data)
            stc.StyledTextCtrl.Copy(self)

    def Paste(self):
        if self.CanEdit() is True:
            #close autocomps and call tip
            if self.AutoCompActive():
                self.AutoCompCancel()
            if self.CallTipActive():
                self.CallTipCancel()

            #get the current pos, paste the text then insert prompts
            startpos = self.GetCurrentPos()
            stc.StyledTextCtrl.Paste(self)
            endpos = self.GetLength()
            line = self.GetTextRange(startpos,endpos)
            
            #first strip any prompts
            line = self._StripPrompts(line,self._ps1)
            line = self._StripPrompts(line,self._ps2)
            line = self._StripPrompts(line,self._ps3)

            #add continuation prompts back in if needed
            if self._reading:
                line = self._AddPrompts(line,self._ps3)
            else:
                line = self._AddPrompts(line,self._ps2)

            #put line into stc
            self.SetSelection(startpos, endpos)
            self.ReplaceSelection(line)
            #reset history search/position when the line is changed.
            self._hist.SetSearchString()

    def Clear(self):
        """Clears the console"""
        self.clear_console()

    def ClearCommand(self):
        """Clears the current command"""
        if (self._busy is True) and ((self._reading is False) or (self._debugging is False)):
            return

        #check engine is still connected
        if self._connected is False:
            log.warning('Attempt to clear command when engine is disconnected!')
            return

        endpos = self.GetTextLength()
        self.SetSelection(self._promptpos, endpos)
        self.ReplaceSelection('')
        #reset history search/position when the line is changed.
        self._hist.SetSearchString()

    #---Command history---------------------------------------------------------
    def CmdHist_ReplacePrevious(self):
        """Move to previous command in command history"""
        #check not busy...
        if self._busy:
            return
        
        #check engine is connected
        if self._connected is False:
            return

        #set the search string to the current line contents if not already in 
        # the history.
        if self._hist.GetPosition()==-1:
            endpos = self.GetTextLength()
            searchstr = self.GetTextRange(self._promptpos,endpos)
            searchstr = self._StripPrompts(searchstr,self._ps2)
            self._hist.SetSearchString(searchstr)

        #get the previous matching command
        cmd = self._hist.GetPreviousCommand()

        #replace text
        endpos = self.GetTextLength()
        self.SetSelection(self._promptpos,endpos)
        cmd = self._AddPrompts(cmd,self._ps2)
        self.ReplaceSelection(cmd)

    def CmdHist_ReplaceNext(self):
        """Move to next command in command history """
        #check not busy...
        if self._busy:
            return

        #check engine is connected
        if self._connected is False:
            return

        #set the search string to the current line contents if not already in 
        # the history.
        if self._hist.GetPosition()==-1:
            endpos = self.GetTextLength()
            searchstr = self.GetTextRange(self._promptpos,endpos)
            searchstr = self._StripPrompts(searchstr,self._ps2)
            self._hist.SetSearchString(searchstr)

        #get the previous matching command
        cmd = self._hist.GetNextCommand()
        
        #replace text
        endpos = self.GetTextLength()
        self.SetSelection(self._promptpos,endpos)
        cmd = self._AddPrompts(cmd,self._ps2)
        self.ReplaceSelection(cmd)

    def CmdHist_Clear(self):
        """Clears the command history"""
        self._hist.Clear()

    #---Autocompletes/Calltips--------------------------------------------------
    def OpenAutoComplete(self):
        """Open the autocomplete list"""
        if (self._busy is True) and (self._debugging is False):
            return

        #check engine is still connected
        if self._connected is False:
            log.warning('Attempt to open autocomps when engine is disconnected!')
            return

        # check if the current position is beyond the current prompt
        currentpos = self.GetCurrentPos()
        if  currentpos < self._promptpos:
            return

        #get the word
        curpos = self.GetCurrentPos()
        wordstart = self.WordStartPosition(curpos,False)
        if wordstart<self._promptpos:
            wordstart=self._promptpos
        word = self.GetTextRange(wordstart,curpos)

        #check it's not a number
        try:
            float(word)
            #is a number return
            return
        except:
            #is not a number carry on
            pass

        #get autocompletes
        wlist,use_n_chars= self.eng.run_task('get_auto_comps',(word,))
        if len(wlist)!=0:
            #convert this list to a string
            stringlist = " ".join(wlist)
            #finally show the list
            self.AutoCompShow(use_n_chars,stringlist)

    def OpenCallTip(self):
        """Opens the calltip"""
        if (self._busy is True) and (self._debugging is False):
            return

        #check engine is still connected
        if self._connected is False:
            log.warning('Attempt to open calltip when engine is disconnected!')
            return

        #get the word
        curpos = self.GetCurrentPos()
        wordstart = self.WordStartPosition(curpos-1,False) #-1 as '(' gets in the way)
        word = self.GetTextRange(wordstart,curpos-1)

        #get the calltip
        tip = self.eng.run_task('get_call_tip',(word,))
        #show it
        if tip is not None:
            self.CallTipShow(curpos,tip)

    def CallTipShow(self,pos,tip):
        """Show the call tip"""
        if (self._busy is True) and (self._debugging is False):
            return

        #close autocomps
        if self.AutoCompActive():
            self.AutoCompCancel()

        if self._calltip.IsLocked():
            return
        #hide it
        self._calltip.Hide()

        #no tip found
        if tip is None:
            return
        
        #set call tip
        self._calltip.SetTip(tip)

        #calculate where to show the calltip
        wpos = self.PointFromPosition(pos)
        spos = self.ClientToScreen(wpos)
        size = self.GetSize()
        maxy = wx.SystemSettings_GetMetric(wx.SYS_SCREEN_Y)
        maxx = wx.SystemSettings_GetMetric(wx.SYS_SCREEN_X)
        spos.x = spos.x+50 #shift over a bit
        if spos.x > maxx:
            spos.x = maxx-size.x
            spos.y = spos.y-size.y-50
        if spos.y>maxy-size.y:
            spos.y = maxy-size.y
        if spos.y<0:
            spos.y = 0
        self._calltip.SetPosition(spos)
        self._calltip.Show()

    def CallTipActive(self):
        return self._calltip.IsShown()

    def CallTipCancel(self):
        self._calltip.Hide()

    #---helper methods----------------------------------------------------------
    def CanEdit(self):
        """Return true if editing should succeed."""
        # check user input is allowed;
        if (self._busy is True) and (self._reading is False) and (self._debugging is False):
            return False

        #check engine is still connected
        if self._connected is False:
            return

        # check if the current position is beyond the current prompt
        currentpos = self.GetCurrentPos()
        if  currentpos < self._promptpos:
            return False

        # check that if text is selected that it does not extend beyond prompt
        if (self.GetSelectionStart()<self._promptpos):
            return False
        if (self.GetSelectionEnd()<self._promptpos):
            return False
        return True

    def IsSelection(self):
        """Returns true if text is selected"""
        result = self.GetSelectionStart() != self.GetSelectionEnd()
        return result

    #---------------------------------------------------------------------------
    # Event handlers
    #---------------------------------------------------------------------------
    def OnKeyDown(self, event):
        """Key down event handler."""

        #get key info
        key         = event.GetKeyCode()
        controlDown = event.ControlDown()
        altDown     = event.AltDown()
        shiftDown   = event.ShiftDown()
        currentpos  = self.GetCurrentPos()
        currentline = self.GetCurrentLine()
        endpos      = self.GetLength()
        selecting   = self.IsSelection()   
        edit        = self.CanEdit() #check if we can edit

        #If the auto-complete window is up let it handle the key
        if self.AutoCompActive():
            event.Skip()
            return
    
        #close any open calltips
        if self.CallTipActive():
            self.CallTipCancel()

        #now check for keys
        # Return (Enter) pressed - Process the line
        if key in [wx.WXK_RETURN, wx.WXK_NUMPAD_ENTER]:
            #check on which line the key was pressed
            if edit is False:
                #key pressed in non-editable region move to endpos
                self.SetAnchor(endpos)
                self.SetCurrentPos(endpos)
                return

            elif currentpos !=endpos:
                #enter pressed  not at the end so insert a new line
                if self._reading:
                    self.AddText('\n'+self._ps3)
                else:
                    self.AddText('\n'+self._ps2)
                return

            else:#enter pressed on last line process it.
                #move to the end of the line and add \n
                self.SetAnchor(endpos)
                self.SetCurrentPos(endpos)
                self.AddText('\n')
                self.PushLine()
                return

        # Cut to the clipboard.
        elif (controlDown and key in (ord('X'), ord('x')) ) \
                 or (shiftDown and key == wx.WXK_DELETE):
            self.Cut()

        # Copy selection to the clipboard removing prompts
        elif controlDown and not shiftDown \
                 and key in (ord('C'), ord('c'), wx.WXK_INSERT):
            self.Copy()

        # Paste from the clipboard.
        elif (controlDown and not shiftDown and key in (ord('V'), ord('v'))) \
                 or (shiftDown and not controlDown and key == wx.WXK_INSERT):
            self.Paste()

        # Command history keys
        elif (controlDown and key in [wx.WXK_UP,wx.WXK_NUMPAD_UP] ):
            self.CmdHist_ReplacePrevious()
        elif (controlDown and key in [wx.WXK_DOWN,wx.WXK_NUMPAD_DOWN]):
            self.CmdHist_ReplaceNext()
        
        # Clear the current, unexecuted command.
        elif(key==wx.WXK_ESCAPE and self.CallTipActive()):
            event.Skip()
        elif(key==wx.WXK_ESCAPE and self.CallTipActive()==False):
            self.ClearCommand()

        # manually invoke AutoComplete
        elif controlDown and key == wx.WXK_SPACE:
            self.OpenAutoComplete()

        # backspace
        elif key == wx.WXK_BACK:
            if (currentpos>(self._promptpos)):
                event.Skip()

        # Don't toggle between insert mode and overwrite mode.
        elif key == wx.WXK_INSERT:
            pass

        elif (key==wx.WXK_SPACE) and (edit is True):
            event.Skip()

        #catch the home key and shift down
        elif edit is True and key == wx.WXK_HOME:
            if shiftDown is True:
                self.SetAnchor(currentpos)
            else:
                self.SetAnchor(self._promptpos)
            self.SetCurrentPos(self._promptpos)

        #Allow navigation in uneditable regions
        elif edit is False and key in NAVKEYS: 
            event.Skip()

        #key press was not caught, so pass it on to the stc...
        elif edit==True:
            event.Skip(); #allow the stc to process the key if ok to edit

    def OnChar(self,event):
        """handles single keypresses - only recieves an event if onKeyDown skips"""
        #get key
        key = event.GetKeyCode()

        if (key==ord('.')):
            self.AddText('.')
            if self._showautocomps is True:
                self.OpenAutoComplete()
        elif (key==ord('(')):
            self.AddText('(')
            if self._showcalltips is True:
                self.OpenCallTip()
        elif key==13: #catch stray '\r' on windows? not on gtk
             pass
        else:
            #reset history search/position when the line is changed.
            self._hist.SetSearchString()
            event.Skip()

    def OnKillFocus(self,event):
        """Close any popups when focus is lost"""
        if self.AutoCompActive():
            self.AutoCompCancel()
        if self.CallTipActive():
            self.CallTipCancel()

    def OnUpdateUI(self, evt):
        # check for matching braces
        braceAtCaret = -1
        braceOpposite = -1
        charBefore = None
        caretPos = self.GetCurrentPos()

        if caretPos > 0:
            charBefore = self.GetCharAt(caretPos - 1)
            styleBefore = self.GetStyleAt(caretPos - 1)

        # check before
        if charBefore and chr(charBefore) in "[]{}()" and styleBefore == wx.stc.STC_P_OPERATOR:
            braceAtCaret = caretPos - 1

        # check after
        if braceAtCaret < 0:
            charAfter = self.GetCharAt(caretPos)
            styleAfter = self.GetStyleAt(caretPos)

            if charAfter and chr(charAfter) in "[]{}()" and styleAfter == wx.stc.STC_P_OPERATOR:
                braceAtCaret = caretPos

        if braceAtCaret >= 0:
            braceOpposite = self.BraceMatch(braceAtCaret)

        if braceAtCaret != -1  and braceOpposite == -1:
            self.BraceBadLight(braceAtCaret)
        else:
            self.BraceHighlight(braceAtCaret, braceOpposite)

    #Context menu events
    def OnContextMenu(self, evt):
        """Create and display a context menu for the shell"""
        menu = wx.Menu()
        menu.Append(wx.ID_CUT, "Cut")
        menu.Append(wx.ID_COPY, "Copy")
        menu.Append(wx.ID_PASTE, "Paste")
        menu.Append(wx.ID_CLEAR, "Clear")
        menu.AppendSeparator()
        menu.Append(wx.ID_SELECTALL, "Select All")
        self.PopupMenu(menu)

    def OnCMenuUpdateUI(self, evt):
        """disable the context menu actions that are not possible"""
        id = evt.Id
        if id == wx.ID_CUT:
            evt.Enable(self.CanEdit() and self.IsSelection())
        elif id == wx.ID_COPY:
            evt.Enable(self.IsSelection())
        elif id == wx.ID_PASTE:
            evt.Enable(self.CanEdit() and self.CanPaste())

    def OnCut(self,evt):
        """Cut event handler"""
        self.Cut()

    def OnCopy(self,evt):
        """Copy event handler"""
        self.Copy()

    def OnPaste(self,evt):
        """Paste event handler"""
        self.Paste()

    def OnClear(self,evt):
        """Clear event handler"""
        self.Clear()

    def OnSelectAll(self,evt):
        """Select all menu handler"""
        self.SelectAll()
    
    def OnDestroy(self, event):
        self.close()
        event.Skip()
