"""
Main console tool

The matlab style main window and console management.


Console Management:
------------------
1a) The EngineManager tool launches/creates an engine and published an 
Engine.Started message
    or
1b) Another tool/external program connects to the message bus as an engine (i.e.
as 'Engine.engname'), EngineManager registers the engine and publishes an 
Engine.Started message

2) The console tool listens for Engine.Started messages and creates a console 
    to control the new engine.

3) The console itself sends a message to the Engine (ENG_SET_CONSOLE) to recieve
    all stdIO and prompt requests from the engine.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('Console')
log.setLevel(logging.DEBUG)

#---Imports---------------------------------------------------------------------
import wx

from ptk_lib.message_bus.mb_node import MBNode

from ptk_lib.core_tools.engine_manager import engman_messages

#Application components
from ptk_lib.tool_manager import Tool

#console components
from history import CmdHistory
from console_frame import ConsoleFrame
from console_settings import ConsoleSettingsPanel, HistorySettingsPanel
import console_messages
import console_icons

#---the tools class-------------------------------------------------------------
class Console(Tool):
    name = 'Console'
    descrip = 'Core tool implementing the main console window'
    version = 11.01        
    author = 'T.Charrett'
    requires = ['TaskIcon']           
    core = True            
    icon = console_icons.console32
    
    def __init__(self):
        Tool.__init__(self)
        log.info('Initialising tool')

        #-----------------------------------------------------------------------
        # Engine/Console management
        #-----------------------------------------------------------------------
        #dictionary of active console 
        self.consoles = {}  #{engname: ConsoleInterface object}
       
        #the command history store object
        self.cmdhistory = CmdHistory(self)

        #Register message handlers
        self.msg_node = MBNode(self.msg_bus, 'Console')
        self.msg_node.set_handler(console_messages.SHOW,self.msg_show)
        self.msg_node.set_handler(console_messages.HIDE,self.msg_hide)
        self.msg_node.set_handler(console_messages.CLEAR,self.msg_clear_console)
        self.msg_node.set_handler(console_messages.EXEC_SOURCE,self.msg_exec_source)

        #subscribe to messages
        self.msg_node.subscribe('App.Init',self.msg_app_init)
        self.msg_node.subscribe('App.Exit',self.msg_app_exit)

        #subscribe to engine manager messages to make sure we know if an engine starts/closes
        self.msg_node.subscribe(engman_messages.ENGINE_STARTED,self.msg_eng_start)  

        #-----------------------------------------------------------------------
        # GUI parts
        #-----------------------------------------------------------------------
        #create the console frame
        self.frame = ConsoleFrame(self)

        #Add taskbar menu item
        taskicon = self.toolmgr.get_tool('TaskIcon')
        bmp = console_icons.console16.GetBitmap()
        taskicon.add_menu_item( wx.NewId(), 'Open the Console',
                                'Open the Console window', 
                                self.on_show, bitmap=bmp)  

        #Add console settings panel
        taskicon.add_settings_item( 'Console', None, 
                                    console_icons.console16.GetBitmap())
        taskicon.add_settings_item( 'Console\\Display', ConsoleSettingsPanel)
        taskicon.add_settings_item( 'Console\\Command history', HistorySettingsPanel)

        log.info('Done Initialising tool')

    #---Interfaces--------------------------------------------------------------
    #gui interfaces
    def show_console(self):
        """
        Show the Console frame
        """
        self.frame.Show()
        self.frame.Raise()

    def hide_console(self):
        """
        Hide the console frame
        """
        self.frame.Hide()

    def add_menu_item(self, menu, id, string, helpstring, evthndlr, bitmap=None, pos=None):
        """
        Add a menu item to the menu.
            menu - string description of menu ('Tools', 'File'. 'Edit' etc)
            string - string to display
            helpstring - help string for item
            evthndlr - wx event handler callable
            bitmap  - bitmap to use / None for no bitmap
            pos - Position in menu, inserted here if specified.
        """
        menu = self.frame.GetMenu(menu)
        item = wx.MenuItem(menu, id, string,helpstring)
        self.frame.Bind(wx.EVT_MENU, evthndlr, id=id)
        if bitmap is not None:
            item.SetBitmap(bitmap)
        if pos is None:
            pos = menu.GetMenuItemCount()
        menu.InsertItem(pos,item)

    def apply_console_options(self):
        """
        Reload and apply the global console settings to all open consoles. 
        This is used to immediately apply the console prefrences without needing
        to restart the app
        """
        #console settings
        for con in self.frame.GetAllConsoles():
            con.LoadOptions()

        #command history settings
        self.cmdhistory.LoadOptions()

    #---message handlers--------------------------------------------------------
    def msg_show(self,msg):
        """
        Message handler to show the console frame
        """
        self.frame.Show()
        self.frame.Raise()

    def msg_hide(self,msg):
        """
        Message handler for Console.Hide
        """
        self.frame.Hide()

    def msg_app_init(self,msg):
        """
        Listener for App.Init message
        Sent when application starts
        """
        #load the main window layouts (in aui mixin class)
        self.frame.LoadLayouts()

    def msg_app_exit(self,msg):
        """
        Listener for App.Exit message
        Save settings on application exit
        """
        #save the main window layouts (in aui mixin class)
        self.frame.SaveLayouts()
        #save command history
        #TODO: move this to cmdhistory tool - use msg to add commands directly 
        # from engine and use message to get history.
        self.cmdhistory.SaveOptions()
        self.cmdhistory.SaveHistory()

    #console management
    def msg_eng_start(self,msg):
        """
        An ENGINE_STARTED mesage is published by the EngineManager to advertise when 
        they a new engine has started.
        Create a console to handle it.   
        """
        engname, engtype = msg.get_data()
        log.info('Adding console from new engine: '+str(engname) + ' type: '+str(engtype)) 

        engmgr = self.toolmgr.get_tool('EngineManager')
        eng = engmgr.get_engine(engname)

        #create a new console
        console = self.frame.AddConsole(engname, engtype)

    #console mesages
    def msg_clear_console(self,msg):
        """
        Message handler for Console.Clear message
        """
        engname = msg.get_data()[0]
        if engname is None:
            console = self.frame.GetCurrentConsole()
        else:
            console = self.frame.GetConsole(engname)

        if console is not None:
            console.clear_console()

    def msg_exec_source(self,msg):
        """
        Message handler for Console.ExecSource message
        runs the source in an engines console.
        """
        #get the source
        engname, source, = msg.get_data() 
 
        if engname is None:
            console = self.frame.GetCurrentConsole()
        else:
            console = self.frame.GetConsole(engname)

        if console is None:
            return

        #show the window
        if self.frame.IsShown() is False:
            self.frame.Show()
        if self.frame.IsIconized():
            self.frame.Restore()
        self.frame.Raise()

        #exec source in the console
        console.exec_source(source)

    #---others------------------------------------------------------------------
    def on_show(self, event):
        """wx event handler for taskbar menu item"""
        self.frame.Show()
        self.frame.Raise()

