"""
CommandHistory

A single CmdHistory object is created and used by all consoles, this allows 
all the engines to share the global command history.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('Console')

#---imports---------------------------------------------------------------------
import os
import wx
import shelve
import time
from ptk_lib.misc import USERDIR

#---CmdHistory class------------------------------------------------------------
class CmdHistory():
    def __init__(self, tool):
        """
        The command history object is used to load/save a global command history
        that is shared among all engines. 
        """
        self.tool = tool 

        #the command history file
        self.file = 'cmdhist.hist'

        #the command history and max length
        self.hist = []
        self.histmax=5000

        #attributes for searching in history
        self.offset = -1
        self.cursearch = ''
    
        #load settings
        self.LoadOptions()

        #load history
        self.LoadHistory()

        #add session started tag
        s = '#Session started : '+time.ctime()
        self.hist.insert(0,s)

    def LoadOptions(self):
        """
        Loads/Reloads the settings for the command history
        """
        cfg = wx.GetApp().GetConfig()
        cfg.SetPath("Console//")
        self.histmax = cfg.ReadInt("max_history_length",5000)
        
        app = wx.GetApp()
        self.file = cfg.Read("history_file",USERDIR + 'cmdhist.hist')

    def SaveOptions(self):
        """
        Save the options
        """
        cfg = wx.GetApp().GetConfig()
        cfg.SetPath("Console//")
        cfg.WriteInt("max_history_length",self.histmax)
        cfg.Write("history_file",self.file)

    def SetFile(self,filename): 
        """
        Set the path to store the current command history.
        """
        self.file = filename
        self.SaveOptions()

    def SaveHistory(self,file=None):
        """
        Save the history to file if None specified the current file is saved
        """
        log.info('Saving command history')
        if file is None:
            file=self.file
        #open shelf file
        shelf = shelve.open(file)
        shelf['history']=self.hist
        #close shelf
        shelf.close()

    def LoadHistory(self,file=None):
        """
        Load the history from a file if None specified the current file is loaded
        """
        log.info('Loading command history')
        if file is None:
            file=self.file
        try:
            shelf = shelve.open(file)
        except:
            app = wx.GetApp()
            self.file = USERDIR + 'cmdhist.shelf'
            shelf = shelve.open(self.file)

        self.hist = shelf.get('history',[])
        shelf.close()

    def ImportHistory(self,file):
        """
        Import the commands from a file into the history
        """
        #check file exists
        if os.path.exists(file) is False:
            return
        root,ext = os.path.splitext(file)
        if ext==".py":
            #load from source
            f = open(file,'r')
            lines = f.readlines()
            f.close()
        else:
            #open as shelf
            try:
                shelf = shelve.open(file)
                lines = shelf.get('history',[])
                lines.reverse()
                shelf.close() 
            except:
                lines = []

        #add commands after removing any extra \n
        for line in lines:
            if line[-1]=='\n':
                line = line[:-1]
            self.hist.insert(0, line )
        
        #check history length
        while len(self.hist)>self.histmax:
            self.hist.pop(-1)
        
        #publish change messaged
        self.tool.msg_node.publish_msg('Console.CmdHistory.Changed',())

    def SetSearchString(self,searchstr=''):
        """
        Sets the string to search for in the history (also resets position)
        """
        if type(searchstr) == unicode:
            try:
                searchstr = searchstr.encode(wx.GetDefaultPyEncoding())
            except UnicodeEncodeError:
                pass # otherwise leave it alone
        self.cursearch = searchstr
        self.offset=-1

    def GetPreviousCommand(self):
        """
        Get the previous(older)command from the history.
        If the search string has been set the hsitory will be searched and the 
        previous match returned.
        """
        self.offset = self.offset+1
        cmd = self.GetCommand(self.offset,self.cursearch)
        if cmd is None:
            #no matches, reset offset and return the search string
            self.offset = -1
            return self.cursearch
        return cmd

    def GetNextCommand(self):
        """
        Get the next (more recent) command from the history. 
        If the search string has been set the hsitory will be searched and the 
        next match returned.
        """
        self.offset = self.offset-1
        cmd = self.GetCommand(self.offset,self.cursearch)
        if cmd is None:
            #no matches, reset offset and return the search string
            self.offset = -1
            return self.cursearch
        return cmd

    def GetPosition(self):
        return self.offset

    def SetPosition(self,n=-1):
        """
        Set the position in the history
        """
        self.offest = n

    def AddCommand(self,cmdstring):
        """
        Add a command to the history.
        """
        #check that the command is not the same as the last command
        if len(self.hist)==0:
            lastcmd=None
        else:
            lastcmd=self.hist[0]
        
        #add to history if not the same as last command
        if (cmdstring!='') and (cmdstring!=lastcmd):
            self.hist.insert(0, cmdstring )
            #publish command added message
            self.tool.msg_node.publish_msg('Console.CmdHistory.Add',(cmdstring,))

        #check history length
        while len(self.hist)>self.histmax:
            self.hist.pop(-1)

    def GetCommand(self,n,startstr=''):
        """
        Get the nth command from the history. If startstr is given then the 
        history will be searched for commands beginning startstr and the nth 
        match returned. If n is greater than matching commmands or the history 
        length then None will be returned.
        """
        #return none from negative n
        if n <0:
            return None

        #by pass the search if startstr is empty
        if startstr == '':
            #check n is not greater than history length
            if n>len(self.hist)-1:
                return None
            #return the command at position n
            return self.hist[n]

        #loop over history until the nth item
        c=-1
        for cmd in self.hist:
            #search for startstr and add to the results list
            if cmd.startswith(startstr):
                c+=1
            #check if n has been reached
            if c==n:
                return cmd
        #if n was never reached return None
        return None

    def ClearHistory(self):
        """Clears the command history"""
        self.hist=[]
        self.SaveHistory()
        #publish hsitory cleared message
        self.tool.msg_node.publish_msg('Console.CmdHistory.Changed',())

