"""
This contains the main editor window:

EditorFrame - The main editor window.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('Editor')
#log.setLevel(logging.DEBUG)

#---imports---------------------------------------------------------------------
import wx                           #for gui elements
import wx.aui as aui

from ptk_lib import VERSION
from ptk_lib.controls import auiframe
from ptk_lib.icons import common22
from ptk_lib.misc import open_help
from ptk_lib.core_tools.fileio import FileDrop
from ptk_lib.core_tools.taskicon import PTKInfoDialog

import editor_icons
from editor_notebook import EditorNotebook
from search_panel    import SearchPanel

#---Editor frame ---------------------------------------------------------------
class EditorFrame(auiframe.AUIFrame):
    """Top level Editor window"""
    def __init__(self, tool):
        """Create editor window"""
        auiframe.AUIFrame.__init__(self, None, -1, "PTK Editor",size=(800,600),pos=(-1,-1))

        #store a reference to the editor tool in the frame.
        self.tool = tool

        #set the window icons
        ib = wx.IconBundle()
        ib.AddIcon(editor_icons.editor16.GetIcon())
        ib.AddIcon(editor_icons.editor32.GetIcon())
        ib.AddIcon(editor_icons.editor48.GetIcon())
        self.SetIcons(ib)

        #create statusbar
        self.CreateStatusBar()
        self.SetStatusText("Python toolkit v"+VERSION)

        #create the menu
        self._CreateMenu()
        self.SetToolbarsMenu(self.menubar.toolbars_menu)

        #create the main tool bar
        self._CreateTools()

        #create the editor notebook then add to the window
        self._CreateNotebook()

        #create search panel
        self._CreateSearchPane()  

        #list of recent files:
        self.filehistory = wx.FileHistory(9)
        cfg = wx.GetApp().GetConfig()
        cfg.SetPath("Editor//")

        self.filehistory.Load(cfg)
        self.filehistory.UseMenu(self.menubar.recent_menu)
        self.filehistory.AddFilesToMenu()
        self.Bind(wx.EVT_MENU_RANGE, self.OnFileHistory, id=wx.ID_FILE1, id2=wx.ID_FILE9)

        #create a droptarget
        self.dt = FileDrop(self.notebook.OpenFile)
        self.SetDropTarget(self.dt) 

        #get window close events
        self.Bind( wx.EVT_CLOSE, self.OnClose)

        #load the frame settings adding layout checkitems to the view menu
        self.SetSavePath('Editor')
        self.SetLayoutsMenu(self.menubar.view_menu)

        #update aui manager
        self.auimgr.Update()

        log.info('Done Initialising Editor Frame')

    def _CreateMenu(self):
        self.menubar = EditorMenu(self)
        #finally create the menubar
        self.SetMenuBar(self.menubar)
 
    def _CreateTools(self):
        self.tools = EditorTools(self)
        #add this toolbar to aui manager
        pane = (aui.AuiPaneInfo().Name('Editor toolbar')
                    .Caption('Editor toolbar').ToolbarPane().CloseButton(True)
                    .CaptionVisible(False)
                    .DestroyOnClose(False).Top().Row(0).LeftDockable(False)
                    .RightDockable(False))

        #add to the window using aui manager
        self.AddToolbar(self.tools,'Editor toolbar',pane,helpstring = 'Show/Hide the editor toolbar')

        self.formattools = FormatTools(self)
        #add this toolbar to aui manager
        pane = (aui.AuiPaneInfo().Name('Format toolbar')
                    .Caption('Format toolbar').ToolbarPane().CloseButton(True)
                    .CaptionVisible(False)
                    .DestroyOnClose(False).Top().Row(0).Position(1).LeftDockable(False)
                    .RightDockable(False))

        #add to the window using aui manager
        self.AddToolbar(self.formattools,'Format toolbar',pane,helpstring = 'Show/Hide the Format toolbar')

    def _CreateNotebook(self):
        self.notebook = EditorNotebook(self)
        #setup how to display this in the aui
        pane = aui.AuiPaneInfo()
        name='Notebook'
        pane.Name(name) #id name
        pane.CentrePane()
        #add the pane
        self.auimgr.AddPane(self.notebook,pane)

    def _CreateSearchPane(self):
        ctrl = SearchPanel(self)
        pane = aui.AuiPaneInfo()
        name='Find and Replace'
        pane.Name(name) #id name
        pane.Caption('Find and Replace')
        pane.CloseButton(True) #close button
        pane.DestroyOnClose(False)
        pane.Floatable(False)
        pane.Resizable(False)
        pane.MinSize( (-1,65))
        pane.MaxSize( (-1,65))
        pane.Bottom()
        pane.Hide()
        #add the pane
        self.auimgr.AddPane(ctrl,pane)
        #store a ref to the pane - needs to be after adding the pane!
        self.search=self.auimgr.GetPane(name)

    #---Interface methods-------------------------------------------------------
    def OpenFile(self,filepath):
        """Opens the file in the editor"""
        self.notebook.OpenFile(filepath)
        self.Show()
        self.Raise()

    def GetMenu(self,menu):
        """
        Return a reference to the menu/menubar:
        menu = 'menubar','file','edit','format','view','toolbars','tools','help'
        """
        menu = menu.lower()
        if menu=='menubar':
            return self.menubar
        if menu=='file':
            return self.menubar.file_menu  #file menu
        if menu=='edit':
            return self.menubar.edit_menu  #edit menu
        if menu=='format':
            return self.menubar.format_menu  #format menu
        if menu=='view':
            return self.menubar.view_menu  #view menu
        if menu=='toolbars':
            return self.menubar.toolbars_menu  #toolbar submenu
        if menu=='tools':
            return self.menubar.tool_menu  #tool menu
        if menu=='help':
            return self.menubar.help_menu  #help menu

    #---Event handlers----------------------------------------------------------
    def OnClose(self,event):
        """Editor frame close event handler"""
        res = self.notebook.CloseAll()
        if res is True:
            self.Hide()

    def OnFileHistory(self, event):
        filenum = event.GetId() - wx.ID_FILE1
        path = self.filehistory.GetHistoryFile(filenum)
        self.notebook.OpenFile(path)

#-------------------------------------------------------------------------------
class EditorTools(wx.ToolBar):
    def __init__(self,parent):

        wx.ToolBar.__init__(self,parent,-1,wx.DefaultPosition, 
                    wx.DefaultSize,wx.TB_FLAT | wx.TB_NODIVIDER)
        self.parent = parent

        #set the icon size
        size=wx.Size(22, 22)
        self.SetToolBitmapSize(size)

        #load some icons
        new_bmp     = common22.document_new.GetBitmap()
        open_bmp    = common22.document_open.GetBitmap()
        save_bmp    = common22.document_save.GetBitmap()

        cut_bmp     = common22.edit_cut.GetBitmap()
        copy_bmp    = common22.edit_copy.GetBitmap()
        paste_bmp   = common22.edit_paste.GetBitmap()

        undo_bmp    = common22.edit_undo.GetBitmap()
        redo_bmp    = common22.edit_redo.GetBitmap()
        search_bmp  = common22.edit_find.GetBitmap()

        untab_bmp   = editor_icons.format_indent_less.GetBitmap()
        tab_bmp     = editor_icons.format_indent_more.GetBitmap()
        com_bmp     = editor_icons.edit_comment.GetBitmap()
        uncom_bmp   = editor_icons.edit_uncomment.GetBitmap()
        sep_bmp     = editor_icons.add_seperator.GetBitmap()
        
        #new
        self.AddLabelTool( wx.ID_NEW, "New", new_bmp,shortHelp='Create a new file',
                        longHelp='Create a new file to edit')
        self.Bind(wx.EVT_TOOL, self.OnNew, id=wx.ID_NEW)

        #open
        self.AddLabelTool(wx.ID_OPEN, "Open", open_bmp,shortHelp='Open a file',
                        longHelp='Open a file to edit')
        self.Bind(wx.EVT_TOOL, self.OnOpen, id=wx.ID_OPEN)
        
        #save
        self.AddLabelTool(wx.ID_SAVE, "Save", save_bmp,shortHelp='Save file',
                        longHelp='Save file to disk')
        self.Bind(wx.EVT_TOOL, self.OnSave, id=wx.ID_SAVE)
        self.AddSeparator()
        
        #cut
        self.AddLabelTool(wx.ID_CUT, "Cut", cut_bmp,shortHelp='Cut',
                        longHelp='Cut selection to clipboard')
        self.Bind(wx.EVT_TOOL, self.OnCut, id=wx.ID_CUT)

        #copy
        self.AddLabelTool(wx.ID_COPY, "Copy", copy_bmp,shortHelp='Copy',
                        longHelp='Copy selection to clipboard')
        self.Bind(wx.EVT_TOOL, self.OnCopy, id=wx.ID_COPY)

        #paste
        self.AddLabelTool(wx.ID_PASTE, "Paste", paste_bmp,shortHelp='Paste',
                        longHelp='Paste selection from clipboard')
        self.Bind(wx.EVT_TOOL, self.OnPaste, id=wx.ID_PASTE)
        self.AddSeparator()

        #undo
        self.AddLabelTool(wx.ID_UNDO, "Undo", undo_bmp,shortHelp='Undo changes',
                        longHelp='Undo changes')
        self.Bind(wx.EVT_TOOL, self.OnUndo, id=wx.ID_UNDO)
        
        #redo
        self.AddLabelTool(wx.ID_REDO, "Redo", redo_bmp,shortHelp='Redo changes',
                        longHelp='Redo changes')
        self.Bind(wx.EVT_TOOL, self.OnRedo, id=wx.ID_REDO)
        
        #search and replace
        self.AddLabelTool(wx.ID_FIND, "Find and replace", search_bmp,
                    shortHelp='Find and replace',longHelp='Find and replace')
        self.Bind(wx.EVT_TOOL, self.OnFind, id=wx.ID_FIND)
        
        self.Realize()
    #---event handlers----------------------------------------------------------
    def OnNew(self,event):
        """New event handler"""
        self.parent.notebook.New()

    def OnOpen(self,event):
        """Open event handler"""
        self.parent.notebook.Open()

    def OnSave(self,event):
        """Save event handler"""
        self.parent.notebook.Save()

    def OnCut(self,event):
        """Cut event handler"""
        self.parent.notebook.Cut()

    def OnCopy(self,event):
        """Copy event handler"""
        self.parent.notebook.Copy()

    def OnPaste(self,event):
        """Paste event handler"""
        self.parent.notebook.Paste()

    def OnUndo(self,event):
        """Undo event handler"""
        self.parent.notebook.Undo()

    def OnRedo(self,event):
        """Redo event handler"""
        self.parent.notebook.Redo()

    def OnFind(self,event):
        """Opens the find/replace pane event handler"""
        if self.parent.search.IsShown():
            self.parent.search.Hide()
        else:
            self.parent.search.Show()
        self.parent.auimgr.Update()

#-------------------------------------------------------------------------------
class FormatTools(wx.ToolBar):
    def __init__(self,parent):
        wx.ToolBar.__init__(self,parent,-1,wx.DefaultPosition, 
                    wx.DefaultSize,wx.TB_FLAT | wx.TB_NODIVIDER)
        self.parent = parent

        #set the icon size
        size=wx.Size(22, 22)
        self.SetToolBitmapSize(size)

        #load some icons
        untab_bmp   = editor_icons.format_indent_less.GetBitmap()
        tab_bmp     = editor_icons.format_indent_more.GetBitmap()
        com_bmp     = editor_icons.edit_comment.GetBitmap()
        uncom_bmp   = editor_icons.edit_uncomment.GetBitmap()
        sep_bmp     = editor_icons.add_seperator.GetBitmap()

        #unindent
        self.AddLabelTool(wx.ID_UNINDENT, "Unident", untab_bmp,
                    shortHelp='Unindent selection',longHelp='Unindent selection')
        self.Bind(wx.EVT_TOOL, self.OnUndent, id=wx.ID_UNINDENT)
        
        #indent
        self.AddLabelTool(wx.ID_INDENT, "Indent", tab_bmp,
                        shortHelp='Indent selection',longHelp='Indent selection')
        self.Bind(wx.EVT_TOOL, self.OnIndent, id=wx.ID_INDENT)
        
        #comment
        id = wx.NewId()
        self.AddLabelTool(id, "Comment", com_bmp,shortHelp='Comment selection',
                        longHelp='Comment selection')
        self.Bind(wx.EVT_TOOL, self.OnComment, id=id)
        
        #uncomment
        id = wx.NewId()
        self.AddLabelTool(id, "Uncomment", uncom_bmp,
                shortHelp='Uncomment selection',longHelp='Uncomment selection')
        self.Bind(wx.EVT_TOOL, self.OnUnComment, id=id)
                
        #insert seperator
        id = wx.NewId()
        self.AddLabelTool(id, "Insert seperator", sep_bmp,
                        shortHelp='Insert seperator',longHelp='Insert seperator')
        self.Bind(wx.EVT_TOOL, self.OnInsertSeperator, id=id)
        
        self.Realize()

    #---event handlers----------------------------------------------------------
    def OnUndent(self,event):
        """Undent event handler"""
        self.parent.notebook.Undent()

    def OnIndent(self,event):
        """Indent event handler"""
        self.parent.notebook.Indent()

    def OnComment(self,event):
        """Comment event handler"""
        self.parent.notebook.Comment()

    def OnUnComment(self,event):
        """Uncomment event handler"""
        self.parent.notebook.UnComment()

    def OnInsertSeperator(self,event):
        """Insert seperator event handler"""
        self.parent.notebook.InsertSeperator()
 
#-------------------------------------------------------------------------------
class EditorMenu(wx.MenuBar):
    def __init__(self,parent):
        wx.MenuBar.__init__(self)

        self.parent = parent

        self.file_menu       = wx.Menu()     #file menu
        self.edit_menu       = wx.Menu()     #edit menu
        self.format_menu     = wx.Menu()     #format menu
        self.view_menu       = wx.Menu()     #view menu
        self.toolbars_menu   = wx.Menu()     #toolbar submenu
        self.tool_menu       = wx.Menu()     #tool menu
        self.help_menu       = wx.Menu()     #help menu
        
        ##add the menus to the menu bar
        self.Append(self.file_menu, "File")
        self.Append(self.edit_menu, "Edit")
        self.Append(self.format_menu, "Format")
        self.Append(self.view_menu, "View")
        self.Append(self.tool_menu, "Tools")
        self.Append(self.help_menu, "Help")

        ##file menu
        self.file_menu.Append(wx.ID_NEW, 'New', 'Create a new file') 
        self.file_menu.Append(wx.ID_OPEN, 'Open', 'Open a file') 
        self.file_menu.Append(wx.ID_SAVE, 'Save', 'Save the curent file') 
        self.file_menu.Append(wx.ID_SAVEAS, 'Save As', 'Save the current file with a differnt name') 
        self.file_menu.AppendSeparator()
        self.recent_menu = wx.Menu()
        self.file_menu.AppendMenu(wx.ID_ANY, "Recent Files", self.recent_menu)
        self.file_menu.AppendSeparator()
        self.file_menu.Append(wx.ID_CLOSE, "Close Editor",'Closes the editor window')
        self.file_menu.Append(wx.ID_EXIT, 'Exit','Exit PTK')

        #event bindings
        self.parent.Bind(wx.EVT_MENU, self.OnNew, id=wx.ID_NEW)
        self.parent.Bind(wx.EVT_MENU, self.OnOpen, id=wx.ID_OPEN) 
        self.parent.Bind(wx.EVT_MENU, self.OnSave, id=wx.ID_SAVE)
        self.parent.Bind(wx.EVT_MENU, self.OnSaveAs, id=wx.ID_SAVEAS) 
        self.parent.Bind(wx.EVT_MENU, self.OnMenuClose, id=wx.ID_CLOSE)
        self.parent.Bind(wx.EVT_MENU, self.OnExit, id=wx.ID_EXIT)
    
        ##edit menu
        self.edit_menu.Append(wx.ID_CUT, 'Cut', 'Cut selection') 
        self.edit_menu.Append(wx.ID_COPY, 'Copy', 'Copy selection') 
        self.edit_menu.Append(wx.ID_PASTE, 'Paste', 'Paste from clipboard') 
        self.edit_menu.AppendSeparator()
        self.edit_menu.Append(wx.ID_UNDO, 'Undo', 'Undo past actions') 
        self.edit_menu.Append(wx.ID_REDO, 'Redo', 'Redo undone actions') 
        self.edit_menu.AppendSeparator()
        self.edit_menu.Append(wx.ID_REPLACE, 'Find and replace', 'Open the find pane')
        #event bindings
        self.parent.Bind(wx.EVT_MENU, self.OnCut, id=wx.ID_CUT)
        self.parent.Bind(wx.EVT_MENU, self.OnCopy, id=wx.ID_COPY) 
        self.parent.Bind(wx.EVT_MENU, self.OnPaste, id=wx.ID_PASTE)
        self.parent.Bind(wx.EVT_MENU, self.OnUndo, id=wx.ID_UNDO)
        self.parent.Bind(wx.EVT_MENU, self.OnRedo, id=wx.ID_REDO)  
        self.parent.Bind(wx.EVT_MENU, self.OnFind, id=wx.ID_REPLACE) 

        ##format menu
        indentid=wx.NewId()
        self.format_menu.Append(indentid, 'Indent', 'Indent selection')
        undentid=wx.NewId()
        self.format_menu.Append(undentid, 'Undent', 'Undent selection') 
        comid=wx.NewId()
        self.format_menu.Append(comid, 'Comment', 'Comment selection') 
        uncomid=wx.NewId()
        self.format_menu.Append(uncomid, 'Uncomment', 'Uncomment selection') 
        self.format_menu.AppendSeparator()
        sepid=wx.NewId()
        self.format_menu.Append(sepid, 'Insert seperator', 'Insert a seperator comment')
        #event bindings
        self.parent.Bind(wx.EVT_MENU, self.OnIndent, id=indentid) 
        self.parent.Bind(wx.EVT_MENU, self.OnUndent, id=undentid) 
        self.parent.Bind(wx.EVT_MENU, self.OnComment, id=comid) 
        self.parent.Bind(wx.EVT_MENU, self.OnUnComment, id=uncomid) 
        self.parent.Bind(wx.EVT_MENU, self.OnInsertSeperator, id=sepid) 
        
        ##view menu
        self.view_menu.AppendSubMenu(self.toolbars_menu , 'Toolbars...','Show Toolbars')
        self.view_menu.AppendSeparator()
        #layouts added in auimix

        ##tools menu
        #Run code
        runid=wx.NewId()
        self.tool_menu.Append(runid, 'Run code', 
                            'Runs the selected code in the command console') 
        #Run as seperate app code
        extrunid=wx.NewId()
        self.tool_menu.Append(extrunid, 'Run code externally', 
                                'Runs the selected code as a new process') 
        #event bindings
        self.parent.Bind(wx.EVT_MENU, self.OnRun, id=runid) 
        self.parent.Bind(wx.EVT_MENU, self.OnExtRun, id=extrunid) 

        ##help menu
        self.help_menu.Append(wx.ID_HELP, 'Help', 'Open the python documentation...') 
        tipid=wx.NewId()
        self.help_menu.Append(tipid, 'Show tips', 'Show tips') 
        self.help_menu.Append(wx.ID_ABOUT, 'About...', 'About this program...') 
        #bindings
        self.parent.Bind(wx.EVT_MENU, self.OnHelp, id=wx.ID_HELP)
        self.parent.Bind(wx.EVT_MENU, self.OnTip, id=tipid)
        self.parent.Bind(wx.EVT_MENU, self.OnAbout, id=wx.ID_ABOUT)

    #---Event handlers----------------------------------------------------------
    #file menu
    def OnNew(self,event):
        """New file event handler"""
        self.parent.notebook.New()

    def OnOpen(self,event):
        """Open file event handler"""
        self.parent.notebook.Open()

    def OnSave(self,event):
        """Save file event handler"""
        self.parent.notebook.Save()

    def OnSaveAs(self,event):
        """Save as event handler"""
        self.parent.notebook.SaveAs()

    def OnMenuClose(self,event):
        """Menu close event handler"""
        self.parent.Close()

    def OnExit(self, event):
        """Called when exit menu item selected"""
        app = wx.GetApp()
        app.Exit()

    #edit menu
    def OnCut(self,event):
        """Edit.Cut event handler"""
        self.parent.notebook.Cut()

    def OnCopy(self,event):
        """Edit.Copy event handler"""
        self.parent.notebook.Copy()

    def OnPaste(self,event):
        """Edit.Paste event handler"""
        self.parent.notebook.Paste()

    def OnUndo(self,event):
        """Edit.Undo event handler"""
        self.parent.notebook.Undo()

    def OnRedo(self,event):
        """Edit.Redo event handler"""
        self.parent.notebook.Redo()

    def OnIndent(self,event):
        """Edit.Indent event handler"""
        self.parent.notebook.Indent()

    def OnUndent(self,event):
        """Edit.Undent event handler"""
        self.parent.notebook.Undent()

    def OnComment(self,event): 
        """Edit.Comment event handler"""
        self.parent.notebook.Comment()

    def OnUnComment(self,event):
        """Edit.UnComment event handler"""
        self.parent.notebook.UnComment()

    def OnInsertSeperator(self,event):
        """Edit.Insert Seperator event handler"""
        self.parent.notebook.InsertSeperator()
        
    def OnFind(self,event):
        """Opens the find/replace pane event handler"""
        self.parent.search.Show()
        self.parent.auimgr.Update()

    #view menu
    #done in auimixin

    #tools menu
    def OnRun(self,event):
        """Run event handler"""
        self.parent.notebook.Run()
    
    def OnExtRun(self,event):
        """Ext Run event handler"""
        self.parent.notebook.ExtRun()

    #help menu
    def OnHelp(self,event):
        """Open the help browser"""
        open_help()  

    def OnTip(self,event):
        """Open the help tips"""
        app=wx.GetApp()
        app.ShowTips(override=True)

    def OnAbout(self,event):
        """Opens the about box"""
        PTKInfoDialog()
