"""
This contains the editor notebook control. This handles and displays the open
files in a tabbed notebook view.

EditorNotebook - the editor notebook class
"""
from subprocess import Popen, PIPE  #until Launcher/ExtEngine/wxProcess implemented
import os                           #for filepath functions

import wx                           #for gui elements
import wx.aui                       #fancy notebook control
import wx.stc

from ptk_lib.core_tools.fileio import FileDrop,DoFileDialog

from editor_page import EditorPage

#---Editor notebook-------------------------------------------------------------
class EditorNotebook(wx.aui.AuiNotebook):   
    def __init__(self,parent):
        style = wx.aui.AUI_NB_DEFAULT_STYLE |wx.aui.AUI_NB_WINDOWLIST_BUTTON
        wx.aui.AuiNotebook.__init__(self,parent,-1,style=style)

        #create a droptarget
        self.dt = FileDrop(self.OpenFile)
        self.SetDropTarget(self.dt)  

        #dictionary of open files, {page:file path}
        self._openfiles = {}

        #internal to clear the paused markers from the correct filename
        # TODO: is there a better way - clear all pages?        
        self._lastpause = None

        #event bindings
        self.Bind(wx.aui.EVT_AUINOTEBOOK_PAGE_CLOSE, self.OnClosePage)

    #---Interface methods------------------------------------------------------- 
    # These are really just wrapper methods for the page methods
    #---------------------------------------------------------------------------
    def New(self):
        """Create a new empty page in the notebook"""
        page = EditorPage(self,-1)
        page.SetSavePoint() #set unchanged state
        #add to the notebook
        num = self.GetPageCount()
        name = 'unnamed ' + str(num+1)
        #add to the openfiles dictionary
        self._openfiles[page]=None
        #add to the notebook
        self.AddPage(page,name)
        self.SetSelection(num) #set new page active
        #get any page events to handle here
        self.Bind(wx.stc.EVT_STC_SAVEPOINTLEFT, self.OnPageSavePoint, page)
        self.Bind(wx.stc.EVT_STC_SAVEPOINTREACHED, self.OnPageSavePoint, page)

    def Open(self):
        """ method for menu/toolbar open dialog"""
        #Create the file open dialog.
        paths,index = DoFileDialog(None, wildcard = "Python source (*.py,*.pyw)|*.py;*.pyw|All files (*,*.*)|*.*;*")
        #Next open each file specified, load the data and create the pages
        for path in paths:
            self.OpenFile(path)
        #ensure the parent frame is shown and raised to the top
        self.Parent.Show()
        self.Parent.Raise()

    def OpenFile(self,path):
        """ Opens a file sepcified by the path """
        path = os.path.abspath(path)
        path = os.path.normcase(path)
        #check file exists
        if os.path.exists(path) is False:
            return

        #check file path against open tabs
        if path in self._openfiles.values():
            #get the page
            page = self.GetPageFromPath(path)
            #show the already open file
            n = self.GetPageIndex(page)
            self.SetSelection(n)
            return

        #not already open so open in a new page and add to the notebook
        page = EditorPage(self,-1)
        page.LoadFile(path)
        name=os.path.basename(path)
        num = self.GetPageCount()
        self.AddPage(page,name)
        self._openfiles[page]=path
        self.SetSelection(num) #set new page active

        #get any page events to handle here
        self.Bind(wx.stc.EVT_STC_SAVEPOINTLEFT, self.OnPageSavePoint, page)
        self.Bind(wx.stc.EVT_STC_SAVEPOINTREACHED, self.OnPageSavePoint, page)

        #add to parents file history    
        self.Parent.filehistory.AddFileToHistory(path)

    def Save(self,num=None):
        """Save the current page, or the page number if given """
        if num is None:
            num  = self.GetSelection()  
            if num==-1: #no page
                return
        page = self.GetPage(num)
        #check if previously saved
        abspath = self._openfiles[page]
        if abspath is None:
            #not saved before, do a saveas
            self.SaveAs(num)
        else:
            page.SaveFile(abspath)
        #check tabname
        self.CheckTabName(num)

    def SaveAs(self,num=None):
        """
        Save the current page with a new filename, or the page number ig given
        """
        if num is None:
            num  = self.GetSelection()
            if num==-1: #no page
                return
        page = self.GetPage(num)

        #save as dialog
        dlg = wx.FileDialog(
                self, message="Save as:",
                defaultDir=os.getcwd(),
                defaultFile="",
                wildcard = "Python source (*.py)|*.py|Python source no command shell (*.pyw)|*.pyw|All files (*.*)|*.*",
                style= wx.SAVE | wx.OVERWRITE_PROMPT
                )
        #only save if not cancelled.
        if dlg.ShowModal() != wx.ID_OK:
            return

        # get the enetered file path
        path = dlg.GetPath()
        #check extension specified use the dialog filters
        root,ext= os.path.splitext(path)
        if ext=='':
            type=dlg.GetFilterIndex()
            if type==0:
                ext = '.py'
            elif type==1:
                ext = '.pyw'
            else:
                ext = ''
        path = root+ext
        #save the file
        page.SaveFile(path)
        #update the openfile dictionary
        path = os.path.abspath(path)
        path = os.path.normcase(path)
        self._openfiles[page]=path
        #check tabname
        self.CheckTabName(num)
        #add to parents file history    
        self.Parent.filehistory.AddFileToHistory(path)
    
    def Cut(self):
        """Cut from the current page"""
        num  = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        page.Cut()

    def Copy(self):
        """Copy from the current page"""
        num  = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        page.Copy()

    def Paste(self):
        """Paste from the current page"""
        num  = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        page.Paste()

    def Undo(self):
        """Undo in the current page"""
        num  = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        page.Undo()

    def Redo(self):
        """Redo in the current page"""
        num  = self.GetSelection()
        if num==-1: #no page
               return
        page = self.GetPage(num)
        page.Redo()

    def Indent(self):
        """Indent in the current page"""
        num = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        page.Indent()

    def Undent(self):
        """Undent in the current page"""
        num = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        page.Undent()

    def Comment(self):
        """Comment in the current page"""
        num = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        page.Comment()

    def UnComment(self):
        """UnComment in the current page"""
        num = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        page.UnComment()

    def InsertSeperator(self):
        """Insert a comment seperator"""
        num = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        page.InsertSeperator()

    def Find(self,stext,back=False,flags=0):
        """
        Find the string in the current page.
        Returns true if found, false otherwise
        """
        num = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        if back is True:
            found = page.FindPrevious(stext,flags)
        else:
            found = page.FindNext(stext,flags)
        return found
    
    def Replace(self,stext,rtext,back=False,flags=0):
        """
        Replace the selection (or find the next occurance if none selected)
        Returns true if found/replaced
        """
        num = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        #see if there is a selection
        sel = page.GetSelectedText()
        #stext found previously replace it
        if sel==stext:
            page.ReplaceSelection(rtext)
            #find the next one
            found = self.Find(stext,back,flags)
        else:
            #not found yet do find next
            found = self.Find(stext,back,flags)

    def ReplaceAll(self,stext,rtext,back=False,flags=0):
        """
        Replace all in the direction specified
        """
        num = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        #start with current selection
        sel = page.GetSelectedText()
        if sel==stext:
            page.ReplaceSelection(rtext)
            #find the next one
            found = self.Find(stext,back,flags)
        else:
            #not found yet do find next
            found = self.Find(stext,back,flags)

        #keep going until no more found
        while found is True:
            #see if there is a selection
            sel = page.GetSelectedText()
            if sel==stext:
                page.ReplaceSelection(rtext)
            #find the next one
            found = self.Find(stext,back,flags)
        return found

    def Run(self):
        """Run the selected code in the console"""
        num = self.GetSelection()
        if num==-1: #no page
            return
        page = self.GetPage(num)
        page.Run()

    def ExtRun(self):
        """Run the select code with file launcher"""
        num = self.GetSelection()
        if num==-1: #no page
            return
        #run as external process
        page = self.GetPage(num)
        abspath = self._openfiles[page]
        if abspath is None:
            #not saved yet
            return 
        cmd='python'
        print '\n Running :',abspath
        p1 = Popen([cmd,abspath], stdout=PIPE,stderr=PIPE)
        p1.wait()
        ret = p1.communicate()
        print '\n Finished'
        print 'stdout:  ',ret[0]
        print 'stderr:  ',ret[1]
    
    def CloseAll(self):
        """Attempts to close all open pages - asks to save if necessary"""
        #loop over open tabs
        numpages = self.GetPageCount()
        for n in range(0,numpages):
            self.SetSelection(0)
            res = self.CheckPageClose(0)
            if res is False:
                return False
            page = self.GetPage(0)
            res = self.RemovePage(0)
            self._openfiles.pop(page)#remove the page from dictionary
            page.Destroy()
        return True
    
    #---events------------------------------------------------------------------
    def OnPageSavePoint(self, event):
        #page has left/reached a save point mark 'dirty'/clean
        num = self.GetSelection()
        if num==-1:
            return
        self.CheckTabName(num)

    def OnClosePage(self,event):
        """This method is called when a tab is closed"""
        num  = self.GetSelection()  
        page = self.GetPage(num)
        result = self.CheckPageClose(num)
        if result is False:
            event.Veto()
        else:
            #remove the page from dictionary
            self._openfiles.pop(page)
            event.Skip()


    #---Checking methods--------------------------------------------------------
    def CheckClose(self):
        """check if open tabs are saved - return true if ok to close"""
        #loop over open tabs
        numpages = self.GetPageCount()
        for page in range(0,numpages):
            ans  = self.CheckPageClose(page)
            #return false if a page doesn't want to close
            if ans is False:
                return False
        return True

    def CheckPageClose(self,num):
        """Check if it is ok to close a page"""
        page = self.GetPage(num)
        if page.GetModify() is True:
            #ask if save/discard or cancel close
            abspath = self._openfiles[page]
            if abspath is None:
                name = self.GetPageText(num)
                if name[-1]=='*':
                    name=name[0:-2]
                msg = "Save file? "+name
            else:
                name=os.path.basename(abspath)
                msg = "Save changes to: "+name+" ?"

            dlg = wx.MessageDialog(self, msg, "Editor",
                    wx.YES_NO | wx.NO_DEFAULT | wx.CANCEL | wx.ICON_QUESTION)
            result=dlg.ShowModal()
            dlg.Destroy()

            #actions depending upon answer
            if result==wx.ID_YES: #if yes try to save
                self.Save(num)
                return True
            elif result==wx.ID_CANCEL: #if cancel veto the close
                return False
            elif result==wx.ID_NO:
                return True
        return True

    def CheckTabName(self,num):
        """Updates the tabname"""
        page = self.GetPage(num)
        abspath = self._openfiles[page]
        #not saved at all case
        if abspath is None:
           name = self.GetPageText(num)
           if name[-1]=='*':
                name=name[0:-2]
        #saved before case
        else:
            name=os.path.basename(abspath)
        #check modification state
        if page.GetModify()==True:
            name = name+' *'
        self.SetPageText(num,name)

    def GetPageFromPath(self,path):
        """Get the editor page that contains the path, None if not opened"""
        path = os.path.abspath(path)
        path = os.path.normcase(path)
        if path not in self._openfiles.values():
            return None
        for i in self._openfiles.items():
            if i[1]==path:
                return i[0]
        return None

    def _UpdateBreakpoints(self):
        """
        Update the breakpoints symbols display in editor pages. 
        For example if the current engine has switched  or the debugger has been toggled on/off
        (the debugger could be disabled)
        """
        num = self.GetPageCount()
        for n in range(0,num):
            page = self.GetPage(n)
            page.UpdateBreakpointMarkers()

    def _UpdatePauseMarkers(self):
        """
        Update any paused indicator symbols in the editor pages.
        For example if the current engine has switched, paused or resumed.
        """
        engmgr = wx.GetApp().toolmgr.get_tool('EngineManager')
        eng = engmgr.get_current_engine()
        if eng is None:
           if self._lastpause is not None:
                self._lastpause.SetPausedMarkers(None)
                self._lastpause = None
           return

        #not paused
        if eng.debugger.filename is None:
            if self._lastpause is not None:
                self._lastpause.SetPausedMarkers(None)
                self._lastpause = None
            return
        filename = os.path.abspath(eng.debugger.filename)
        filename = os.path.normcase(filename)

        page =   self.GetPageFromPath( filename )     
        if page is not None:
            self._lastpause = page
            page.SetPausedMarkers(eng.debugger.lineno)

