"""
Editor tool.

A simple python code editor with just enough features.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('Editor')

#---Imports---------------------------------------------------------------------
import wx
import wx.aui as aui

from ptk_lib.message_bus.mb_node import MBNode
from ptk_lib.tool_manager import Tool
from ptk_lib.engine import messages

from ptk_lib.core_tools.fileio import Importer, DoFileDialog
from ptk_lib.core_tools.engine_manager import engman_messages

from editor_frame import EditorFrame
import editor_messages
import editor_icons

#---the tools class-------------------------------------------------------------
class Editor(Tool):
    name = 'Editor'
    descrip = 'Core tool providing a simple python code editor with just enough features'
    version = '11.01'  
    author = 'T.Charrett'
    requires = ['TaskIcon','Console','FileIO']           
    core = True            
    icon = editor_icons.editor32

    def __init__(self):
        Tool.__init__(self)
        log.info('Initialising tool')

        #create the editor frame
        self.frame =  EditorFrame(self)

        #Add taskbar menu item
        taskicon = self.toolmgr.get_tool('TaskIcon')
        bmp = editor_icons.editor16.GetBitmap()
        taskicon.add_menu_item(wx.NewId(), 'Open the Editor',
                   'Open the Editor window', self.on_show, bmp)

        #Add menu item to console window
        console = self.toolmgr.get_tool('Console')
        bmp = editor_icons.editor16.GetBitmap()
        console.add_menu_item('tools', wx.NewId(), 'Editor',
                    'Open the Editor window', self.on_show, bmp)

        #Register message handlers
        self.msg_node = MBNode(self.msg_bus, 'Editor')
        self.msg_node.set_handler(editor_messages.NEW,   self.msg_new)  #open the a new file
        self.msg_node.set_handler(editor_messages.OPEN,  self.msg_open) #open the file(s)
        self.msg_node.set_handler(editor_messages.SHOW,  self.msg_show) #show the editor
        self.msg_node.set_handler(editor_messages.HIDE,  self.msg_hide) #hide the editor

        #subscribe to application messages
        self.msg_node.subscribe('App.Init',self.msg_app_init)
        self.msg_node.subscribe('App.ExitCheck',self.msg_app_exitcheck)
        self.msg_node.subscribe('App.Exit',self.msg_app_exit)

        #Add the edit command when a new engine starts.
        self.msg_node.subscribe(engman_messages.ENGINE_STARTED,self.msg_engine_start)

        #on debug toggled messages - update the editor breakpoint markers
        self.msg_node.subscribe(messages.ENGINE_DEBUG_TOGGLED,self.msg_debug_toggled)

        #on debug paused/resumed update paused markers 
        self.msg_node.subscribe(messages.ENGINE_DEBUG_PAUSED, self.msg_debug_paused)
        self.msg_node.subscribe(messages.ENGINE_DEBUG_RESUMED,self.msg_debug_paused)

        #on engine switched messages update breakpoint markers and paused markers
        self.msg_node.subscribe(engman_messages.ENGINE_SWITCHED,self.msg_engine_switched)

        #register file importer
        fileio = self.toolmgr.get_tool('FileIO')
        self.importer = EditorImporter(self.frame)
        fileio.register_importer( self.importer )

        log.info('Done Initialising tool')

    #---Interfaces--------------------------------------------------------------
    def show_console(self):
        """
        Show the Console frame
        """
        self.frame.Show()
        self.frame.Raise()

    def hide_console(self):
        """
        Hide the console frame
        """
        self.frame.Hide()

    def add_menu_item(self, menu, id, string, helpstring, evthndlr, bitmap=None, pos=None):
        """
        Add a menu item to the menu.
            menu - string description of menu ('Tools', 'File'. 'Edit' etc)
            string - string to display
            helpstring - help string for item
            evthndlr - wx event handler callable
            bitmap  - bitmap to use / None for no bitmap
            pos - Position in menu, inserted here if specified.
        """
        menu = self.frame.GetMenu(menu)
        item = wx.MenuItem(menu, id, string,helpstring)
        self.frame.Bind(wx.EVT_MENU, evthndlr, id=id)
        if bitmap is not None:
            item.SetBitmap(bitmap)
        if pos is None:
            pos = menu.GetMenuItemCount()
        menu.InsertItem(pos,item)

    #---Message handlers-------------------------------------------------------
    def msg_show(self,msg):
        """
        Message handler for Editor.Show and wx.Event from menu
        """
        self.frame.Show()
        self.frame.Raise()

    def msg_hide(self,msg):
        """
        Message handler for Editor.Hide
        """
        self.frame.Hide()

    def msg_new(self,msg):
        """
        Message handler for Editor.New
        """
        self.frame.notebook.New()
        self.frame.Show()
        self.frame.Raise()

    def msg_open(self,msg):
        """
        Message handler for Editor.Open
        data=list of files to open in editor
        """
        filepaths = msg.get_data()
        if filepaths is ():
            #Create the file open dialog.
            filepaths,index = DoFileDialog(None,wildcard = "Python source (*.py,*.pyw)|*.py;*.pyw|All files (*,*.*)|*.*;*")
            if filepaths==[]:
                return

        if filepaths is not None or filepaths!=[]:
            #open the file requested
            for path in filepaths:
                self.frame.notebook.OpenFile(path)
        self.frame.Show()
        self.frame.Raise()

    def msg_app_init(self,msg):
        """
        Listener for App.Init message
        Sent when application starts
        """
        #load the main window layouts (in aui mixin class)
        self.frame.LoadLayouts()

    def msg_app_exitcheck(self,msg):
        """
        Check its ok for the application to exit
        """
        #check for unsaved files
        res= self.frame.notebook.CheckClose()
        if res is False:
            self.app.VetoExit()

    def msg_app_exit(self,msg):
        """
        Listener for App.Exit message
        Save settings on application exit
        """
        #save the main window layouts (in aui mixin class)
        self.frame.SaveLayouts()
        #save the recent files list
        cfg = self.app.GetConfig()
        cfg.SetPath("Editor//")
        self.frame.filehistory.Save(cfg)
        cfg.Flush()

    def msg_engine_start(self,msg):
        """
        Engine manager - Engine.Started message handler
        """
        #When an engine is started add the edit() command
        log.debug('Adding edit() command to new engine')
        engname,engtype = msg.get_data()
        app = wx.GetApp()
        engmgr = app.toolmgr.get_tool('EngineManager')
        eng = engmgr.get_engine(engname)
        eng.add_builtin(edit, 'edit')

    def msg_debug_toggled(self, msg):
        """
        Debugger enabled/disabled message
        """
        self.frame.notebook._UpdateBreakpoints()

    def msg_debug_paused(self, msg):
        """
        Debugger paused/resumed message handler
        """
        self.frame.notebook._UpdatePauseMarkers()

    def msg_engine_switched(self, msg):
        """
        Engine switched message handler 
        """
        self.frame.notebook._UpdatePauseMarkers()
        self.frame.notebook._UpdateBreakpoints()

    #---other-------------------------------------------------------------------
    def on_show(self, event):
        """wx event handler for the taskbar and console menu items"""
        self.frame.Show()
        self.frame.Raise()

#-------------------------------------------------------------------------------
# Engine magic command - this is added to engines __builtin__ module.
#-------------------------------------------------------------------------------
def edit(filename):
    """
    Edit a file in the ptk editor
    """
    import os
    import __main__

    #check file exists
    filepath = os.path.abspath(filename)
    cwd = os.getcwd()
    #a full path given
    if os.path.exists(filepath) is False:
        raise Exception('File does not exist: '+filename)
    #send the editor message
    __main__.engine.send_msg('Editor','Open',(filepath,))

#-------------------------------------------------------------------------------
#FileIO importer
#-------------------------------------------------------------------------------
class EditorImporter(Importer):
    def __init__(self,frame):
        Importer.__init__( self,'Open in the Editor', ['','.py','.pyw','.txt'],
                            data=False,
                            wildcards=["Python source (*.py,*.pyw)|*.py;*.pyw;",
                                        "Text file (*.txt)|*.txt"],
                            descrip = 'Open the file in the Editor' )
        self.frame = frame

    def __call__(self,filename):
         self.frame.OpenFile(filename)


