"""
A general purpose control for object addresses for use with tools such as 
Namespace browser and Inspector.

Features autocompletion of addresses fetched from the active engine.
Use the EVT_ENGINE_ADDRESS to get address changes
Automatically sends an EVT_ENGINE_ADDRESS when the engine state changes
to allow windows to refresh.

The engine name and address can be found from the event as event.engname and 
event.address. 
The address can be:
    ''   - top level namespace
or
    object name
or
    None if no engine is set as current
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('AddressCtrl')


import wx
from wx.lib.newevent import NewEvent
EvtEngineAddress , EVT_ENGINE_ADDRESS = NewEvent()

class AddressCtrl(wx.TextCtrl):
    def __init__(self,parent, id, msg_node, size=(200,-1)):
        wx.TextCtrl.__init__(self,parent,id, '',
                                size=size,style=wx.TE_PROCESS_ENTER)

        self.Bind( wx.EVT_KILL_FOCUS, self.OnKillFocus, self )
        self.Bind( wx.EVT_TEXT , self.OnText, self )
        self.Bind( wx.EVT_TEXT_ENTER , self.OnTextEnter, self )
        self.Bind( wx.EVT_KEY_DOWN , self.OnKeyDown, self )
        self.Bind(wx.EVT_LISTBOX_DCLICK, self.OnListBoxDClick)

        #messagebus node to use
        self.msg_node = msg_node

        #get a reference to the engine manager tool
        self.engmgr = wx.GetApp().toolmgr.get_tool('EngineManager')

        #create popup window...
        self.dropdown =  wx.PopupWindow(self)
        self.dropdown.SetSize( (size[0],120))

        #...containing a listbox
        self.listbox = wx.ListBox(self.dropdown,-1,size=(size[0],120),
                choices=[])

        #attributes to keep track of address
        self._autoupdate = False #automatically update when engine state changes.
        self.objname = None
        self.partname = None
        self.cur_add = ''
        self.cur_eng = None

        #enginename: address dictionary for engine swicthing memory
        self.memory = {}
        self.msg_node.subscribe('Engine.Switched', self.msg_eng_switched)
        self.msg_node.subscribe('Engine.Closed', self.msg_eng_closed)
        self.msg_node.subscribe('Engine.StateChange', self.msg_eng_change)
        self.msg_node.subscribe('Engine.Done', self.msg_eng_change)

    def __del__(self):
        self.msg_node.unsubscribe('Engine.Switched', self.msg_eng_switched)
        self.msg_node.unsubscribe('Engine.Closed', self.msg_eng_closed)
        self.msg_node.unsubscribe('Engine.StateChanged', self.msg_eng_change)
        self.msg_node.unsubscribe('Engine.Done', self.msg_eng_change)

    #---Interfaces--------------------------------------------------------------
    def SetEngine(self,engname):

        #store current address to memory
        self.memory[self.cur_eng] = self.cur_add
        if engname is not None:
            #check engine name
            if engname not in self.engmgr.get_engine_names():
                    raise NameError('Engine does not exist!')
        #set engine and address
        self.cur_eng = engname
        if self.cur_eng is None:
            self.Disable()
        else:
            self.Enable()
        address= self.memory.get(engname,'')
        self.SetAddress(address)

    def GetEngine(self):
        """Get the current engine name this address control is using"""
        return self.cur_eng
        
    def SetAddress(self,address,engname=None):
        """
        Set the address. If engname is None this will set the address for the 
        current engine, otherwise it will set the address in the engine memory.
        """
        #get the current engine
        if engname is None:
            engname = self.cur_eng

        #check the address
        if engname is not None:
            exists = self._CheckAddress(engname, address)
            #if the address is invalid set the control back to the current
            if exists is False:
                #check cur address is ok
                if self._CheckAddress(engname, self.cur_add):
                    address = self.cur_add
                else:
                    #use top level.
                    address = ''

        #not the current engine store address if it exists and return.
        if engname!=self.cur_eng:
            self.memory[engname] = address
            return

        #set address in text control, store the new address as current
        self.SetValue(address)
        self.cur_add = address
        self._ShowDropDown(False)   #hide autocomplete list

        #raise address event
        evt = EvtEngineAddress(engname=self.cur_eng, address=address)
        wx.PostEvent(self, evt)

    def GetAddress(self, engname=None):
        """
        Get the address. If engname is None this is for the current engine
        otherwise the address from the engine memory is returned.
        """
        if engname is None:
            engname = self.cur_eng

        if engname!=self.cur_eng:
            return self.memory.get(engname, '')

        return self.GetValue()
    
    def MoveUpLevel(self):
        """Change the address up one level"""
        address = self.GetValue()
        new = address.rpartition('.')[0]
        self.SetAddress(new)

    def RefreshAddress(self):
        """Refresh the current address"""
        self.SetAddress(self.cur_add)

    def SetAutoUpdate(self, flag=False):
        """
        Set whether an address event should be sent when the engine state 
        changes.
        """
        self._autoupdate = flag

    #---internal methods--------------------------------------------------------
    def _CheckAddress(self, engname, address):
        """
        Check an address. Returns True if ok, False if the address is invalid or
        the engine does not exist.
        """
        try:
            eng = self.engmgr.get_engine(engname)
        except:
            return False

        #check address
        if address!='':
            #not top level (top level is ok. but doesn't exist)
            return eng.run_task('object_exists',(address,))
        else:
            return True

    def _ShowDropDown(self, show=True):
        """
        Either display the drop down list (show = True) or hide it (show = False).
        """
        if show is False:
            self.dropdown.Show(False)
            return

        res = self._RefreshDropDown()
        if res is False:
            #could not get autocomps - no engine?
            return 

        #Finally show the list
        #get width of dropdown and text control and position
        dw,dh  = self.dropdown.GetSizeTuple()
        tw, th = self.GetSizeTuple()
        x, y = self.ClientToScreenXY(0,th)
        #check width of dropdown
        if dw!=tw:
            self.dropdown.SetSize( (tw,dh))
            self.listbox.SetSize( (tw,dh) )
        self.dropdown.SetPosition( (x, y) )
        self.dropdown.Show(show)

    def _RefreshDropDown(self):
        """
        Refresh the drop down list
        """
        #get engine
        if self.cur_eng is None:
           return False

        eng = self.engmgr.get_engine(self.cur_eng)
        if eng is None:
            return False

        #update contents of list
        value = self.GetValue()

        #split the partial address into objname and part
        if value.count('.')==0:
            self.objname = ''
            self.partname = value
        else:
            self.objname,self.partname = value.rsplit('.',1)

        #use  get_auto_comps task to get possibles addresses and number of chars 
        #to replace
        comps,nchars = eng.run_task('get_auto_comps',(value,))
        self.listbox.SetItems(comps)

        #scroll to first match
        self._ScrollToFirstMatch()
        return True

    def _ScrollToFirstMatch(self):
        """Internal method to get autocomps and display the list"""
        value = self.GetValue()

        #find best current match
        if value.endswith('.'):
            #show all options after a dot
            found=True
            n=0
        else:
            comps = self.listbox.GetItems()
            found = False
            for n in range(0,len(comps)):
                if comps[n].startswith(self.partname):
                    found = True
                    break

        #set listbox selection and show if a match found
        if found is True:
            self.listbox.SetSelection(n)            

    #---event handlers----------------------------------------------------------
    def OnText(self,event):
        part=self.GetValue()
        if part!='':
            #find matches and show list box
            self._ShowDropDown(True)
        else:
            #nothing entered hide listbox
            self._ShowDropDown(False)
        event.Skip()

    def OnKillFocus(self, event):
        self._ShowDropDown( False )
        event.Skip()

    def OnKeyDown(self, event):
        visible = self.dropdown.IsShown()
        sel = self.listbox.GetSelection()

        KC = event.GetKeyCode()
        if KC == wx.WXK_DOWN and visible:
            if (sel < (self.listbox.GetCount() - 1)) and sel!=-1:
                self.listbox.SetSelection(sel+1)
            skip = False

        elif KC == wx.WXK_UP and visible:
            if (sel > 0):
                self.listbox.SetSelection(sel-1)
            skip = False

        elif KC in [wx.WXK_TAB, wx.WXK_RETURN] and visible:
            #get value
            name = self.listbox.GetString(sel)
            if len(self.objname)==0:
                new = name
            else:
                new = self.objname+'.'+name
            self.SetAddress(new)
            self._ShowDropDown(False)
            self.SetInsertionPoint(len(new))
            skip = False

        elif KC == wx.WXK_ESCAPE and visible:
            self._ShowDropDown(False)
            skip = False
        else:
            skip = True

        if skip :
            event.Skip()

    def OnListBoxDClick(self,event):
        sel = self.listbox.GetSelection()
        #get value
        name = self.listbox.GetString(sel)
        if len(self.objname)==0:
            new = name
        else:
            new = self.objname+'.'+name
        self.SetValue(new)
        self._ShowDropDown(False)
        self.SetInsertionPoint(len(new))

    def OnTextEnter(self,event):
        #raise address event
        address = self.GetValue()
        self.SetAddress(address)

    #---message handlers--------------------------------------------------------
    def msg_eng_switched(self,msg):
        """
        Message handler for Engine.Switched.
        Load the previous address for the new engine (if it exists).
        """
        neweng,oldeng = msg.get_data()
        self.SetEngine(neweng)

    def msg_eng_closed(self,msg):
        """
        Message handler for Engine.Closed.
        Remove engine's address history.
        """
        engname = msg.get_data()
        self.memory.pop(engname,None)

    def msg_eng_change(self,msg):
        """
        Message handler for Engine.Done / Engine.StateChanged.
        Refresh the dropdown list
        """
        if self._autoupdate is False:
            return
        engname  = msg.data[0]
        self.RefreshAddress()
        self._RefreshDropDown()

#-------------------------------------------------------------------------------
def test():
    f= wx.Frame(None,-1,'Test')
    s = wx.BoxSizer(wx.VERTICAL)
    a = AddressCtrl(f,-1)
    s.Add(a, 0, wx.ALL,5)
    f.SetSizer(s)
    f.Show()
    return f
