"""

Interfaces
----------

EngineInterface:
A conveinence interface to send engine control and interogation mesages.
Created automatically by the EngineManager
The engine state is also kept upto date by the EngineManager
All the actions can also be done by sending the approiate message via the messagebus
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('Engine.Interface')
log.setLevel(logging.DEBUG)

#---imports----------------------------------------------------------------------
import marshal

from ptk_lib.message_bus.mb_node import MBNode
from ptk_lib.message_bus import mb_protocol
from ptk_lib.engine import messages

#-------------------------------------------------------------------------------
class EngineInterface(object):
    def __init__(self, msg_node, engname, engtype):
        #internal attributes
        self.engname = engname
        self.engtype = engtype
        self.engnode = 'Engine.'+engname
        self.msg_node = msg_node          #message bus node used to send/recieve

        #send message directly via the message bus
        self.send_msg = msg_node.send_msg
        self.publish_msg = msg_node.publish_msg

        #engine state flags
        self.busy = False
        self.debug = False
        self.profile = False

        #debugger interface
        self.debugger = DebuggerInterface(self)

        #profiler interface
        self.profiler = ProfilerInterface(self)
        
    #---interface methods-------------------------------------------------------
    def register_task( self, task):
        """
        Register a complex task with the engine.
        """
        s = marshal.dumps(task.func_code)
        res = self.send_msg(self.engnode, messages.ENG_REGISTER_TASK, (s,), get_result=True)
        return res

    def run_task(self, taskname, args=(),kwargs={}, scope=None):
        """
        Run a complex task in the engine (must have been previously registered 
        using register_task).
        If the debugger is active and an integer scope is given the code will
        be exectuted in the scope at that level (scope=0 is the users namespace 
        dictionary), if scope is None the code will be executed in the user 
        namespace.
        """
        res = self.send_msg(self.engnode, messages.ENG_RUN_TASK, (taskname,args,kwargs,scope), get_result=True)
        return res

    def execute(self, source, scope=None):
        """
        Execute source in the engine. 
        If the debugger is active and an integer scope is given the code will
        be exectuted in the scope at that level (scope=0 is the users namespace 
        dictionary), if scope is None the code will be executed in the user 
        namespace.
        """
        res = self.send_msg(self.engnode, messages.ENG_EXEC_COMMAND, (source,scope), get_result=True)
        return res

    def evaluate(self, source, scope=None):
        """
        Evaluate source in the engine.
        If the debugger is active and an integer scope is given the code will
        be exectuted in the scope at that level (scope=0 is the users namespace 
        dictionary), if scope is None the code will be executed in the user 
        namespace.
        """
        res = self.send_msg(self.engnode, messages.ENG_EVAL_COMMAND, (source,scope), get_result=True)
        return res

    def add_builtin(self, func, name):
        """
        Add a function to the engines builtin module.
        """
        s = marshal.dumps(func.func_code)
        res = self.send_msg( self.engnode, messages.ENG_ADD_BUILTIN, (name,s), get_result=True )
        return res

    def get_registered_tasks(self):
        """
        Returns a list of the registered engine task names.
        """
        res = self.send_msg( self.engnode, messages.ENG_GET_TASKS, (), get_result=True)
        return res

    def notify_change(self):
        """
        Publish an engine state change message to notify that the engine state 
        may have changed.
        """
        busy, debug, profile = self.get_state()
        data = (self.engname, busy, debug, profile)
        self.publish_msg(messages.ENGINE_STATE_CHANGE, data)

    def set_compiler_flag(self, flag, setto=True):
        """
        Set or unset a __future__ compiler flag.
        """
        res = self.send_msg( self.engnode, messages.ENG_FUTURE_FLAG, (flag,setto), get_result=True)
        return res

    def stop(self):
        """
        Stop a running command.
        """
        success = self.send_msg(self.engnode, messages.ENG_STOP, ())

    def get_state(self):
        """
        Returns a tuple of engine state flags=(busy, debug, profile)
        """
        return (self.busy, self.debug, self.profile)

    def enable_debug(self, flag=True):
        """
        Enable the debugger.
        """
        res = self.send_msg(self.engnode, messages.ENG_DEBUG_TOGGLE,
                            (flag,),True)
        return res

    def enable_profile(self, flag=True):
        """
        Enable the profiler.
        """
        res = self.send_msg(self.engnode, messages.ENG_PROFILE_TOGGLE,
                            (flag,), True)
        return res

#-------------------------------------------------------------------------------
class DebuggerInterface(object):
    def __init__(self, eng):
        """
        Interface to the engine debugger 
        - the state is updated by the Enginemanager
        """
        self.eng = eng              #ref to the engine interface
        self.engnode = eng.engnode  #engines node name

        #debugger state flags
        self.paused = False
        self.can_stepin = False
        self.can_stepout = False
        #scopes
        self.scopes = []
        self.active_scope = None
        #paused at
        self.scope_name = None
        self.filename = None
        self.lineno   = None
        
    #---debugger interfaces-----------------------------------------------------
    def pause(self, flag=True):
        """
        Pause a running command.
        """
        if flag is False:
            return self.resume()
        res = self.eng.send_msg(self.engnode, messages.ENG_DEBUG_PAUSE, (),True)
        self.paused = res
        return res

    def end(self):
        """
        End debugging and finish running command.
        """
        res = self.eng.send_msg(self.engnode, messages.ENG_DEBUG_END, (),True)
        return res

    def resume(self):
        """
        Resume a running command.
        """
        res=self.eng.send_msg(self.engnode, messages.ENG_DEBUG_RESUME, (),True)
        self.paused = not res

    def step(self):
        """
        Step to the next line.
        """
        self.eng.send_msg(self.engnode, messages.ENG_DEBUG_STEP, ())

    def step_in(self):
        """
        Step into a new scope.
        """
        self.eng.send_msg(self.engnode, messages.ENG_DEBUG_STEP_INTO, ())

    def step_out(self):
        """
        Step out of a scope.
        """
        self.eng.send_msg(self.engnode, messages.ENG_DEBUG_STEP_OUT, ())

    def set_breakpoint(self, bpdata):
        """
        Set a break point
        id     =  a unique identifier for this breakpoint
        bpdata =  (filename, lineno, condition, ignore_count, trigger_count)
        """
        res = self.eng.send_msg(    self.engnode, 
                                    messages.ENG_DEBUG_SET_BREAKPOINT,
                                    (bpdata,), True )
        return res

    def modify_breakpoint(self, id, **kwargs):
        """
        Modify an existing breakpoint
    
        Only modify the keywords given (from filename, lineno, condition,
        trigger_count and ignore_count).

        e.g. modify_breakpoint( id=1, filename='test.py', lineno=23) will modify
        the breakpoint filename and lineno.
        """
        res = self.eng.send_msg(    self.engnode, 
                                    messages.ENG_DEBUG_MODIFY_BREAKPOINT,
                                    (id, kwargs), True )
        return res

    def clear_breakpoint(self, id):
        """
        Clear an existing breakpoint.

        id = Unique breakpoint identifier.
        If None all breakpoints will be cleared.
        """
        res = self.eng.send_msg(    self.engnode, 
                                    messages.ENG_DEBUG_CLEAR_BREAKPOINT,
                                    (id,), True )
        return res

    def set_active_scope(self, level):
        """
        Set the currently active scope for interogation
        """
        self.eng.send_msg(  self.engnode, 
                            messages.ENG_DEBUG_SET_SCOPE, (level,))

    def get_state(self):
        """
        Returns information on the debugger state.
        flags = (paused, can_stepin, can_stepout)
        """
        return (self.paused, self.can_stepin, self.can_stepout)

    def get_scopelist(self):
        """
        Returns the engine available scopes - only updated when paused!
        """
        return self.scopes

    def get_active_scope(self):
        """
        Returns the currently active scope being interogated - only valid when paused!
        """
        return self.active_scope

    def get_paused_at(self):
        """
        Returns the (scopename, filename, lineno) that the debugger is paused at.
        If not paused returns (None,None)
        """
        return self.scope_name, self.filename, self.lineno

#-------------------------------------------------------------------------------
class ProfilerInterface(object):
    def __init__(self, eng):
        """
        Interface to the engine profiler
        - the state is updated by the Enginemanager
        """
        self.eng = eng              #ref to the engine interface
        self.engnode = eng.engnode  #engines node name

    #---profiler interfaces-----------------------------------------------------

    
