"""
EngineManager tools

Tool to handle engine starting/closing/switching and debugger breakpoints.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('EngineManager')

#---imports---------------------------------------------------------------------

import imp                      #for checking if gui libaries are installed
import os                       #for launching external engines
import subprocess               #for launching external engines
import sys                      #for launching external engines
import pickle
import __future__

import wx
import wx.aui as aui

from ptk_lib.tool_manager import Tool
from ptk_lib.message_bus.mb_node import MBNode
from ptk_lib.message_bus import mb_protocol
from ptk_lib.engine import messages, eng_misc

import engman_messages
import engman_icons
from eng_interface import EngineInterface
from engman_settings import AutostartPanel, EnvironmentPanel
from engman_toolbars import EngineTools, DebugTools


#---the tools class-------------------------------------------------------------
class EngineManager(Tool):
    name = 'EngineManager'
    descrip = 'Core tool providing engine switching and management'
    version = '11.01' 
    author = 'T.Charrett'
    requires = []      
    core = True
    icon = engman_icons.engine32

    def __init__(self):
        log.info('Initialising tool')
        Tool.__init__(self)

        self.msg_node = MBNode(self.msg_bus, 'EngineManager')
        self.internal_yield = wx.YieldIfNeeded    #internal engine yield function

        #Dictionaries of active engines
        self.engines = {}   #{engname: EngineInterface object}
        self.processes = {} #(engname: subprocess process object}

        #break points (set in all engines)
        self.bpoints = eng_misc.DictList()
        self.bp_counter = 0

        #current engine name
        self.current_engine = None

        #find which engines are available
        self.engtypes = []       #list of engtypes strings
        self.engdescrip = {}     #{engtype string, description string
        self.engtypes, self.engdescrip = eng_misc.get_available_engines()

        #subscribe to MessageBus messages to make sure we know if an 
        #engine disconnects
        self.msg_node.subscribe(mb_protocol.SYS_NODE_CONNECT,self.msg_sys_connect)
        self.msg_node.subscribe(mb_protocol.SYS_NODE_DISCONNECT ,self.msg_sys_disconnect)     
        self.msg_node.subscribe(mb_protocol.SYS_NODE_ERR_DISCONNECT,self.msg_sys_disconnect)  

        #register for engine message to keep internal state of the engine interfaces
        self.msg_node.subscribe(messages.ENGINE_BUSY, self.msg_eng_busy)
        self.msg_node.subscribe(messages.ENGINE_DONE, self.msg_eng_done)

        self.msg_node.subscribe(messages.ENGINE_DEBUG_TOGGLED, self.msg_debug_toggled)
        self.msg_node.subscribe(messages.ENGINE_DEBUG_PAUSED, self.msg_debug_paused)
        self.msg_node.subscribe(messages.ENGINE_DEBUG_RESUMED, self.msg_debug_resumed)
        self.msg_node.subscribe(messages.ENGINE_DEBUG_SCOPECHANGED, self.msg_debug_scope)

        self.msg_node.subscribe(messages.ENGINE_PROFILE_TOGGLED, self.msg_profile_toggled)

        #subscribe to application messages
        self.msg_node.subscribe('App.Init',self.msg_app_init)

        #add enginemanager settings panels
        taskicon = self.toolmgr.get_tool('TaskIcon')
        taskicon.add_settings_item('Engine Manager', None, engman_icons.engine16.GetBitmap())
        taskicon.add_settings_item('Engine Manager\\Startup engines', AutostartPanel, None)
        taskicon.add_settings_item('Engine Manager\\Python environment', EnvironmentPanel, None)

        #add Engine toolbar to console frame
        console = self.toolmgr.get_tool('Console')
        self.engtools = EngineTools(console.frame, self)
        pane = (aui.AuiPaneInfo().Name('Engine toolbar')
                    .Caption('Engine toolbar').ToolbarPane().CloseButton(True)
                    .CaptionVisible(False)
                    .DestroyOnClose(False).Top().Row(1).Position(0)
                    .LeftDockable(False).RightDockable(False))
        console.frame.AddToolbar(self.engtools,'Engine toolbar',pane,
                        helpstring = 'Show/hide the Engine toolbar')   

        ##add Debuger toolbar to console frame
        self.debugtools = DebugTools(console.frame, self)
        pane = (aui.AuiPaneInfo().Name('Debuger toolbar')
                    .Caption('Debuger toolbar').ToolbarPane().CloseButton(True)
                    .CaptionVisible(False)
                    .DestroyOnClose(False).Top().Row(1).Position(1)
                    .LeftDockable(False).RightDockable(False))
        console.frame.AddToolbar(self.debugtools,'Debuger toolbar',pane,
                        helpstring = 'Show/hide the Debugger toolbar') 

        log.info('Done Initialising tool')

    #---general interfaces------------------------------------------------------
    #engine starting/closing
    def start_engine(self, engname, engtype):
        """
        Start a new engine. 
        This creates the Engine message bus node (either by creating the 
        InternalEngine object or launching an external process which connects to
        the message bus). The engine interfaces are not created 
        until the EngineNode has actually connected (see msg_sys_connect).
        """
        log.info('Starting engine: '+engtype+' name: '+engname)
        #check engine name is ok
        if self.engines.has_key(engname):
            log.exception('Engine with name already exists: '+str(engname))
            raise Exception('Engine with name already exists: '+str(engname))
        
        if engtype not in self.engtypes:
            raise Exception('Unknown/Unavailable engine type: '+str(engtype))
        
        if engtype=='Internal':
            if engname!='Internal':
                raise Exception('Can only call the internal engine: Internal')
            self.internal = eng_misc.start_internal_engine(self.msg_bus, 
                                    dict={}, internal_yield=self.internal_yield)

        else:
            #start the engine process
            process = eng_misc.start_external_engine(engname,engtype, self.msg_bus)
            #store process in dictionary to clear up when the engine ends
            self.processes[engname] = process
        log.debug('Engine launched '+engname)

    def close_engine(self, engname):
        """
        Close the engine called engname.
        """
        if self.engines.has_key(engname) is False:
            raise Exception('No engine with name: '+str(engname))
        eng_misc.close_engine(engname, self.msg_bus)
        #This requests the engine to close, the eng interface is removed when
        #the engine disconnects from the message bus

    #multiple engine management
    def set_current_engname(self, engname):
        """
        Set the current active engine
        """
        if self.engines.has_key(engname) is False:
            raise Exception('No engine with name: '+str(engname))
        oldname = self.current_engine
        self.current_engine = engname
        log.debug('setting current engine: '+engname)
        #publish Engine Switched message
        self.msg_node.publish_msg(engman_messages.ENGINE_SWITCHED,(engname, oldname))

    def get_current_engname(self):
        """
        Get the current active engine name
        """
        return self.current_engine
    
    def get_engine_names(self):
        """
        Get a list of engine names
        """
        return self.engines.keys()

    def get_engine_types(self):
        """
        Get a dict of available engine types.
        """
        #return a copy of the types list.
        return list(self.engtypes)

    def get_engine_descriptions(self):
        """
        Get a dict of available engine types : engine description
        """
        #return a copy of the global.
        return dict(self.engdescrip)

    def get_suggested_name(self, engtype, engname=None):
        """
        Get a suggest name for a new engine.
        """
        if engtype=='Internal':
            return 'Internal'
        if (engname in self.engines.keys()) or (engname is None):
            engname = 'Engine-' + str(len(self.engines)+1)
        return engname

    def get_engine(self, engname):
        """
        Get the EngineInterface object for interrogating the engine given by 
        engname.
        """
        eng = self.engines.get(engname, None)
        if eng is None:
            raise Exception('No engine with name: '+str(engname))
        return eng

    def get_current_engine(self):
        """
        Get the EngineInterface object for interogating the current engine.
        Returns None if no current engine.
        """
        if self.current_engine is None:
            return None
        eng = self.engines.get(self.current_engine, None)
        return eng

    def get_all_engines(self):
        """
        Returns the started engines dictionary {name:engine instance}
        """
        return self.engines

    #debugger interfaces
    def set_breakpoint(self, filename,lineno,
                        condition=None,ignore_count=None,trigger_count=None):
        """
        Set a breakpoint for all engines.
        Returns the new breakpoint id (used to edit/clear breakpoints).
        """
        log.debug('In set breakpoint')
        #check if there is already a breakpoint at this location, if there is
        #modify it, if not create a new breakpoint.
        bps = self.bpoints.filter( ('filename','lineno'),(filename,lineno) )
        if len(bps)!=0:
            #existing breakpoint at that location return None.
            raise Exception('Breakpoint already set at filename,lineno - use edit_breakpoint')
        
        #create new id.
        id = self.bp_counter
        self.bp_counter+=1 
        #store in mkeydict.
        bpdata = {  'id':id,'filename':filename, 'lineno':lineno,
                    'condition':condition, 'ignore_count':ignore_count,
                    'trigger_count':trigger_count }
        self.bpoints.append(bpdata)
        
        #set the breakpoint in each engine.
        for engname in self.engines:
            eng = self.engines[engname]
            res = self.msg_node.send_msg( eng.engnode, 
                    messages.ENG_DEBUG_SET_BREAKPOINT,(bpdata,), True )

        #publish a breakpoint set message
        self.msg_node.publish_msg(  engman_messages.ENGMAN_BREAKPOINT_SET,
                                    (bpdata,)  )        
        return id

    def clear_breakpoint(self, id):
        """
        Clear the debugger breakpoint with the id given for all engines.
        If id is None all breakpoints will be cleared.
        """
        bps = self.bpoints.filter( ('id',),(id,) )
        if len(bps)==0:
            raise Exception('No breakpoint with id '+str(id))
        bpdict = bps[0]

        #clear the breakpoint in each engine
        for engname in self.engines:
            eng = self.engines[engname]
            res = self.msg_node.send_msg( eng.engnode, 
                                    messages.ENG_DEBUG_CLEAR_BREAKPOINT,
                                    (id,), True )
        
        #remove from internal breakpoint list
        if id is not None:
            self.bpoints.remove(bpdict)
        else:
            self.bpoint.clear()

        #published a breakpoint set message
        self.msg_node.publish_msg(  engman_messages.ENGMAN_BREAKPOINT_CLEARED,
                                    (id,)  )        

    def modify_breakpoint(self, id, **kwargs):
        """
        Modify a breakpoint.

        Only modify the keywords given (from: filename, lineno, condition,
        trigger_count and ignore_count).

        e.g. modify_breakpoint( id=1, filename='test.py', lineno=23) will modify
        the breakpoint filename and lineno. To pass a dictionary of changes use 
        **dict.
        """
        bps = self.bpoints.filter( ('id',),(id,) )
        if len(bps)==0:
            raise Exception('No breakpoint with id '+str(id))
        bpdict = bps[0]

        #modify the breakpoint in each engine
        for engname in self.engines:
            eng = self.engines[engname]
            res = self.msg_node.send_msg( eng.engnode, 
                                    messages.ENG_DEBUG_MODIFY_BREAKPOINT,
                                    (id, kwargs), True )

        #modify the breakpoint data dictionary.
        bpdict.update( kwargs )

        #published a breakpoint set message
        self.msg_node.publish_msg(  engman_messages.ENGMAN_BREAKPOINT_CHANGED,
                                    (id,kwargs)  )     

    def get_breakpoint(self, id):
        """
        Get the breakpoint by id
        """
        res = self.bpoints.filter( keys=('id',), values=( id,) )
        if res==[]:
            return None
        return res[0]

    def get_breakpoints(self, filename=None):
        """
        Get the breakpoints currently set in the filename given if None return
        all breakpoints
        """
        if filename is None:
            return self.bpoints.items()
        return self.bpoints.filter( keys=('filename',), values=(filename,) )

    #---Message handlers--------------------------------------------------------
    def msg_app_init(self,msg):
        """
        Handler for App.Init message
        """
        #autostart engines and setup defualt environment
        cfg = self.app.GetConfig()
        cfg.SetPath("EngineManager//")
        s = cfg.Read("auto_start_engines","")  #a list of engine name, type tuples to autostart
        try:
            engs = pickle.loads(str(s))
        except:
            engs=[("Engine-1","wxEngine")]

        #start each engine
        for engname,engtype in engs:
            log.info('Auto starting engine '+str(engname)) 
            eng = self.start_engine(engname,engtype)

        #console tool create console on Engine.Started message

    def msg_sys_connect(self,msg):
        """
        When a node connects to the message bus check if it is an engine.
        If it is register it and publish EngineStarted message.
        """
        node_name, = msg.get_data()
        if node_name.startswith('Engine.') is False:
            return

        engname = node_name.partition('.')[2]
        engtype = self.msg_node.send_msg(node_name, messages.ENG_GET_TYPE, get_result=True)
        log.info('New engine connected: '+str(engname) + ' type: '+str(engtype)) 

        #create a new EngineInterface object 
        eng = EngineInterface(self.msg_node, engname, engtype)
        self.engines[engname] = eng

        #set any breakpoints
        for bpdata in self.bpoints:
           res = self.msg_node.send_msg( eng.engnode, 
                    messages.ENG_DEBUG_SET_BREAKPOINT,(bpdata,), True )    

        #do other engine start tasks
        self.on_engine_start(eng)         

        #publish an engine started message
        self.msg_node.publish_msg(engman_messages.ENGINE_STARTED, (engname,engtype))

        #set as current engine
        self.set_current_engname(engname)

    def msg_sys_disconnect(self, msg):
        """
        Message handler for MessageBus system message indicating if an engine
        node disconnected.
        """
        node_name = msg.get_data()[0]
        #not an engine
        if node_name.startswith('Engine.') is False:
            return

        engname = node_name.partition('.')[2]
        log.info('Engine disconnected, engname: '+engname)

        #find the engine interface and remove it
        eng = self.engines.pop(engname,None)

        #remove the process from the internal dicionary and communicate with it 
        #to prevent orphaned processes.
        process = self.processes.pop(engname,None)
        if process is not None:
            #todo: start process with a pipe then read stdout/stderr and print this to console here.
            process.poll() #communicate(None)

        #if this is was the current engine set it to None
        if self.current_engine == engname:
            #publish Engine Switched message
            self.current_engine = None
            self.msg_node.publish_msg(engman_messages.ENGINE_SWITCHED,(None, engname))

        #publish an 'ENGINE_CLOSED' message to other tools.
        self.msg_node.publish_msg(engman_messages.ENGINE_CLOSED, (engname,))

    #messages to keep engine interfaces upto date.
    def msg_eng_busy(self,msg):
        engname, debug, profile = msg.data
        eng = self.engines.get(engname, None)
        if eng is None:
            log.warning('Busy message from nonexistant engine! '+str(engname))
            return
        eng.busy = True
        eng.debug = debug
        eng.profile = profile

    def msg_eng_done(self,msg):
        engname, debug, profile = msg.data
        eng = self.engines.get(engname, None)
        if eng is None:
            log.warning('Done message from nonexistant engine! '+str(engname))
            return
        #reset engine state
        eng.busy = False
        eng.debug = debug
        eng.profile = profile
        #update the debugger interface state
        #flags
        eng.debugger.paused = False
        eng.debugger.can_stepin = False
        eng.debugger.can_stepout = False
        #paused at
        eng.debugger.scope_name = None
        eng.debugger.filename = None
        eng.debugger.lineno = None
        #scopes
        eng.debugger.scopes = []
        eng.debugger.active_scope = None

    def msg_debug_toggled(self,msg):
        engname, debug = msg.data
        eng = self.engines.get(engname, None)
        if eng is None:
            log.warning('Debug toggled message from nonexistant engine! '+str(engname))
            return
        eng.debug = debug

    def msg_debug_paused(self,msg):
        engname, paused_at, scope_list, active_scope, flags = msg.data
        eng = self.engines.get(engname, None)
        if eng is None:
            log.warning('Debug paused message from nonexistant engine! '+str(engname))
            return
        #update the debugger interface state
        #flags
        eng.debugger.paused = True
        can_stepin, can_stepout = flags
        eng.debugger.can_stepin = can_stepin
        eng.debugger.can_stepout = can_stepout
        #paused at
        scope_name,filename, lineno = paused_at
        eng.debugger.scope_name = scope_name
        eng.debugger.filename = filename
        eng.debugger.lineno = lineno
        #scopes
        eng.debugger.scopes = scope_list
        eng.debugger.active_scope = active_scope

    def msg_debug_resumed(self,msg):
        engname, = msg.data
        eng = self.engines.get(engname, None)
        if eng is None:
            log.warning('Debug resumed message from nonexistant engine! '+str(engname))
            return
        #update the debugger interface state
        #flags
        eng.debugger.paused = False
        eng.debugger.can_stepin = False
        eng.debugger.can_stepout = False
        #paused at
        eng.debugger.scope_name = None
        eng.debugger.filename = None
        eng.debugger.lineno = None
        #scopes
        eng.debugger.scopes = []
        eng.debugger.active_scope = None

    def msg_debug_scope(self, msg):
        engname, scopes, active_scope = msg.data
        eng = self.engines.get(engname, None)
        if eng is None:
            log.warning('Debug scope change message from nonexistant engine! '+str(engname))
            return
        #scopes
        eng.debugger.scopes = scopes
        eng.debugger.active_scope = active_scope

    def msg_profile_toggled(self,msg):
        engname, profile = msg.data
        eng = self.engines.get(engname, None)
        if eng is None:
            log.warning('Profile toggled message from nonexistant engine! '+str(engname))
            return
        eng.profile = profile

    #---internal methods-------------------------------------------------------
    def on_engine_start(self, eng):
        """
        Do the engine start up tasks
        """
        #set up engine environment
        #execute startup script if option selected
        cfg = self.app.GetConfig()
        cfg.SetPath("EngineManager//")
        flag = cfg.ReadBool("exec_startup",True)
        if flag is True:
            res = eng.run_task('execute_startup_script')
            log.debug('Started script executed - success ='+str(res))

        #add builtin commands
        eng.add_builtin(ptk_help, 'ptk_help')
        eng.add_builtin(clear, 'clear')

        #set compiler flags
        flag = cfg.ReadBool("future_div",False)
        eng.set_compiler_flag(__future__.CO_FUTURE_DIVISION,flag)
        flag = cfg.ReadBool("future_import",False)
        eng.set_compiler_flag(__future__.CO_FUTURE_ABSOLUTE_IMPORT,flag)
        flag = cfg.ReadBool("future_print",False)
        eng.set_compiler_flag(__future__.CO_FUTURE_PRINT_FUNCTION,flag)
        flag = cfg.ReadBool("future_unicode",False)
        eng.set_compiler_flag(__future__.CO_FUTURE_UNICODE_LITERALS,flag)


#-------------------------------------------------------------------------------
def ptk_help():
    """
    Display PTK specfic help
    """
    doc = (
    "Python Toolkit console keys:\n"+
    "   ctrl+up/down    -   cycle through command history\n"+
    "                       (searches for a partially typed command)\n"+
    "   escape          -   clear the current command\n"+
    "   ctrl+space      -   open the autocomplete list\n"+
    "\n"+
    "Engines:\n"+
    "The internal engine (python interpretor) runs in the same process as the\n"+
    "interface so if your program crashes or hangs so will PTK. Use an external\n"+
    "engine (a python interpretor in a separate process) to avoid blocking the \n"+
    "GUI when running long commands or when interactive use of a GUI mainloop \n"+
    "is required (wxPython can be used from the internal engine as well).\n"+
    "Multiple external engines can be used simutanously, but only a single \n"+
    "internal engine is allowed"
    )
    print doc

#-------------------------------------------------------------------------------
#Uses engine message CON_CLEAR
def clear():
    """
    Clear the PTK console
    """
    #send a engine interface message to clear the console
    import __main__
    __main__.engine.send_msg(__main__.engine.console, 'Con_Clear',(None,))


