import wx
import engman_icons
from engman_dialogs import EngineChoiceDialog

#---Engine toolbar--------------------------------------------------------------
class EngineTools(wx.ToolBar):
    def __init__(self,parent, tool):

        wx.ToolBar.__init__(self,parent,-1,wx.DefaultPosition, 
                    wx.DefaultSize,wx.TB_FLAT | wx.TB_NODIVIDER)
        self.parent = parent
        self.tool = tool

        #set the icon size
        size=wx.Size(16, 16)
        self.SetToolBitmapSize(size)

        #load some icons
        bmp_engnew   = engman_icons.engine_new.GetBitmap()
        bmp_stop     = engman_icons.engine_stop.GetBitmap() 
        bmp_debug    = engman_icons.engine_debug.GetBitmap()
        bmp_profile  = engman_icons.engine_profile.GetBitmap() 

        #engine list
        self.AddControl(wx.StaticText(self, wx.ID_ANY , 'Current engine:   ',size=(-1,16)))
        self.englist = wx.Choice(self, -1, size=(150, -1), choices=[])
        self.AddControl(self.englist)
        self.englist.SetMaxSize((-1,20))
        self.englist.SetSelection(0)
        self.Bind(wx.EVT_CHOICE, self.OnEngineChoice, self.englist)

        #new engine
        id = wx.NewId()
        self.AddLabelTool( id, "New engine", bmp_engnew,shortHelp='New engine',
                        longHelp='Create a new engine')
        self.Bind(wx.EVT_TOOL, self.OnNewEngine, id=id)
        self.AddSeparator()

        #Stop running code
        self.stopid = wx.NewId()
        self.AddLabelTool( self.stopid, "Stop", bmp_stop,
                            shortHelp='Stop running command',
                            longHelp='Stop running command')
        self.Bind(wx.EVT_TOOL, self.OnStop, id=self.stopid)

        #toggle debug
        self.debugid = wx.NewId()
        self.AddCheckLabelTool(self.debugid, "Enable debugger", bmp_debug,
                shortHelp="Enable engine debugger",
                longHelp="Enable the engine debugger mode")
        self.Bind(wx.EVT_TOOL, self.OnDebug, id=self.debugid)

        #toggle profile
        self.profileid = wx.NewId()
        self.AddCheckLabelTool(self.profileid, "Enable profiler", bmp_profile,
                shortHelp="Enable engine profiler",
                longHelp="Enable the engine profiler mode")
        self.Bind(wx.EVT_TOOL, self.OnProfile, id=self.profileid)

        self.Realize()

        #subscribe to engine messages
        self.tool.msg_node.subscribe('Engine.Started',  self.msg_eng_start_close)
        self.tool.msg_node.subscribe('Engine.Closed',   self.msg_eng_start_close)
        self.tool.msg_node.subscribe('Engine.Switched', self.msg_eng_switched)
        self.tool.msg_node.subscribe('Engine.Busy',    self.msg_eng_busy)
        self.tool.msg_node.subscribe('Engine.Done',    self.msg_eng_done)
        self.tool.msg_node.subscribe('Engine.Debug.Toggled', self.msg_eng_debug)
        self.tool.msg_node.subscribe('Engine.Profile.Toggled', self.msg_eng_profile)

    def _update_tools(self):
        """update tool button states"""
        eng = self.tool.get_current_engine()
        #disable tools if no engine.
        if eng is None:
            self.EnableTool(self.stopid  , False)
            self.EnableTool(self.debugid  , False)
            self.EnableTool(self.profileid  , False)
        else:
            #enable the tools
            busy, debug, profile = eng.get_state()

            self.EnableTool(self.debugid, True)
            self.EnableTool(self.profileid, True)
            self.EnableTool(self.stopid, busy)

            self.ToggleTool(self.debugid , debug)
            self.ToggleTool(self.profileid, profile)

        #update the engine list
        avail_engines = self.tool.get_engine_names()
        if eng is None:
            engname=None
        else:
            engname = eng.engname

        if avail_engines==[]:
            avail_engines= ['None']
            self.englist.SetItems(avail_engines)
            self.englist.SetSelection(0)
            self.englist.Disable()
        elif engname==None:
            self.englist.Enable()
            self.englist.SetItems(avail_engines)
        else:
            self.englist.Enable()
            self.englist.SetItems(avail_engines)
            self.englist.SetSelection(avail_engines.index(engname))

    #---events------------------------------------------------------------------
    def OnEngineChoice(self,event):
        """Current engine changed"""
        engname = self.englist.GetStringSelection()
        #switch to this engine
        self.tool.set_current_engname(engname)

    def OnNewEngine(self,event):
        """
        Create a new engine
        """
        #show a dialog here to create a new engine and console.
        engnames = self.tool.get_engine_names()
        d = EngineChoiceDialog(self.Parent, 'Start a new engine', engnames)
        res=d.ShowModal()
        if res==wx.ID_OK:
            engname,engtype = d.GetValue()
            self.tool.start_engine(engname, engtype)
        d.Destroy()

    def OnStop(self,event):
        eng = self.tool.get_current_engine()
        if eng is None:
            return
        eng.stop()

    def OnDebug(self,event):
        """Enable engine debugger"""
        flag = event.IsChecked()
        eng = self.tool.get_current_engine()
        if eng is None:
            return
        res = eng.enable_debug(flag)
        self.ToggleTool(self.debugid, res)

    def OnProfile(self,event):
        """Enable engine profiler"""
        flag = event.IsChecked()
        eng = self.tool.get_current_engine()
        if eng is None:
            return
        res = eng.enable_profile(flag)
        self.ToggleTool(self.profileid, res)

    #---messages----------------------------------------------------------------
    def msg_eng_switched(self,msg):
        """Message handler for EngineSwitched"""
        self._update_tools()
        
    def msg_eng_start_close(self,msg):
        """Message handler for engine start/close messages"""
        self._update_tools()

    def msg_eng_busy(self,msg):
        engname = msg.data[0]
        if engname !=self.tool.get_current_engname():
            return
        self.EnableTool(self.stopid, True)

    def msg_eng_done(self,msg):
        engname = msg.data[0]
        if engname !=self.tool.get_current_engname():
            return
        self.EnableTool(self.stopid, False)

    def msg_eng_debug(self, msg):
        engname,debug = msg.data
        if engname != self.tool.get_current_engname():
            return
        #toggle the debug button
        self.ToggleTool(self.debugid , debug)

    def msg_eng_profile(self, msg):
        engname,profile = msg.data
        if engname != self.tool.get_current_engname():
            return
        #toggle the profile button
        self.ToggleTool(self.profileid , profile)


#---console toolbar-------------------------------------------------------------
class DebugTools(wx.ToolBar):
    def __init__(self, parent, tool):

        wx.ToolBar.__init__(self,parent,-1,wx.DefaultPosition, 
                    wx.DefaultSize,wx.TB_FLAT | wx.TB_NODIVIDER)

        #store a reference to the console tool.
        self.tool = tool

        #set the icon size
        size=wx.Size(16, 16)
        self.SetToolBitmapSize(size)

        #load some icons
        bmp_pause    = engman_icons.engine_pause.GetBitmap() 
        bmp_resume   = engman_icons.engine_resume.GetBitmap()
        bmp_end      = engman_icons.engine_end_debug.GetBitmap()
        bmp_step     = engman_icons.engine_step.GetBitmap()
        bmp_stepin   = engman_icons.engine_stepin.GetBitmap()
        bmp_stepout  = engman_icons.engine_stepout.GetBitmap()

        #-----------------------------------------------------------------------
        #pause/resume
        self._paused = False
        self.pauseid = wx.NewId()
        self.AddCheckLabelTool( self.pauseid, "Pause", bitmap=bmp_pause,
                        shortHelp='Pause running command',
                        longHelp='Pause running command')
        self.Bind(wx.EVT_TOOL, self.OnPause, id=self.pauseid)

        #end debugging
        self.endid = wx.NewId()
        self.AddLabelTool( self.endid, "End", bitmap=bmp_end,
                        shortHelp='End debugging',
                        longHelp='End debugging and finish running command')
        self.Bind(wx.EVT_TOOL, self.OnEnd, id=self.endid)

        self.AddSeparator()

        #step
        self.stepid = wx.NewId()
        self.AddLabelTool( self.stepid, "Step", bitmap=bmp_step,
                        shortHelp='Step to next line',
                        longHelp='Step to next line of running command')
        self.Bind(wx.EVT_TOOL, self.OnStep, id=self.stepid)

        #step in
        self.stepinid = wx.NewId()
        self.AddLabelTool( self.stepinid, "Step In", bitmap=bmp_stepin,
                        shortHelp='Step into the new scope',
                        longHelp='Step into the new scope')
        self.Bind(wx.EVT_TOOL, self.OnStepIn, id=self.stepinid)

        #step out
        self.stepoutid = wx.NewId()
        self.AddLabelTool( self.stepoutid, "Step Out", bitmap=bmp_stepout,
                        shortHelp='Step out of the current scope',
                        longHelp='Step out of the current scope')
        self.Bind(wx.EVT_TOOL, self.OnStepOut, id=self.stepoutid)

        #scope list
        self.AddControl(wx.StaticText(self, wx.ID_ANY , 'Debugger scope:   ',size=(-1,16)))
        self.scopelist = wx.Choice(self, -1, size=(150, -1), choices=['Main'])
        self.AddControl(self.scopelist)
        self.scopelist.SetMaxSize((-1,20))
        self.Bind(wx.EVT_CHOICE, self.OnScopeChoice, self.scopelist)

        self.Realize()

        #subscribe to engine messages
        self.tool.msg_node.subscribe('Engine.Switched', self.msg_eng_switched)
        self.tool.msg_node.subscribe('Engine.Busy', self.msg_eng_busy)
        self.tool.msg_node.subscribe('Engine.Done', self.msg_eng_done)
        self.tool.msg_node.subscribe('Engine.Debug.Toggled', self.msg_eng_debug)
        self.tool.msg_node.subscribe('Engine.Debug.Paused', self.msg_debug_paused)
        self.tool.msg_node.subscribe('Engine.Debug.Resumed', self.msg_debug_resumed)
        self.tool.msg_node.subscribe('Engine.Debug.ScopeChanged', self.msg_debug_scope)

    #---messages----------------------------------------------------------------
    def msg_eng_busy(self, msg):
        engname, debug, profile = msg.data
        eng = self.tool.get_current_engine()
        if engname != eng.engname:
            return
        #enable the tools
        self._enable_tools(True,eng.debug,False,False,False)
    
    def msg_eng_done(self, msg):
        engname, debug, profile = msg.data
        eng = self.tool.get_current_engine()
        if engname != eng.engname:
            return

        #disable the tools 
        self._enable_tools(False,eng.debug,False,False,False)

    def msg_eng_debug(self, msg):
        engname, debug = msg.data
        eng = self.tool.get_current_engine()
        if engname != eng.engname:
            return
        #disable the tools 
        self._enable_tools(eng.busy,debug,eng.debugger.paused,False,False)

    def msg_debug_paused(self, msg):
        engname, paused_at, scope_list,active_scope, flags = msg.data
        eng = self.tool.get_current_engine()
        if engname != eng.engname:
            return
        can_stepin,can_stepout = flags
        
        #update scopelist
        self.scopelist.SetItems(scope_list)
        self.scopelist.SetSelection(active_scope)

        #enable/disable tools                
        self._enable_tools(eng.busy,eng.debug,True,can_stepin,can_stepout)

    def msg_debug_resumed(self, msg):
        engname, = msg.data
        eng = self.tool.get_current_engine()
        if engname != eng.engname:
            return

        #set tools
        self._enable_tools(eng.busy,eng.debug,False,False,False)

    def msg_debug_scope(self, msg):
        """Message handler for Engine Debugger scope change messages"""
        engname, scope_list, active_scope  = msg.data
        cureng = self.tool.get_current_engname()
        if engname != cureng:
            return
        self.scopelist.SetItems(scope_list)
        self.scopelist.SetSelection(active_scope)

    def msg_eng_switched(self,msg):
        """Message handler for EngineSwitched"""
        #check debugger state.
        eng = self.tool.get_current_engine()

        #no engines disable all tools
        if eng is None:
            self._enable_tools(False,False,False,False,False)
            return

        #check engine state and update buttons
        busy, debug, profile = eng.get_state()
        
        #not in debug mode - disable all tools
        if debug is False:
            self._enable_tools(False,False,False,False,False)
            return

        #check debugger state
        paused,can_stepin, can_stepout = eng.debugger.get_state()
        
        #update the tools, and the scopelist       
        if paused:
            scopes = eng.debugger.get_scopelist()
            index  = eng.debugger.get_active_scope()
            self.scopelist.SetItems(scopes)
            self.scopelist.SetSelection(index)
        self._enable_tools( busy, debug, paused, can_stepin, can_stepout)

    def _enable_tools(self, busy, debug, paused, can_stepin, can_stepout):
        #paused button
        self.EnableTool(self.pauseid, debug)
        self.ToggleTool(self.pauseid, paused)
        self.EnableTool(self.endid, (busy and debug))

        if paused is True:
            self.SetToolNormalBitmap(self.pauseid, engman_icons.engine_resume.GetBitmap())
            self.SetToolShortHelp(self.pauseid,'Resume running command')
            self.SetToolLongHelp(self.pauseid,'Resume running command')
        else:
            self.SetToolNormalBitmap(self.pauseid, engman_icons.engine_pause.GetBitmap())
            self.SetToolShortHelp(self.pauseid,'Pause running command')
            self.SetToolLongHelp(self.pauseid,'Pause running command')

        #steps
        self.EnableTool(self.stepid  , (paused and busy))            
        self.EnableTool(self.stepinid  , can_stepin)
        self.EnableTool(self.stepoutid  , can_stepout)
        #scopelist
        self.scopelist.Enable((paused and busy))
        if paused is False:
            self.scopelist.SetSelection(0)


    #---events------------------------------------------------------------------
    def OnPause(self,event):
        eng = self.tool.get_current_engine()
        if eng is None:
            return
        #check if we are pausing or resumeing
        if eng.debugger.paused :
            eng.debugger.resume()
        else:
            eng.debugger.pause()
        #update pause button
        paused,can_stepin, can_stepout = eng.debugger.get_state()
        self._enable_tools( eng.busy, eng.debug, paused, can_stepin, can_stepout)        

    def OnEnd(self,event):
        eng = self.tool.get_current_engine()
        if eng is None:
            return
        eng.debugger.end()

    def OnStep(self, event):
        eng = self.tool.get_current_engine()
        if eng is None:
            return
        eng.debugger.step()
    
    def OnStepIn(self, event):
        eng = self.tool.get_current_engine()
        if eng is None:
            return
        eng.debugger.step_in()

    def OnStepOut(self, event):
        eng = self.tool.get_current_engine()
        if eng is None:
            return
        eng.debugger.step_out()

    def OnScopeChoice(self, event):
        eng = self.tool.get_current_engine()
        if eng is None:
            return
        level = self.scopelist.GetSelection()
        eng.debugger.set_active_scope(level)

