"""
PathList control

A drop down list which handles path changes
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('PathManager')

#---Imports---------------------------------------------------------------------
import wx
import os

#---the pathlist control--------------------------------------------------------
class PathList(wx.ComboBox):
    """The current working directory selection toolbar item"""
    def __init__(self,parent, tool):
        """Create the pathlist control - provides a drop list of paths"""

        #get a reference to the EngineManager tool
        app = wx.GetApp()
        self.engmgr = app.toolmgr.get_tool('EngineManager')

        #pathmanager tool
        self.tool = tool

        #create combobox
        wx.ComboBox.__init__(self, parent, -1, "", choices=[],
                                size=(300,-1), style=wx.CB_DROPDOWN)
        
        #bind events
        self.Bind(wx.EVT_COMBOBOX, self.OnChange, self)
        self.Bind(wx.EVT_TEXT_ENTER, self.OnText, self)

        #current path
        self.cur = ''

        #subcribe to these subjects
        self.tool.msg_node.subscribe('Engine.Switched',self.msg_eng_switched)

    def LoadPaths(self):
        """Load previous paths from config"""
        cfg = wx.GetApp().GetConfig()

        #get the recent paths
        paths = [os.getcwd()]
        for n in range(0,10):
            if cfg.Exists("PathManager//recent"+str(n)) :
               path = cfg.Read("PathManager//recent"+str(n))
               if paths.count(path)==0 and os.path.isdir(path):
                    paths.append(path)
        self.SetItems(paths)
        self.cur = paths[0]
        self.SetSelection(0)

    def SavePaths(self):
        """Save the path list history to config"""
        cfg = wx.GetApp().GetConfig()
        #save the current working directory paths
        paths = self.GetItems()
        #cut down to 10 only
        paths = paths[:10]
        n=0
        for path in paths: 
            cfg.Write("PathManager//recent"+str(n), path)
            n=n+1
        cfg.Flush() 

    def OnChange(self,event):
        new = event.GetString()
        result = self.ChangeDir(new)

    def OnText(self,event):
        new = event.GetString()
        self.ChangeDir(new)

    def ChangeDir(self,path):
        #change the engine path
        eng=self.engmgr.get_current_engine()
        if eng is None:
            os.chdir(path)
            ok = True
        else:
            ok = eng.run_task('set_cwd',(path,))

        if ok is True:
            self.SetControlPath(path)
        else:
            self.SetString(self.GetSelection(),self.cur)
        #publish message
        self.tool.msg_node.publish_msg('PathManager.CWDChanged',(path,))

    def SetControlPath(self,path):
        """Set the list to display the path"""
        #check if this dir is in the list
        self.cur = path
        items = self.GetItems()
        if items.count(path)>0:
            #it is so remove it
            n = items.index(path)
            item = items.pop(n)
        #not in list, insert at 0
        items.insert(0,path)
        self.SetItems(items)
        self.SetSelection(0)

    def CheckControlPath(self):
        """Check the control is showing the right path"""
        eng=self.engmgr.get_current_engine()
        if eng is not None:
            cwd_engine = eng.run_task('get_cwd')
        else:
            cwd_engine = os.getcwd()

        #check control's display
        if cwd_engine!=self.cur:
            self.SetControlPath(cwd_engine)

    #---Message handlers--------------------------------------------------------
    def msg_eng_switched(self,msg):
        """
        Message handler for Engine.Switched message - these are emitted when 
        the active engine is changed, here we check the current working 
        directory is ok.
        """
        new,old = msg.get_data()
        if new is None:
            return
        eng=self.engmgr.get_engine(new)
        ok = eng.run_task('set_cwd', (self.cur,))
        if ok is False:
            log.error('Error setting engine cwd for path: '+str(self.cur))
