"""
PathManager Dialog

PathManDialog -  A dialog containing...
PathManPanel - A panel containing the controls for setting up python system paths etc.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('PathManager')

#---Imports---------------------------------------------------------------------
import os
import wx
from ptk_lib.controls import dialogs
import pathman_icons
import pathman_tasks                               

#-------------------------------------------------------------------------------
class PathManDialog(wx.Dialog):
    def __init__(self, parent):
        wx.Dialog.__init__(self, parent, id=-1, title='Edit system paths', 
                            size=(400, 375),
                            style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
        #set the window icon
        icon = pathman_icons.pathmanager16.GetIcon()
        self.SetIcon(icon)

        panel = PathManPanel(self)
        line = wx.StaticLine(self)

        hbox = wx.BoxSizer(wx.HORIZONTAL)
        okButton = wx.Button(self, wx.ID_OK, 'Done')
        hbox.Add(okButton, 1,wx.TOP|wx.LEFT,5)

        vbox = wx.BoxSizer(wx.VERTICAL)
        vbox.Add(panel,1,wx.EXPAND|wx.ALL,0)
        vbox.Add(line,0,wx.EXPAND|wx.TOP,10)
        vbox.Add(hbox, 0, wx.ALIGN_RIGHT|wx.ALIGN_BOTTOM| wx.ALL, 5)

        self.SetSizer(vbox)
        self.SetMinSize((400,250))


#---The path manager panel------------------------------------------------------
class PathManPanel(wx.Panel):
    def __init__(self,parent):
        wx.Panel.__init__(self,parent,id=-1)
        vsizer = wx.BoxSizer(wx.VERTICAL)

        #get a reference to the EngineManager tool
        app = wx.GetApp()
        self.engmgr = app.toolmgr.get_tool('EngineManager')

        #register enginetasks
        eng = self.engmgr.get_current_engine()
        tasks =  eng.get_registered_tasks()
        if 'get_sys_path' not in tasks:
            eng.register_task(pathman_tasks.get_sys_path)
        if 'add_to_sys_path' not in tasks:
                eng.register_task(pathman_tasks.add_to_sys_path)
        if 'remove_from_sys_path' not in tasks:
                eng.register_task(pathman_tasks.remove_from_sys_path)
        if 'move_up_sys_path' not in tasks:
                eng.register_task(pathman_tasks.move_up_sys_path)
        if 'move_down_sys_path' not in tasks:
                eng.register_task(pathman_tasks.move_down_sys_path)

        #add a list box of sys.path
        text = wx.StaticText(self,-1,'Python module search path (sys.path):')
        paths = eng.run_task( 'get_sys_path' )
        self.plist = wx.ListBox(self,-1, choices=paths, 
                    style=wx.LB_SINGLE|wx.HSCROLL)
        vsizer.Add(text,0,wx.ALIGN_LEFT|wx.ALL,10)
        vsizer.Add(self.plist,1,
                wx.EXPAND|wx.ALIGN_CENTER_HORIZONTAL|wx.ALIGN_BOTTOM|wx.LEFT|wx.RIGHT,10)

        #add remove/add buttons
        self.add    = wx.Button(self, -1, "Add")
        self.Bind(wx.EVT_BUTTON, self.OnAdd, self.add)

        self.remove = wx.Button(self, -1, "Remove")
        self.Bind(wx.EVT_BUTTON, self.OnRemove, self.remove)

        #add updown buttons
        self.up    = wx.Button(self, -1, "Move Up")
        self.Bind(wx.EVT_BUTTON, self.OnMoveUp, self.up)

        self.down = wx.Button(self, -1, "Move Down")
        self.Bind(wx.EVT_BUTTON, self.OnMoveDown, self.down)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        hsizer.Add(self.add,1,wx.EXPAND|wx.ALL,2)
        hsizer.Add(self.remove,1,wx.EXPAND|wx.ALL,2)
        hsizer.Add(self.up,1,wx.EXPAND|wx.ALL,2)
        hsizer.Add(self.down,1,wx.EXPAND|wx.ALL,2)
        vsizer.Add(hsizer,0,wx.ALIGN_CENTER_HORIZONTAL|wx.ALL,0)
        self.SetSizer(vsizer)

    
    def OnAdd(self,event):
        """Add a new path to the sys.path"""
        dirname = wx.DirSelector(defaultPath=os.getcwd())
        if dirname!='':
            eng = self.engmgr.get_current_engine()
            #add the dir to the engines sys.path
            paths = eng.run_task('add_to_sys_path',(dirname,))
            #update the list
            self.plist.SetItems(paths)

    def OnRemove(self,event):
        """Removes the currently selected path from sys.path"""
        #first get the selection
        n = self.plist.GetSelection()
        if n==-1:
            return
        msg='Delete selected path from search paths?'
        title='Confirm path removal'
        ans=dialogs.ConfirmDialog(msg,title)
        if ans is True:
            dirname = self.plist.GetString(n)
            #remove the dir from the engines sys.path
            eng = self.engmgr.get_current_engine()
            paths = eng.run_task('remove_from_sys_path',(dirname,))
            #update the list
            self.plist.SetItems(paths)

    def OnMoveUp(self,event):
        #first get the selection
        n = self.plist.GetSelection()
        if n==-1:
            return
        dirname = self.plist.GetString(n)
        #move the dir up in the engines sys.path
        eng = self.engmgr.get_current_engine()
        paths,newn = eng.run_task('move_up_sys_path',(dirname,))
        #update the list
        self.plist.SetItems(paths)
        self.plist.SetSelection(newn)

    def OnMoveDown(self,event):
        #first get the selection
        n = self.plist.GetSelection()
        if n==-1:
            return
        dirname = self.plist.GetString(n)
        #move the dir down in the engines sys.path
        eng = self.engmgr.get_current_engine()
        paths,newn = eng.run_task('move_down_sys_path',(dirname,))
        #update the list
        self.plist.SetItems(paths)
        self.plist.SetSelection(newn)
