"""
Console.py 
----------

ConsoleInterface:
The console interface to engines - a subclass of the MessageBus MBNode with a 
node name 'Console.engname'.
A console class should subclass this and implement the console methods prompt, 
read_stdin etc to define what to do when communicating with the engine.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('Engine.console')
log.setLevel(logging.DEBUG)

from ptk_lib.message_bus.mb_node import MBNode
from ptk_lib.message_bus import mb_protocol

import messages

class ConsoleInterface(MBNode):
    def __init__(self, msg_bus, engname):
        MBNode.__init__(self, msg_bus, 'Console.'+engname)

        #store engine name and node address for communications
        self.engname = engname
        self.engnode = 'Engine.'+engname

        #flags to keep track of engine state
        self._connected = False
        self._busy      = False

        #default engine prompts
        self._ps1 = '>>> '      #normal prompt              
        self._ps2 = '... '      #continuation prompt
        self._ps3 = '??? '      #reading prompt
        self._ps4 = 'DB> '      #debugging prompt

        #set handlers for messages
        #disconnection messages
        self.subscribe(mb_protocol.SYS_NODE_DISCONNECT ,self.msg_sys_disconnect)     
        self.subscribe(mb_protocol.SYS_NODE_ERR_DISCONNECT,self.msg_sys_disconnect)  

        #prompt messages
        self.set_handler( messages.CON_PROMPT      , self.msg_prompt)
        self.set_handler( messages.CON_PROMPT_STDIN, self.msg_prompt_stdin)
        self.set_handler( messages.CON_PROMPT_DEBUG, self.msg_prompt_debug)

        self.set_handler( messages.CON_WRITE_STDOUT, self.msg_write_stdout)
        self.set_handler( messages.CON_WRITE_STDERR, self.msg_write_stderr)
        self.set_handler( messages.CON_WRITE_DEBUG , self.msg_write_debug)

        self.set_handler( messages.CON_CLEAR, self.msg_console_clear)
        self.set_handler( messages.CON_EXEC_SOURCE, self.msg_exec_source)

        #finally set the engines console to be this console.
        self.connect()

    #---console methods---------------------------------------------------------
    def connect(self):
        """ 
        Connect to the engine node by sending the SetConsole message.
        """
        self._connected = True
        res = self.send_msg(self.engnode, messages.ENG_SET_CONSOLE,
                             (self._name,), get_result=True)

        #reply is welcome message and prompt strings
        welcome, self._ps1, self._ps2, self._ps3 = res
        
        #write the welcome message and prompt for first command
        self.write_stdout(welcome)
        self.prompt()

    def disconnect(self):
        """
        Called when the engine disconnects.
        """
        #write an error
        self._connected = False
        self.engnode = None
        self.write_stderr('\nEngine disconnected\n')
        #remove the console node from the message bus.
        if self._connected is True:
            self.close()

    def push(self, line):
        """
        Push a line of user command to the engine.
        """
        self._busy=True #assume busy until prompted
        #send the line to the engine
        self.send_msg( self.engnode, messages.ENG_PUSH, (line,))

    #Overload these in subclass 
    def prompt(self, more=False):
        """
        Called when the engine prompts for more user commands, more is True if 
        the previously pushed line is imcomplete.
        """
        #cannot be busy if the engine is prompting
        self._busy   = False
        #do stuff here in subclasses

    def prompt_stdin(self, more=False):
        """
        Called when the engine prompts for user input on std_in, 
        more is True if multiple lines are required or None to disable the
        reading prompt.
        """
        pass

    def prompt_debug(self,more=False):
        """
        Called when the debugger prompts for debugger commands.
        more=False then a new debugger prompt.
        more=True then it is a line continuation
        more=None to disable debugger prompt.
        """
        pass
    
    def write_stdout(self, string):
        """
        Called when the engine wants to write a string to the std_out.
        """
        pass
    
    def write_stderr(self, string):
        """
        Called when the engine wants to write a string to the std_err.
        """
        pass

    def write_debug(self, string):
        """
        Called when the engine wants to write a debugger message
        """
        pass

    def clear_console(self):
        """
        Clear the console.
        """
        pass

    def exec_source(self, source):
        """
        Add the source to the console and push it to the engine as if the user 
        typed the code.
        """
        pass

    #---message handlers--------------------------------------------------------
    def msg_sys_disconnect(self, msg):
        """
        Message handler for MessageBus system message indicating if an engine
        node disconnected.
        """
        node_name = msg.get_data()[0]
        if node_name!=self.engnode:
            return
        self.disconnect()

    def msg_prompt(self, msg):
        more = msg.data[0]
        self.prompt(more)

    def msg_prompt_stdin(self, msg):
        more = msg.data[0]
        self.prompt_stdin(more)

    def msg_prompt_debug(self, msg):
        more = msg.data[0]
        self.prompt_debug(more)

    def msg_write_stdout(self, msg):
        string = msg.data[0]
        self.write_stdout(string)

    def msg_write_stderr(self, msg):
        string = msg.data[0]
        self.write_stderr(string)

    def msg_write_debug(self, msg):
        string = msg.data[0]
        self.write_debug(string)

    def msg_console_clear(self, msg):
        self.clear_console()

    def msg_exec_source(self, msg):
        source= msg.data[0]
        self.exec_source(source)

    
