"""
The engine is the object that actually executes user commands and handles 
debugger and profiler functions.

It can exists in a seperate process (external engines) or within the interface
process (internal engine). 

It uses a message bus node to commuication with it's console and other tools, so
other extensions can be added to the engine by defining new message types and 
registering handlers with the engine object. This can be found by importing the
__main__ namespace, at __main__.engine

Engine structure:
-----------------
Engine
    MessageBus Node -   for sending/recieving control and IO messages.
    Compiler        -   Handles the compilation of python code and display of errors.
    Debugger        -   debugger traceback object.
    Profiler        -   profiler function object.

Exiting engines:
----------------
Three methods: 
eng.close() - used to close the message bus connection and trigger the exit.
node.on_close() - trigger by the node closing.
eng._close() - called in on_close the actually exit the engine process.

PTK closes engine via SYS_NODE_CLOSE message or SystemExit:
        call eng.close()
            - sets _exiting flag to prevent new prompt message
            - closes msgbus node.
        node.on_close() is called this calls eng._close method.
            - stop running code if busy via Keyboard interupt.
                (prevent next prompt with exiting flag)
            - exit process.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('Engine')
#log.setLevel(logging.DEBUG)

#---imports---------------------------------------------------------------------
from codeop import _maybe_compile, Compile
from code import softspace
import sys
import __main__
import __builtin__                      #for adding builtin commands
import marshal                          #for task/builtins
import types                            #for task/builtins
from threading import Event             #for readline events
import thread                           #to interupt, running code.

from eng_compiler import EngineCompiler
from eng_debugger import EngineDebugger
from eng_profiler import EngineProfiler
import messages                     #standard engine message types
import eng_tasks                        #engine task utils  


#---pseudo file object----------------------------------------------------------
class PseudoFile:
    def __init__(self,readline=None,readlines=None,write=None):
        """Create a file-like object."""
        #setup inputs:
        if readline is not None:
            self.readline  = readline
        if readlines is not None:
            self.readlines = readlines
        #setup output:
        if write is not None:
            self.write = write

    def readline(self):
        pass
    def write(self, s):
        pass
    def writelines(self, l):
        map(self.write, l)
    def flush(self):
        pass
    def isatty(self):
        return 1

#---Engine----------------------------------------------------------------------
class Engine(object):
    engtype = 'Engine'
    def __init__(self, engname, userdict=None,node=None):
        """
        Base class implementation of an Engine, 
        A subclass should provide a MessageBus node object to send/recieve 
        messages.
        """
        #-----------------------------------------------------------------------
        # Init message bus node
        #-----------------------------------------------------------------------
        if node is None:
            raise Exception('No MessageBus node object provided!')
        self.node = node  
        self.console = None     #name of controlling console

        #add shortcuts to the node methods
        self.send_msg = node.send_msg
        self.set_handler = node.set_handler
        self.remove_handler = node.remove_handler
        self.publish_msg = node.publish_msg
        self.subscribe = node.subscribe
        self.unsubscribe = node.unsubscribe
     
        #-----------------------------------------------------------------------
        # Attributes
        #-----------------------------------------------------------------------
        self.engname = engname
        self._ps1 = '>>> '              #engine prompts are changable, these are
        self._ps2 = '... '              #what is shown in the console, so an
        self._ps3 = '??? '              #engine may want to display it's name.

        #flags
        self._busy     = False          #running a user command
        self._debug    = False          #use traceback debugger
        self._profile  = False          #use profiler
        self._exiting  = False          #used to prevent prompt/EngineStop when exiting
        self._stop     = False          #used to catch stop running code exceptions

        #dictionary to execute user commands in.
        if userdict is None:
            userdict = __main__.__dict__
        self._userdict = userdict

        #store registered engine tasks {name:func}
        self._tasks = {}

        #Compiler handles the compilation of code.
        self.compiler = EngineCompiler(self)

        #Debugger handles the exectution of code with the debugger enabled
        self.debugger = EngineDebugger(self)

        #Profiler handles the execution of code with the profiler enabled
        self.profiler = EngineProfiler(self)

        #-----------------------------------------------------------------------
        # Set up the working environment
        #-----------------------------------------------------------------------
        #the builtins module
        exec('import __builtin__ as __builtins__',self._userdict)
        
        #store the engine process object in the main namespace for use by 
        #extensions
        __main__.engine = self

        #register standard engine tasks
        self.register_task(eng_tasks.object_exists)
        self.register_task(eng_tasks.get_type_string)
        self.register_task(eng_tasks.execute_startup_script)
        self.register_task(eng_tasks.get_cwd)
        self.register_task(eng_tasks.set_cwd)

        #-----------------------------------------------------------------------
        # redirect standard input/output
        #-----------------------------------------------------------------------
        #attributes for reading from stdin
        self._isreading = False     #flag indicating we are waiting for input.
        self._readend = ''          #string that indicates enough.
        self._readresult = ''       #read input goes here for collection.
        self._readevent = Event()   #event to wake readline

        #this redirects all sys.std* io to the gui process
        self.stdin  = PseudoFile(readline=self.readline, readlines=self.readlines)
        self.stdout = PseudoFile(write=self.write_stdout)
        self.stderr = PseudoFile(write=self.write_stderr)

        sys.stdin   = self.stdin
        sys.stdout  = self.stdout
        sys.stderr  = self.stderr

        #-----------------------------------------------------------------------
        #Set up the interface/engine communications
        #   - compiler/debugger/profiler also register message handlers
        #-----------------------------------------------------------------------  
        self.set_handler(messages.ENG_GET_TYPE, self.msg_get_type)      
        self.set_handler(messages.ENG_SET_CONSOLE, self.msg_set_console)
        self.set_handler(messages.ENG_PUSH, self.msg_push)
        self.set_handler(messages.ENG_STOP, self.msg_stop)

        self.set_handler(messages.ENG_DEBUG_TOGGLE,  self.msg_toggle_debug)
        self.set_handler(messages.ENG_PROFILE_TOGGLE, self.msg_toggle_profile)

        self.set_handler(messages.ENG_EXEC_COMMAND, self.msg_exec)
        self.set_handler(messages.ENG_EVAL_COMMAND, self.msg_eval)
        self.set_handler(messages.ENG_RUN_TASK, self.msg_run_task)
        self.set_handler(messages.ENG_REGISTER_TASK, self.msg_register_task)
        self.set_handler(messages.ENG_ADD_BUILTIN, self.msg_add_builtin)
        self.set_handler(messages.ENG_GET_TASKS, self.msg_get_tasks)
        
    #---------------------------------------------------------------------------
    # Interface methods
    #---------------------------------------------------------------------------
    def push_line(self, line):
        """A new line of user input"""
        ##line is user command compile it
        ismore,code,err = self.compiler.compile(line)

        #need more 
        #   - tell the console to prompt for more 
        if ismore:
            self.send_msg(self.console, messages.CON_PROMPT,(True,))
            return

        #syntax error
        #   - compiler will output the error
        #   - tell the console to prompt for new command 
        if err:
            self.send_msg(self.console, messages.CON_PROMPT,(False,))
            return

        #no code object - could be a blank line
        if code is None:
            self.send_msg(self.console, messages.CON_PROMPT,(False,))
            return

        ##have some code to execute so call run_user_command to do something with it
        self.run_user_command(code)

    def run_code(self, code):
        """
        Run the code object in the engine
        """
        #set busy flag and send busy message
        self._busy = True
        data = (self.engname, self._debug, self._profile)
        try:
            self.publish_msg( messages.ENGINE_BUSY,data)
        except:
            log.exception('Cannot publish message!')

        #enable debugger?
        if self._debug is True:
            trace_func = self.debugger
        else:
            trace_func = None

        #enable profiler?
        if self._profile is True:
            profile_func = self.profiler
        else:
            profile_func = None

        #run the code
        try:
            sys.settrace(trace_func)
            sys.setprofile(profile_func)
            exec code in self._userdict
            sys.settrace(None)
            sys.setprofile(None)

        #system exit  - close engine
        except SystemExit:
            sys.settrace(None)
            sys.setprofile(None)
            self._busy = False
            log.debug('system exit in runnning code')
            self.close()

        #keyboard interrupt stopped running code
        except KeyboardInterrupt:
            sys.settrace(None)
            sys.setprofile(None)
            self._busy = False

            #engine stopped code in order to exit
            if self._exiting:
                return

            #user stopped code
            self._stop = False
            if self._isreading: #cancel the read prompt.
                self.send_msg( self.console, messages.CON_PROMPT_STDIN, data=(None,))
            sys.stderr.write('STOP: User forced running code to stop.\n\n')
        

        #other exception - could be an error 1) caused by the engine exiting 2) a different error caused by
        #the KeyboardInterrupt (wxpython doesn't play nice!) or 3) a user code error
        except:
            sys.settrace(None)
            sys.setprofile(None)
            self._busy = False

            #1) engine is exiting   -  probably some error caused by engine exiting
            if self._exiting:
                log.exception('Exception raised to stop running code? - engine wants to exit.')
                return

            #2) user stopped code
            if self._stop is True:
                self._stop = False
                if self._isreading: #cancel the read prompt.
                    self.send_msg( self.console, messages.CON_PROMPT_STDIN, data=(None,))
                sys.stderr.write('STOP: User forced running code to stop.\n\n')

            #3) error in user code.
            self.compiler.show_traceback()

        #If exiting skip the rest.
        if self._exiting is True:
            return

        #softspace makes the print statement work correctly when using final 
        #comma to supress newlines.
        if softspace(sys.stdout, 0):
            print 

        #reset internal state flags
        self._busy = False
        self._isreading = False
        if self._debug is True:
            self.debugger.reset()

        #publish an engine done message
        try:
            data = (self.engname, self._debug, self._profile)
            self.publish_msg(messages.ENGINE_DONE, data)  
        except:
            pass

        #prompt the console for new command
        try:
            self.send_msg(self.console, messages.CON_PROMPT,(False,))
        except:
            pass

    def stop_code(self):
        """
        Attempt to stop the running code by raising a keyboard interrupt in
        the main thread
        """
        if self._busy is False:
            return
        #set the stop flag to catch the error correctly
        self._stop = True

        #make sure we are not stuck in readline(s) 
        if self._isreading:
            self._readevent.set()

        #make sure the debugger is not paused.
        if self._debug is True:
            self.debugger.stop_code()
            return

        #try a keyboard interrupt - this will not work for the internal engine 
        # as the error is raised here instead of the running code, hence put in 
        #try clause.
        try:
            thread.interrupt_main()
        except:
            pass

    def close(self):
        """
        Close the engine 
        - by closing the connection to the message bus
        """
        #all close tasks are done in _close which is called from the message bus
        #nodes on_close method.
        self.node.close()

    def _close(self):
        """
        Engine is about to close (called when node is closed) 
        -   stop any running code.
        """
        self._exiting = True
        #restore stdIO
        sys.stdin   = sys.__stdin__
        sys.stdout  = sys.__stdout__
        sys.stderr  = sys.__stderr__
        if self._busy is True:
            log.info('Closing engine '+self.engname+' - stopping running code')
            #stop any running code
            self.stop_code()
        log.info('Closing engine '+self.engname+'- doing engine exit process tasks')
        self.exit_process()

    def evaluate(self,expression):
        """
        Evaluate expression in engine and return result (like builtin eval)
        
        It is intened to be used to provide functionality to the GUI, so commands
        should be fairly quick to process to avoid blocking.
        
        Returns None on errors
        """
        globals = self._userdict
        locals = self._userdict

        #try to evaluate the expression
        try:
            result = eval(expression, globals, locals)
        except:
            result = None
        return result

    def execute(self,expression):
        """
        Execute expression in engine (like builtin exec)

        It is intened to be used to provide functionality to the GUI interface,
        so commands should be fairly quick to process to avoid blocking the 
        communications thread.
        """
        globals = self._userdict
        locals = self._userdict

        #execute the expression
        try:
            exec(expression, globals, locals)
        except:
            pass

    def run_task(self,taskname, args=(), kwargs={}):
        """
        Run a complex task in the engine; taskname is the name of a registered 
        task function taking the following arguments: 
                task(globals, locals, *args, **kwargs)

        Returns None on errors.
        """
        #get the task
        task = self.get_task(taskname)
        if task is None:
            return none

        globals = self._userdict
        locals = self._userdict
        
        #run the task
        try:
            result=task(globals, locals, *args, **kwargs)
        except:
            log.exception('run_task failed :'+task.func_name)
            result=None
        return result

    def register_task(self,task):
        """
        Register a task with the process.
        """
        log.debug('Registering task '+task.func_name)
        if self._tasks.has_key(task.func_name):
            result = False
        else:
            self._tasks[task.func_name]=task
            result=True
        return result

    def get_task(self, taskname):
        """
        Get a registered task callable.
        """
        #get the task
        if self._tasks.has_key(taskname) is False:
            log.warning('Unknown task name: '+str(taskname))
            return None
        task = self._tasks[taskname]
        return task

    def enable_debug(self,flag=True):
        """ Enable the debugger - returns debug state """

        #already set to the correct state
        if flag == self._debug:
           return self._debug

        #engine is busy
        if (self._busy is True):
            #cannot enable/disable debugging when running a command.
            return self._debug

        #engine is inactive set to flag
        self._debug = flag

        #publish debug toggle message
        try:
            data = (self.engname, self._debug)
            self.publish_msg(messages.ENGINE_DEBUG_TOGGLED, data)  
        except:
            pass
        return self._debug

    def enable_profile(self, flag=True):
        """ Enable the profiler - returns profiler state """
        #cannot enable/disable profiler if running a command.
        if self._busy is True:
            return self._profile

        #enable profiler/disable debugger
        if flag is True:
            self._profile = True    
        else:
            self._profile = False

        #publish debug toggle message
        try:
            data = (self.engname, self._profile)
            self.publish_msg(messages.ENGINE_PROFILE_TOGGLED, data)  
        except:
            pass

        #disable debugger if enabling profiler
        if (flag is True) and (self._debug is True):
           self.enable_debug(False)

        return self._profile

    #---------------------------------------------------------------------------
    # standard IO methods
    #---------------------------------------------------------------------------
    def readline(self):
        """std in readline directs here- this calls the stdin handler for input"""
        #set flag
        self._isreading = True
        self._readend = '\n'  #string to indicate enough
        self._readresult = ''
        self._readevent.clear()

        #request input from the interface
        self.send_msg(self.console, messages.CON_PROMPT_STDIN,(False,))

        #wait until _readevent is set
        self._readevent.wait()
        self._isreading = False

        #reset the event for next read
        self._readevent.clear()
        #reset the read result buffer
        result = self._readresult
        self._readresult = ''
        return result

    def readlines(self):
        """std in readlines directs here"""
        #set flag
        self._isreading = True
        self._readend = '\n\n' #string to indicate enough
        self._readresult = ''
        self._readevent.clear()

        #request input from the interface
        self.send_msg(self.console, messages.CON_PROMPT_STDIN,(False,))

        #wait until _readevent is set
        self._readevent.wait()
        self._isreading = False

        #reset the event for next read
        self._readevent.clear()
        #reset the read result buffer
        result = self._readresult
        self._readresult = ''
        return result

    def push_stdin(self, line):
        #check if we have enough input
        if line.endswith(self._readend):
            #store here to be collected in readline(s)
            self._readresult = line 
            #wake readline
            self._readevent.set()
        else:
            self.send_msg(self.console, messages.CON_PROMPT_STDIN,(True,))

    def write_stdout(self,string):
        """std out write redirects here"""
        try:
            self.send_msg(self.console, messages.CON_WRITE_STDOUT,(string,))        
        except:
            pass        

    def write_stderr(self,string):
        """std err write redirects here"""
        try:
            self.send_msg(self.console, messages.CON_WRITE_STDERR,(string,))
        except:
            pass        

    #---------------------------------------------------------------------------
    # Methods to implement in subclasses
    #---------------------------------------------------------------------------
    def run_user_command(self,code):
        """
        Called to run the compiled user command that was pushed to the engine 
        process. Typically this needs to be run in the main
        thread - not the comms thread which calls this method, so an event of 
        some sort should be raised here to wake the main thread.
        """ 
        pass

    def exit_process(self):
        """
        Things to do when exiting the process
        - such as wake the main loop.
        """
        pass

    def get_welcome(self):
        """Return the standard part of engines welcome message"""
        ver = sys.version
        plat = sys.platform
        info = "Type \"help\", \"copyright\", \"credits\" or \"license\" for more information.\n"
        welcome = ( "Python "+ver+" on "+plat+" \n"+info)
        return welcome
    #---------------------------------------------------------------------------
    # message handlers
    #---------------------------------------------------------------------------
    def msg_get_type(self, msg):
        """
        Return the engine type string - sent by the enginemanger to a newly 
        connected engine to find its type.
        """
        return self.engtype

    def msg_set_console(self, msg):
        """
        Set the name of the controlling console - used to send messages and
        return the welcome message.
        """
        con_name, = msg.get_data()
        self.console = con_name
        welcome = self.get_welcome()
        return welcome, self._ps1, self._ps2, self._ps3

    def msg_push(self, msg):
        """Process a line from the console"""
        line, = msg.get_data()
        ##data is a line of input
        if self._isreading:
            log.debug('stdin data recieved '+str(line))
            self.push_stdin(line)
        ##debugger is active and we are running a command
        elif self._debug and self._busy:
            log.debug('debugger cmd recieved '+str(line))
            self.debugger.push_line(line)
        ##a user command
        else:
            log.debug('user command recieved '+str(line))
            self.push_line(line)

    def msg_toggle_debug(self, msg):
        """Enable the debugger traceback"""
        res = self.enable_debug(msg.data[0])
        return res

    def msg_toggle_profile(self, msg):
        """Enable the profiler"""
        res = self.enable_profile(msg.data[0])
        return res

    def msg_stop(self, msg):
        self.stop_code()

    def msg_exec(self, msg):
        """Execute expression"""
        #data can have optional level argument for use with debugger
        if len(msg.data)==2:
            expression, level = msg.data
        else:
            expression, = msg.data
            level = None

        #if debugging and paused use the debugger interfaces
        if self._debug and self.debugger._paused:
            self.debugger.execute(expression, level)
        else:
            self.execute(expression)

    def msg_eval(self, msg):
        """Evaluate expression and return result"""
        #data can have optional level argument for use with debugger
        if len(msg.data)==2:
            expression, level = msg.data
        else:
            expression, = msg.data
            level = None

        #if debugging and paused use the debugger interfaces
        if self._debug and self.debugger._paused:
            result = self.debugger.evaluate(expression, level)
        else:
            result = self.evaluate(expression)
        return result

    def msg_run_task(self, msg):
        """Run an pre-registered engine task and return the result"""
        #data is a task name, args, kwargs plus optional level argument for use 
        #with debugger
        if len(msg.data)==4:
            taskname,args,kwargs,level = msg.data
        else:
            taskname,args,kwargs = msg.data
            level = None

        #if debugging and paused use the debugger interfaces
        if self._debug and self.debugger._paused:
            result = self.debugger.run_task(taskname, args, kwargs, level)
        else:
            result = self.run_task(taskname, args, kwargs)
        return result

    def msg_register_task(self, msg):
        """Register a new task with the engine"""
        #data is name and marshalled code object
        s = msg.data[0]
        c = marshal.loads(s)
        task = types.FunctionType(c,{'__builtins__':__builtins__},None)
        result = self.register_task(task)
        return result

    def msg_get_tasks(self, msg):
        """Return the registered task names"""
        return self._tasks.keys()

    def msg_add_builtin(self, msg):
        """add a command to the builtins module"""
        name, s = msg.get_data()
        code = marshal.loads(s)
        cmd = types.FunctionType(code,{'__builtins__':__builtins__},None)
        __builtin__.__dict__[name] = cmd
        return True
    
