"""
Internal engine - an engine subclass for the special case of an 
internal (same process) engine. This just uses a MessageBus MBNode to connect to
the message bus and for communication.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('InternalEngine')

#---imports---------------------------------------------------------------------
import time

from engine import Engine                                   #base class
from ptk_lib.message_bus.mb_node import MBNode              #for communications

#-------------------------------------------------------------------------------
class InternalEngNode(MBNode):
    def __init__(self, eng, msg_bus, name):
        """
        Subclassed message bus node for internal engine - calls eng._close when
        the node disconnects from the message bus.
        """
        self.eng = eng
        MBNode.__init__(self, msg_bus, name)

    def on_close(self):
        self.eng._close()

#-------------------------------------------------------------------------------
class PseudoEvent():
    """
    An object with the same interface as a threading.Event for the internal 
    engine. This prevents the readline/readlines and debugger from blocking
    when waiting for user input.

    This calls the doyield function and sleeps until the event is set.
    """
    def __init__(self, eng):
        self._eng = eng
        self._doyield = eng.doyield
        self._set = False

    def set(self):
        self._set = True

    def clear(self):
        self._set = False

    def wait(self, timeout=None):
        if timeout is not None:
            raise NotImplementedError('Timeout not implemented')
        log.debug('waiting: '+str(self._doyield))
        while (self._set is False) and (self._eng._stop is False):
            self._doyield()
            time.sleep(0.05)

#-------------------------------------------------------------------------------
class InternalEngine(Engine):
    engtype = 'Internal'
    def __init__(self, mb, userdict={}, doyield=None):
        """
        Create an internal engine object.
        mb          -   MessageBus instance
        userdict    -   Dictionary to execute user commands in.
        doyield     -   A callable to yield to a running mainloop (i.e. wx.Yield)
                        Used to allow the GUI to update and return stdinput and
                        ensure message are sent in the correct order.
        """
        #store a reference to the message bus and create a new node.
        self.mb = mb
        node =  InternalEngNode(self, mb, 'Engine.Internal' )

        #store reference to the user supplied callables
        self.doyield = doyield

        #call base class init
        Engine.__init__(self, 'Internal', userdict,node)

        #for internal engines we need to replace the readevent and debugger
        #resume events and the send_msg/publish_msg methods to allow the GUI 
        #interface a chance to run
        self._readevent = PseudoEvent(self)
        self.send_msg = self._send_msg
        self.publish_msg = self._publish_msg

    def _send_msg(self, to, subject, data=(), get_result=False):
        """
        Send a message via the MessageBus.
        to      -   Node to send message to.
        subject -   The message subject.
        data    -   A list or tuple data object.
        get_result  -   True/False wait for and return the result.

        Returns - the result value if get_result is True, or None
        """
        self.node.send_msg(to, subject, data, get_result)
        #yield to the gui mainloop to ensure message is sent.
        if self.doyield is not None:
            self.doyield()
        
    def _publish_msg(self, subject, data=()):
        self.node.publish_msg(subject, data)
        #yield to the gui mainloop to ensure message is sent.
        if self.doyield is not None:
            self.doyield()

    #---------------------------------------------------------------------------
    def run_user_command(self,code):
        """
        Called to run the compiled user command that was pushed to the engine 
        process.

        As this is in the main process it will only be called in the main 
        thread so we can just call run_code.
        """ 
        #run code
        self.run_code(code) 

    def exit_process(self):
        """
        Things to do when exiting the process
        """
        import __main__
        import sys
        del __main__.engine 

    def get_welcome(self):
        """Return the engines welcome message"""
        welcome = Engine.get_welcome(self) + "\n\nRunning in the interface process\n"
        return welcome

    def readline(self):
        """std in readline directs here- this calls the stdin handler for input"""
        if self._stop:
            raise KeyboardInterrupt
        return Engine.readline(self)

    def readlines(self):
        """std in readlines directs here"""
        if self._stop:
            raise KeyboardInterrupt
        return Engine.readlines(self)

    def write_stdout(self,string):
        """std out write redirects here"""
        if self._stop:
            raise KeyboardInterrupt
        Engine.write_stdout(self, string)

    def write_stderr(self,string):
        """std err write redirects here"""
        if self._stop:
            raise KeyboardInterrupt
        Engine.write_stderr(self, string)

    def enable_debug(self,flag=True):
        """ 
        Enable the debugger - returns debug state 
        Not available in internal engine so always return False        
        """
        self.write_stderr('Debugger is not available in Internal engine')
        return False

    
