"""
MessageBus Node

These objects register with the MessageBus to send and recieve messages.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('MessageBus.MBNode')

#---Imports---------------------------------------------------------------------
import threading

#Other messenger components
import mb_protocol
from mb_misc import AsyncResult, WeakRefHandler, AddressError

#-------------------------------------------------------------------------------
# A standard MessageBus Node which can be used to send, recieve and publish 
# messages.
# Callables can be registerd for particular message subjects via 
# register_handler() and for published message with particular subjects via 
# subscribe().
#-------------------------------------------------------------------------------
class MBNode(object):
    def __init__(self, msg_bus, name):
        """
        Create a standard MessageBus Node which can be used to send, recieve and
        publish messages. 

        Subclass and overload the recive_msg and recieve_published methods to 
        handle incomming messages.
        """
        #node name
        self._msg_bus = msg_bus
        self._name = name
        self._registered = False

        #Use to make things thread safe
        self._lock = threading.Lock()   
     
        #message counter - used to generate unique message ids
        self._count = 0

        #message handlers for addressed messages
        self._msg_hndlrs = {}  #dictionary of {subject:handler}
        
        #handlers for published messagess
        self._sub_hndlrs = {} #dictionary of {subject:[handlers]}

        self.set_handler(mb_protocol.SYS_NODE_CLOSE, self.process_sys_msg)

        #egister the node with the parent MessageBus
        self._msg_bus.register_node( self, self._name )

    #---------------------------------------------------------------------------
    def close(self):
        """
        Close the node and disconnect from the MessageBus
        """
        if self._registered is False:
            return
        self._registered = False
        self._msg_bus._remove_node(self._name, err=False)
        #call on call to do subclass specific method.
        self.on_close()

    #---addressed messages------------------------------------------------------
    def send_msg(self, to, subject, data=(), get_result=False):
        """
        Send a message via the MessageBus.
        to      -   Node to send message to.
        subject -   The message subject.
        data    -   A list or tuple data object.
        get_result  -   True/False wait for and return the result.

        Returns - the result value if get_result is True, or None
        """
        if self._registered is False:
            raise Exception('Node not registered with a MessageBus!')

        #construct the message object
        self._lock.acquire()
        msgid = str(self._name)+':'+str(self._count)
        self._count+=1
        self._lock.release()
        if get_result is True:
            msgtype = mb_protocol.MSG_ADD_RESULT
        else:
            msgtype = mb_protocol.MSG_ADD_NO_RESULT

        msg = mb_protocol.Message( msgid, msgtype,from_node=self._name,
                       to_node=to, subject=subject, data=data)

        #result object allows the reply/result to be returned to the sending 
        #thread
        if get_result is True:
            res_obj = AsyncResult(msgid, self._msg_bus._timeout)
            msg.set_result = res_obj.set_result
            msg.get_result = res_obj.get_result

        #send the message to the parent MessageBus
        self._msg_bus.process_msg(msg)

        if get_result is False:
            return None

        #get the reply/result - this will block the sending thread until a reply
        #is set using reply(res)
        res = msg.get_result()

        #check if an error
        if isinstance(res, Exception):
            raise res

        return res

    def set_handler(self, subject, handler):
        """
        Register a message handler for the addressed message subject given.
        """
        #get the lock
        self._lock.acquire()
        #store handler into dictionary
        self._msg_hndlrs[subject]=WeakRefHandler(handler) 
        #release the lock
        self._lock.release()

    def remove_handler(self, subject):
        """
        Remove the message handler for the message subject.
        """
        #get the lock
        self._lock.acquire()
        #store handler into dictionary
        self._msg_hndlrs.pop(subject,None) 
        #release the lock
        self._lock.release()

    def process_msg(self, msg):
        """
        Handles incomming messages addressed to this node
        msg         -   the message object
        """
        #get handler
        self._lock.acquire()
        handler = self._msg_hndlrs.get(msg.subject,None)
        self._lock.release()
        if handler is not None:
            #check if the handler is dead
            if handler.isdead():
                #remove it
                self._lock.acquire()
                self._msg_hndlrs.pop(msg.subject)
                self._lock.release()
            else:   
                #call it        
                try:
                    res = handler(msg)
                except:
                    log.exception('Handler failed: '+str(handler))
                    res=None
        else:
            log.info('No handler found for subject: '+str(msg.subject))
            res = None
        #return the result via the reply object/callable at msg.set_result
        msg.set_result(res)

    #---published messages------------------------------------------------------
    def publish_msg(self, subject, data=()):
        if self._registered is False:
            raise Exception('Node not registered with a MessageBus!')

        #construct the message object
        self._lock.acquire()
        msgid = str(self._name)+':'+str(self._count)
        self._count+=1
        self._lock.release()
        msg = mb_protocol.Message( msgid, msgtype=mb_protocol.MSG_PUBLISHED, 
                        from_node=self._name, to_node='', 
                        subject=subject, data=data )

        #publish the message via the parent MessageBus
        self._msg_bus.process_published(msg)

    def subscribe(self, subject, handler):
        """
        Subscribe a message handler for the published message subject given.
        """
        #get the lock
        self._lock.acquire()

        #get current list of handlers
        hlist = self._sub_hndlrs.get(subject,[])
        #add a weakref to the handler
        hlist.append(WeakRefHandler(handler)) 
        #store back into dictionary
        self._sub_hndlrs[subject]=hlist 

        #release the lock
        self._lock.release()

    def unsubscribe(self, subject, handler):
        """
        Remove the message handler subscribed for the message type.
        """
        self._lock.acquire()
        hlist = self._sub_hndlrs.get(subject,[])
        #remove the handler
        for n in range(0,len(hlist)):
            h = hlist[n]
            if h == handler:
                n = hlist.index(h)
                hlist.pop(n)
        #store back into dict
        self._sub_hndlrs[subject] = hlist
        self._lock.release()

    def process_published(self, msg):
        """
        Handles incomming published messages.
        """
        #get message handlers
        self._lock.acquire()
        hlist = self._sub_hndlrs.get(msg.subject,[])
        self._lock.release()
        
        #call all handlers
        new_hlist=[]
        for handler in hlist:
            #check if the handler is dead
            if handler.isdead():
                pass
            else: 
                new_hlist.append(handler)
                try:
                    handler(msg)
                except:
                    log.exception('Handler failed: '+str(handler))

        #store checked handler list back
        self._lock.acquire()
        self._sub_hndlrs[msg.subject]=new_hlist
        self._lock.release()

    #---System messages---------------------------------------------------------
    def process_sys_msg(self, msg):
        """
        Handles system message to close the node.
        """
        if msg.subject == mb_protocol.SYS_NODE_CLOSE:
            self.close()

    def on_close(self):
        """
        Called after the node is closed - overload this in a subclass.
        """
        pass
