"""
This module defines the Messenger msgtypes for use with the protocol socket 
communications module.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('MBProtocol')

#-------------------------------------------------------------------------------
try:
   import cPickle as pickle
except:
   import pickle

#---Globals---------------------------------------------------------------------
PROTOCOL   = '<<MB_PROTOCOL>>'      #welcome string
TERMINATOR = '<<END>>'              #string to delimit messages
SEPERATOR  = '<<,>>'                #string to seperator message components

#---Message types---------------------------------------------------------------
#a Message that was published:
MSG_PUBLISHED    = 'MSG_PUBLISHED'    

#a Message that was sent to a registered address, result returned
MSG_ADD_RESULT  = 'MSG_ADD_RESULT'  

#a Message that was sent to a registered address, no result returned
MSG_ADD_NO_RESULT  = 'MSG_ADD_NO_RESULT'  

#a Message object containing a system command/notification for external connections
MSG_SYSTEM      = 'MSG_SYSTEM'

#---system message subjects-----------------------------------------------------
SYS_NODE_CLOSE  = 'SYS_NODE_CLOSE'       #close the connection request
SYS_NODE_CLOSING= 'SYS_NODE_CLOSING'     #sent by an external node to notify it is closeing
SYS_RESULT      = 'SYS_RESULT'           #a result to a previous message

#Node connected, disconnected and error disconnected pubslished when nodes join or leave
#the message bus.
SYS_NODE_CONNECT        = 'SYS_NODE_CONNECT'
SYS_NODE_DISCONNECT     = 'SYS_NODE_DISCONNECT'  
SYS_NODE_ERR_DISCONNECT = 'SYS_NODE_ERR_DISCONNECT'

#---Message object--------------------------------------------------------------
class Message(object):
    def __init__(self,  msgid, msgtype=MSG_PUBLISHED,
                        from_node='', to_node='',
                        subject='', data=()):
        """
        Create a new message object:
            msgid       - a unique identifier.
            msgtype     - the message type - PUBLISHED/ADDRESSED
            from_node   - the sender node name
            to_node     - the destination node name (addressed message only)
            subject     - the message subject
            data        - the message data
        """
        self.msgid  = msgid                 # unique msg id
        self.msgtype= msgtype               # the message type
        self.from_node = from_node          # sender node name
        self.to_node = to_node              # destination node name
        self.subject = subject              # message subject
        # message data
        if type(data) not in [tuple,list]:
            raise Exception('msg data should be a tuple or list!')
        self.data = data             
        
    def get_msgid(self):
        """ Return the unique msgid """
        return self.msgid

    def get_msgtype(self):
        """ Return the message type """
        return self.msgtype

    def get_to(self):
        """ Return the destination node name (addressed messages) """
        return self.to_node

    def get_from(self):
        """ Return the sender node name (published messages) """
        return self.from_node

    def get_subject(self):
        """ Return the message subject"""
        return self.subject

    def get_data(self): 
        """ Return the message data tuple """
        return self.data

    def set_result(self, result):
        """ Callable to set the message result - this is replaced if a result is required"""
        pass

    def get_result(self):
        """ Callable to get the message result - this is replaced if a result is required"""
        return None



#-------------------------------------------------------------------------------
class PickleError():
    """
    A representation of data that was not pickled successfully, when being sent
    to an external client
    """
    def __init__(self,data):
        try:
            self._str = str(data)
        except:
            self._str = 'unknown data'
    def __call__(self,*args,**kwargs):
        raise Exception('Attempt to call message data that was not pickled successfully')
    def __repr__(self):
        return 'Message data could not be pickled: '+self._str

class UnPickleError():
    """
    A representation of data that was not unpickled successfully, when recieved
    by an external client
    """
    def __init__(self,data):
        self._str = str(data)

    def __call__(self,*args,**kwargs):
        raise Exception('Attempt to call message data that was not unpickled successfully')
    def __repr__(self):
        return 'Message data could not be unpickled: '+self._str
#-------------------------------------------------------------------------------
def dumps(msg):
    """
    Dumps message to a string
    """
    #header
    s = str(msg.msgid) + SEPERATOR          #msgid
    s = s + str(msg.msgtype) + SEPERATOR    #type
    s = s + str(msg.from_node)+ SEPERATOR   #from node name
    s = s + str(msg.to_node) + SEPERATOR    #to node name
    s = s + str(msg.subject)     #subject

    #pickle the data, pickle error strings replace unpicklable objects.
    for d in msg.data:
        try:
            s_d = pickle.dumps(d)
        except:
            s_d = pickle.dumps(PickleError(d))
        s = s + SEPERATOR + s_d
    #terminate
    s= s + TERMINATOR
    return s


def loads(datastring):
    """
    Recreate a message from the datastring
    """
    parts = datastring.split(SEPERATOR)
    try:
        msgid       = parts[0]
        msgtype     = parts[1]
        from_node   = parts[2]
        to_node     = parts[3]
        subject     = parts[4]
        s_data      = parts[5:]
    except:
        log.exception('Failed to split message string - invalid format?\n '+str(parts)+'\n\n')

    #unpickle the data
    data = []
    for s in s_data:
        try:
            i = pickle.loads(s)
        except:
            i = UnPickleError(s)
        data.append(i)
    data = tuple(data)
    msg = Message( msgid, msgtype, from_node, to_node, subject, data)
    return msg
#-------------------------------------------------------------------------------
