"""
MBServer/ MBSocketNode

an asyncore Channel to monitor a port for incomming connections. When a 
connection is made a MBSocketNode  object is created to control the connection 
and added to the parent MessageBus
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('MBServer')
#log.setLevel(logging.DEBUG)

#---imports---------------------------------------------------------------------
import asyncore
import socket
import threading

#Other messenger components
import mb_protocol
from mb_misc import SocketResult
from mb_socket_channel import MBSocketChannel

#-------------------------------------------------------------------------------
# A Nobe representing a client connection via a socket.
# This forwards all messages over the socket and sends all incomming messages to
# the MessageBus. It is created automatically by the MBServer
#-------------------------------------------------------------------------------
class MBSocketNode(MBSocketChannel):
    def __init__(self, msg_bus, name, conn):
        MBSocketChannel.__init__(self,name, msg_bus._map, msg_bus._timeout)

        #node name
        self._msg_bus = msg_bus
        self._name = name
        self._registered = False

        #Use to make things thread safe
        self._lock = threading.Lock()   
     
        #message counter - used to generate unique message ids
        self._count = 0
        
        #register with the messagebus and set socket
        self._name = self._msg_bus.register_node( self, self._name )

        #send the assigned name in case it is different than that requested
        conn.sendall(self._name+mb_protocol.TERMINATOR)

        self.set_socket(conn, self._msg_bus._map)

    #---channel methods---------------------------------------------------------
    def on_msg_in(self, msg):
        """
        Handles incomming addressed messages- pass to MessageBus
        """
        #Addressed messsages
        if (msg.msgtype in [mb_protocol.MSG_ADD_RESULT, mb_protocol.MSG_ADD_NO_RESULT]):
            self._msg_bus.process_msg(msg)
            return

        #a message to publish on the message bus
        if (msg.msgtype == mb_protocol.MSG_PUBLISHED):
                self._msg_bus.process_published(msg)

    def on_close( self):
        """
        Called after the socket closed normally.
        """
        log.debug('Node closed')
        self._msg_bus._remove_node(self._name, err=False)

    def on_err_close(self):
        """
        Called after the socket closed abnormally.
        """
        log.debug('Node err closed')
        self._msg_bus._remove_node(self._name, err=True)

    def process_msg(self, msg):
        """
        Handles incomming messages addressed to this node - this will be called 
        from MessageBus in the main thread - send over the socket to the client.
        """
        self.forward_msg(msg)

    def process_published(self, msg):
        """
        Handles incomming published messages - this will be called from 
        MessageBus in the main thread - send over the socket to the client.
        """
        #only forward back to client if it didn't come from the client!
        fromnode = msg.get_from()
        if fromnode!=self._name:
            self.forward_msg(msg)

    def close_node(self):
        """
        Close the node and disconnect from the MessageBus
        """
        self.close() #close socket - node removed in on_close


#---Messenger server------------------------------------------------------------
class MBServer(asyncore.dispatcher):
    """
    This object is an asyncore Channel to monitor a port for incomming 
    connections. When a connection is made a MBSocketNode  object is 
    created to control the connection and added to the parent MessageBus
    """
    def __init__(self, msg_bus, port=6667,allow_ext=False):
        asyncore.dispatcher.__init__ (self,map=msg_bus._map)

        #store a reference to the parent messagebus
        self._msg_bus = msg_bus

        #create the socket and listen for connections
        self.create_socket (socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()

        if allow_ext is True:
            host = socket.gethostname()
        else:
            host = 'localhost'
        address = (host,port)
        self.bind(address)

        #start listening for connections
        self.listen(5)

    def handle_accept(self):
        #create a new channel object here
        log.debug('in handle_accept')
        conn,add = self.accept()

        #send the handshake message
        log.debug('Sending handshake message...')
        conn.sendall(mb_protocol.PROTOCOL)
        conn.setblocking(True)
        #read the reply should be the same
        log.debug('Waiting for handshake message...')
        buffer = ''
        while len(buffer)<len(mb_protocol.PROTOCOL):
            try:
                buffer = buffer + conn.recv(1)
            except:
                log.exception('Timeout waiting for handshake message')
                conn.close()
                raise Exception('Timeout waiting for handshake message')

        #check protocol
        log.debug('Checking protocol...'+buffer)
        if buffer !=mb_protocol.PROTOCOL:
            conn.close()
            log.exception('Handshake message was not received')
            raise Exception('Handshake message was not received')

        log.info('Handshake OK, socket connected.')

        #read node name
        name = ''
        try:
            while name.endswith(mb_protocol.TERMINATOR) is False:
                name = name + conn.recv(1)
        except:
            log.exception('Timeout waiting for node name')
            conn.close()
            raise Exception('Timeout waiting for node name')
        name = name.rstrip(mb_protocol.TERMINATOR)
        #create the node to handle this client
        node = MBSocketNode(self._msg_bus, name, conn)

        log.debug('MBSocketNode created')
