"""
MBSocketChannel

Base clas to handle external connections to the MessageBus via sockets, used for
both MBSocketNode and MBSocketClient
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('MBSocketChannel')

#-------------------------------------------------------------------------------
import asynchat
import socket
import threading

from mb_misc import SocketResult
import mb_protocol

#-------------------------------------------------------------------------------
class MBSocketChannel(asynchat.async_chat):
    def __init__(self, name, cmap, timeout=30):
        asynchat.async_chat.__init__(self, None, cmap)
        #name
        self._name = name
        self._timeout= timeout

        #Use to make things thread safe
        self._lock = threading.Lock()   

        #message counter - used to generate unique message ids
        self._count = 0

        #replies dictionary {msgid:reply_obj}
        self._replies={}
        
        #set the terminator
        self.set_terminator (mb_protocol.TERMINATOR)

        #store incoming data until complete
        self._buffer = ''

        #flags
        self._closing = False           #set when closing connection to avoid recieving multiple close events

    #---asyncore methods--------------------------------------------------------
    def handle_close(self):
        #this should close the socket
        asynchat.async_chat.close(self)
        if (self._closing is False):
            log.warning('channel closed unexpectedly [handle_close]!')
            self.on_err_close()
        else:
            log.debug('channel closed [handle_close]')
            self.on_close()

    def handle_connect(self):
        pass

    def handle_accept(self):
        pass

    def handle_expt(self):
        """
        Avoid connecting errors on windows when nothing is listening on the port
        that we are trying to connect to.
        """
        log.error('in handle_exp')
        asynchat.async_chat.handle_expt(self)

    def handle_error(self):
        asynchat.async_chat.handle_error(self)

    def handle_write(self):
        self._lock.acquire()
        asynchat.async_chat.handle_write(self)
        self._lock.release()

    def collect_incoming_data(self, data):
        self._lock.acquire()
        self._buffer = self._buffer + data
        self._lock.release()

    def found_terminator(self):
        """Process message"""
        #empty the buffer
        self._lock.acquire()
        data = self._buffer
        self._buffer = ''
        self._lock.release()

        #recreate the message object
        msg = mb_protocol.loads(data)

        #A system message
        if (msg.msgtype == mb_protocol.MSG_SYSTEM):

            #a result to a previous message
            if msg.subject == mb_protocol.SYS_RESULT:
                log.debug('reply message data: '+str(msg.data))
                msgid,res = msg.data
                sent_msg = self._replies.pop(msgid, None)
                if sent_msg is not None:
                    #call the set_result method object to return the result
                    sent_msg.set_result(res)
                else:
                    log.error('Result recieved for unknown msgid: '+msgid)
                return

            #a close request
            if msg.subject  == mb_protocol.SYS_NODE_CLOSE:
                log.debug('Node close request')
                self.close()
                return

            #a close notification
            if msg.subject  == mb_protocol.SYS_NODE_CLOSING:
                log.debug('Node closing notification')
                self._closing = True
                return

        #Addressed messsage to this node requiring a result returned.
        if (msg.msgtype == mb_protocol.MSG_ADD_RESULT):
            #return the result via SocketResult and SocketChannel.send_result
            res_obj = SocketResult( self, msg.msgid) 
            msg.set_result = res_obj.set_result
            msg.get_result = res_obj.get_result

        #call the subclassed handler
        self.on_msg_in(msg)

    def close(self):
        """ Close the channel after sending a SYS_NODE_CLOSE message"""
        #already closing/closed
        if (self._closing is False):
            #set flags
            self._closing = True
            #send a system close message
            msg = mb_protocol.Message( 'SYS', msgtype=mb_protocol.MSG_SYSTEM,
                        from_node=None, to_node=None,
                        subject=mb_protocol.SYS_NODE_CLOSING, data=())
            s = mb_protocol.dumps(msg)
            self._lock.acquire()
            self.push(s)
            self._lock.release()

        #close socket
        asynchat.async_chat.close(self)
        self.on_close()

    def set_socket(self, sock, map):
        #set socket options to keep things zippy...
        sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY,1)
        asynchat.async_chat.set_socket(self, sock, map)

    #---socket methods----------------------------------------------------------
    def forward_msg(self, msg):
        """
        Forward a message over the socket connection
        """
        #dump msg to string
        s = mb_protocol.dumps(msg)

        #need to wait for the result?
        if msg.msgtype == mb_protocol.MSG_ADD_RESULT:
            #store msg internally until reply is recieved
            self._replies[msg.msgid] = msg

        #send
        self._lock.acquire()
        self.push(s)
        self._lock.release()

    def send_result( self, msgid, result):
        """
        Send a result to an addressed message - this is called by the 
        SocketResult object when it is called via the msg.set_result method.
        """
        #construct the message object
        msg = mb_protocol.Message( 'REPLY TO:'+str(msgid),
                        msgtype=mb_protocol.MSG_SYSTEM,
                        from_node=None, to_node=None,
                        subject=mb_protocol.SYS_RESULT, data=(msgid,result,))
        #dump message to string
        s = mb_protocol.dumps(msg)
        #send
        self._lock.acquire()
        self.push(s)
        self._lock.release()

    #---------------------------------------------------------------------------
    def on_msg_in(self, msg):
        """
        Handles incomming messages- subclass this to define what to do
        """
        pass

    def on_close(self):
        """
        Called after the socket closed normally.
        """
        pass

    def on_err_close(self):
        """
        Called after the socket closed abnormally.
        """
        pass
