"""
MBSocketClient
--------------
A simple client object that connects to a MessageBus in a different process via 
a socket connection. All messages are handled in the communications thread 
monitoring the socket - this should be subclassed and the incomming_msg / 
incomming_msg_published methods changed for other behaviour.
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('MBSocketClient')
#log.setLevel(logging.DEBUG)

#-------------------------------------------------------------------------------
import threading
import socket
import time

import mb_protocol
from mb_socket_channel import MBSocketChannel
from mb_misc import SocketResult, AsyncResult, WeakRefHandler, MBThread, AddressError
 
#-------------------------------------------------------------------------------
class MBSocketClient(MBSocketChannel):
    def __init__(self, name, address=('localhost',6667), timeout=10):
        #asyncore channel map
        self._map = {}
        MBSocketChannel.__init__(self, name, self._map, timeout)
        
        #socket address to use
        self._address=address

        #message handlers for addressed messages
        self._msg_hndlrs = {}  #dictionary of {subject:handler}
        
        #handlers for published messagess
        self._sub_hndlrs = {} #dictionary of {subject:[handlers]}

        #communications thread
        self._thread = None

    #---channel methods---------------------------------------------------------
    def close(self):
        MBSocketChannel.close(self)
        if (self._thread is not None) and (self._thread!=threading.current_thread()):
            self._thread.join()
    
    def on_msg_in(self, msg):
        """
        Handles incomming (from socket) messages.
        This calls handlers in the comms thread - subclass raise an event and
        call _do_send to do this from main thread using events
        """
        log.debug('Incomming message: '+str(msg.msgtype)+' '+str(msg.subject) + ' from '+str(msg.from_node))
       
        #Addressed messsage to this node
        if (msg.msgtype in [mb_protocol.MSG_ADD_RESULT, mb_protocol.MSG_ADD_NO_RESULT]):
            self._do_send(msg)

        #a message to publish locally
        elif (msg.msgtype == mb_protocol.MSG_PUBLISHED):
            self._do_publish(msg)

    #---addressed messages------------------------------------------------------
    def send_msg(self, to, subject, data=(), get_result=False):
        """
        Send a message via the MessageBus.
        to      -   Node to send message to.
        subject -   The message subject.
        data    -   A list or tuple data object.
        get_result  -   True/False wait for and return the result.

        Returns - the result value if get_result is True, or None
        """
        #construct the message object
        self._lock.acquire()
        msgid = str(self._name)+':'+str(self._count)
        self._count+=1
        self._lock.release()
        if get_result is True:
            msgtype = mb_protocol.MSG_ADD_RESULT
        else:
            msgtype = mb_protocol.MSG_ADD_NO_RESULT

        msg = mb_protocol.Message( msgid, msgtype,from_node=self._name,
                       to_node=to, subject=subject, data=data)

        #result object replaces the set_result/get_result methods of the msg 
        #the message is stored in a dictionary to allow the comms thread to store 
        #the result when it arrives, using the set_result method.
        if get_result is True:
            res_obj = AsyncResult(msgid, self._timeout)
            msg.set_result = res_obj.set_result
            msg.get_result = res_obj.get_result
            self._replies[msgid] = msg

        #send over link
        log.debug('Sending message to: '+msg.to_node+' subject:'+msg.subject)
        self.forward_msg(msg)
        res = msg.get_result()

        #check if an AddressError
        if isinstance(res, Exception):
            raise res

        return res

    def set_handler(self, subject, handler):
        """
        Register a message handler for the addressed message subject given.
        """
        log.debug('In register handler '+subject)
        #get the lock
        self._lock.acquire()
        #store handler into dictionary
        self._msg_hndlrs[subject]=WeakRefHandler(handler) 
        #release the lock
        self._lock.release()

    def remove_handler(self, subject):
        """
        Remove the message handler for the message subject.
        """
        #get the lock
        self._lock.acquire()
        #store handler into dictionary
        self._msg_hndlrs.pop(subject,None) 
        #release the lock
        self._lock.release()

    def _do_send(self, msg):
        #get handler
        self._lock.acquire()
        handler = self._msg_hndlrs.get(msg.subject,None)
        self._lock.release()
        if handler is not None:
            #check if the handler is dead
            if handler.isdead():
                #remove it
                self._lock.acquire()
                self._msg_hndlrs.pop(msg.subject)
                self._lock.release()
            else:   
                #call it
                try:
                    res = handler(msg)
                except:
                    log.exception('Handler failed: '+str(handler))
                    res = None
        else:
            log.info('No handler found for subject: '+str(msg.subject))
            res = None
        #return the result via msg.set_result()
        msg.set_result(res)

    #---published messages------------------------------------------------------
    def publish_msg(self, subject, data=()):
        #construct the message object
        self._lock.acquire()
        msgid = str(self._name)+':'+str(self._count)
        self._count+=1
        self._lock.release()
        msg = mb_protocol.Message( msgid, msgtype=mb_protocol.MSG_PUBLISHED, 
                        from_node=self._name, to_node='', 
                        subject=subject, data=data )

        #send over link
        log.debug('Publishing message: '+msg.subject)
        self.forward_msg(msg)

    def subscribe(self, subject, handler):
        """
        Subscribe a message handler for the published message subject given.
        """
        log.debug('In subscribe '+subject)
        #get the lock
        self._lock.acquire()

        #get current list of handlers
        hlist = self._sub_hndlrs.get(subject,[])
        #add a weakref to the handler
        hlist.append(WeakRefHandler(handler)) 
        #store back into dictionary
        self._sub_hndlrs[subject]=hlist 

        #release the lock
        self._lock.release()

    def unsubscribe(self, subject, handler):
        """
        Remove the message handler subscribed for the message type.
        """
        hlist = self._sub_hndlrs.get(subject,[])
        #remove the handler
        for n in range(0,len(hlist)):
            h = hlist[n]
            if h == handler:
                n = hlist.index(h)
                hlist.pop(n)
        #store back into dict
        self._sub_hndlrs[subject] = hlist

    def _do_publish(self, msg):
        #get message handlers
        self._lock.acquire()
        hlist = self._sub_hndlrs.get(msg.subject,[])
        self._lock.release()

        #call all handlers
        new_hlist=[]
        for handler in hlist:
            #check if the handler is dead
            if handler.isdead():
                pass
            else: 
            #not dead call it
                new_hlist.append(handler)
                try:
                    handler(msg)
                except:
                    log.exception('Handler failed: '+str(handler))

        #store checked handler list back
        self._lock.acquire()
        self._sub_hndlrs[msg.subject]=new_hlist
        self._lock.release()

    #---close callbacks---------------------------------------------------------
    def on_close( self):
        """
        Called after the socket closed normally.
        """
        #publish a SYS_NODE_DISCONNECT message locally - i.e. not over the link.
        self._lock.acquire()
        msgid = str(self._name)+':'+str(self._count)
        self._count+=1
        self._lock.release()
        msg = mb_protocol.Message( msgid, msgtype=mb_protocol.MSG_PUBLISHED, 
                        from_node=self._name, to_node='', 
                        subject=mb_protocol.SYS_NODE_DISCONNECT, data=(self._name,) )
        self.on_msg_in(msg)

    def on_err_close(self):
        """
        Called after the socket closed abnormally.
        """
        #publish a SYS_NODE_ERR_DISCONNECT message locally - i.e. not over the link.
        self._lock.acquire()
        msgid = str(self._name)+':'+str(self._count)
        self._count+=1
        self._lock.release()
        msg = mb_protocol.Message( msgid, msgtype=mb_protocol.MSG_PUBLISHED, 
                        from_node=self._name, to_node='', 
                        subject=mb_protocol.SYS_NODE_ERR_DISCONNECT, data=(self._name,) )
        self.on_msg_in(msg)

    #---Interface methods-------------------------------------------------------
    def connect(self):
        """
        Connect to a listening MessageBus.
        """
        #create the socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(self._timeout)

        #try to connect
        log.info('connecting on: '+str(self._address)+'...')
        nmax=10
        for n in range(0,nmax+1):
            log.info('try'+str(n))
            try:
                sock.connect(self._address)
                break
            except:
                if n==nmax:
                    raise Exception('Could not connect!')
                time.sleep(1)

        #do handshake
        self.handshake(sock)

    def listen(self):
        """Allow the master MessageBus to connect to us"""
        
        #listen
        sock = listen(port=self._address[1],timeout=30)
        
        #do handshake
        self.handshake(sock)

    def handshake(self,conn):
        #send the handshake message
        log.debug('Sending handshake message...')
        conn.sendall(mb_protocol.PROTOCOL)

        #read the reply should be the same
        log.debug('Waiting for handshake message...')
        buffer = ''
        while len(buffer)<len(mb_protocol.PROTOCOL):
            try:
                buffer = buffer + conn.recv(1)
            except:
                log.exception('Timeout waiting for handshake message')
                conn.close()
                raise Exception('Timeout waiting for handshake message')

        #check protocol
        log.debug('Checking protocol...'+buffer)
        if buffer !=mb_protocol.PROTOCOL:
            conn.close()
            log.exception('Handshake message was not received')
            raise Exception('Handshake message was not received')
        log.info('Handshake OK, socket connected.')

        #send node name
        log.info('Checking name...')
        conn.sendall(self._name+mb_protocol.TERMINATOR)
        name=''
        try:
            while name.endswith(mb_protocol.TERMINATOR) is False:
                name = name + conn.recv(1)
        except:
            log.exception('Timeout waiting for node name')
            conn.close()
            raise Exception('Timeout waiting for node name')
        assigned_name = name.rstrip(mb_protocol.TERMINATOR)
        log.debug('Name is: '+assigned_name)
        if assigned_name !=self._name:
            log.warning('Unable to connect using preferred name')
            self._name = assigned_name

        #set socket
        self.set_socket(conn,self._map)

        #no comms thread start one.
        if self._thread is None:
            self._thread = MBThread(self._map)
            self._thread.start()

        #previous thread is dead - replace
        if self._thread.is_alive() is False:
            self._thread = MBThread(self._map)
            self._thread.start()

