"""
New MessageBus system to unify the EngineMessage and Message types and allow 
EngineProcess to publish messages (i.e. 'Engine_Paused' etc)

                                             ______     ______
                                            | Coms |   | Coms |
              ____________                  |thread|   |thread|
    Node1 ___|            |___ MBSocketNode.|......|...|......|.ExternalClient1
    Node2 ___| MessageBus |___ MBSocketNode.|......|...|......|.ExternalClient2
    Node3 ___|____________|___ MBSocketNode.|......|...|......|.ExternalClient3
                                            |______|   |______|
Addressed messages:
-------------------
1)Nodes send message to a node name with a subject string and data inside a 
Message object together with a Reply object or callable(result) to recieve the 
reply/result
2) The Message and Reply objects are passed to the MessageBus which sends it to 
the correct Node.
3) The Nodes recieve the message and reply objects, does what it wants with the 
message and returns the result by calling the reply object i.e. reply(result)
4) The reply object/callable then returns the reply to the sending node directly
"""
#---logging---------------------------------------------------------------------
import logging
log = logging.getLogger('MessageBus')
#log.setLevel(logging.DEBUG)

import threading

#Other messenger components
from mb_misc import AsyncResult, MBThread, AddressError
from mb_server import MBServer
import mb_protocol

#-------------------------------------------------------------------------------
class MessageBusBase():
    def __init__(self, name='MessageBus', timeout=5):
        """
        Create a MessageBus instance - this should be done only once and the 
        instance stored where it can be used.
        """
        self._name = name
        self._timeout = timeout

        #message counter 
        #   - used to generate unique message ids for retrieving replies
        self._count = 0
    
        #nodes
        self._lock = threading.Lock()
        self._nodes = {} #{name:node} dictionary

        #attributes required to handle external clients
        self._thread = None     #comms thread
        self._map    = {}       #channel map
        self._server = None     #reference to the server if started
        self._shutdown = False  #shutdown flag used to shutdown quietly

    #---nodes-------------------------------------------------------------------
    def has_node(self, node_name):
        """
        Check if a node exists on the message bus.
        """
        return self._nodes.has_key(node_name)

    def register_node(self, node, node_name):
        """
        Add a node to the Messenger system - this is given a unique name which 
        can be used to send messages from point to point. The node itself should
        have a recieve(msg) method to handle messages sent to the node which may
        have different subjects.
        """
        self._lock.acquire()

        if self._nodes.has_key(node_name):
            new_name = 'Node'+str(len(self._nodes))
            log.warning('Node name '+node_name+' already registered using assigned name: '+new_name)
            node_name = new_name

        #store reference to the node
        self._nodes[node_name] = node
        self._lock.release()
        node._registered = True

        log.debug('Registered node : '+str(node_name))
        self.publish_msg( mb_protocol.SYS_NODE_CONNECT, (node_name,))
        return node_name

    def _remove_node(self, node_name, err=False):
        """
        This is called internally by closing nodes.
        Remove the node with the name given from the Messenger system and publish
        a SYS_NODE_CLOSED or SYS_NODE_ERR_CLOSED message.
        """
        self._lock.acquire()
        node = self._nodes.pop(node_name,None)
        self._lock.release()
        if node is None:
            return

        if self._shutdown is True:
            return
        if err is True:
            self.publish_msg( mb_protocol.SYS_NODE_ERR_DISCONNECT, (node_name,))
        else:
            self.publish_msg( mb_protocol.SYS_NODE_DISCONNECT, (node_name,))
        log.debug('Node removed: '+str(node_name)+' error= '+str(err))

    def close_node(self, node_name):
        """
        Send a message to a node requesting it closes.
        """
        log.debug('Closing node ' + node_name)
        #send a system close message
        self._lock.acquire()
        msgid = str(self._name)+':'+str(self._count)
        self._count+=1
        self._lock.release()

        msgtype = mb_protocol.MSG_SYSTEM
        subject = mb_protocol.SYS_NODE_CLOSE
        msg = mb_protocol.Message( msgid, msgtype,from_node=self._name,
                       to_node=node_name, subject=subject, data=())
        #send via message bus
        self.process_msg(msg)


    def start_server(self,port=6668,allow_ext=False):
        """
        Start a thread to monitor for external clients.
            port        - port to listen on.
            allow_ext   - allow clients external to this computer. 
        """
        if self._server is not None:
            log.exception('Server already started')
            raise exception('Server already started')

        #create a MBServer channel and start the asyncore thread
        log.debug(  'Starting server '+str(port)+
                    ' allow external ='+str(allow_ext))
        self._server = MBServer(self,port,allow_ext)
        self.port = port

        #start a communications thread if needed.
        if self._thread is None:
            start = True
        elif self._thread.is_alive() is False:
            start = True
        else:
            start = False
        if start is True:
            self._thread = MBThread(self._map)
            self._thread.start()

    def stop_server(self):
        """
        Stop the server monitoring for external clients
        """
        if self._server is not None:
            self._server.close()
            self._server = None
            log.debug('Server stopped')

    def has_server(self):
        """
        Returns True if server is active, False otherwise
        """
        if self._server is None:
            return False
        else:
            return True

    def shutdown(self):
        """Close external nodes and shutdown the communications thread"""
        log.debug('Shutting down')
        self._shutdown = True

        #close all nodes - they remove themselves from the MessageBus and close 
        #sockets etc.
        for node in self._nodes.values():
            node.close()

        #stop server
        self.stop_server()

        #wait for thread to stop to avoid errors on application exit
        if self._thread is None:
            return
        log.debug('waiting for communications thread to end')
        self._thread.join()

    #---send addressed messages-------------------------------------------------
    def send_msg(self, to, subject, data=(), get_result=False):
        """
        Send a message directly on the MessageBus.
        to      -   Node to send message to.
        subject -   The message subject.
        data    -   A list or tuple data object.        
        get_result  -   True/False wait for and return the result.

        Returns - the result value if get_result is True, or None
        """
        #construct the message object
        self._lock.acquire()
        msgid = str(self._name)+':'+str(self._count)
        self._count+=1
        self._lock.release()
        if get_result is True:
            msgtype = mb_protocol.MSG_ADD_RESULT
        else:
            msgtype = mb_protocol.MSG_ADD_NO_RESULT

        msg = mb_protocol.Message( msgid, msgtype,from_node=self._name,
                       to_node=to, subject=subject, data=data)

        #result object allows the reply/result to be returned to the sending 
        #thread
        if get_result is True:
            res_obj = AsyncResult(msgid, self._timeout)
            msg.set_result = res_obj.set_result
            msg.get_result = res_obj.get_result

        #send the message via the MessageBus
        self.process_msg(msg)

        if get_result is False:
            return None

        #get the reply/result - this will block the sending thread until a result
        #is set using msg.set.result(res)
        res = msg.get_result()

        #check if an AddressError
        if isinstance(res, AddressError):
            raise Exception('Unknown address:'+res._address)

        return res

    def process_msg(self, msg):
        """
        Send a message on the MessageBus.
        msg         -   the message object to send
        """
        #addressed message?
        if (msg.msgtype not in [mb_protocol.MSG_ADD_RESULT, mb_protocol.MSG_ADD_NO_RESULT, mb_protocol.MSG_SYSTEM]):
            raise Exception('Not an addressed message type')

        #check if this is the main thread or not.
        mainthread = isinstance(threading.current_thread(), threading._MainThread)
        if (mainthread is True):
            self._do_send(msg)
        else:
            self._do_thread_send(msg)

    def _do_send(self, msg):
        """
        Send the message in the current thread - this is only used in the 
        main thread.
        """
        log.debug('Sending message to: '+msg.to_node+' subject: '+msg.subject)

        #get the node
        self._lock.acquire()       #get the lock
        node = self._nodes.get(msg.to_node,None)
        self._lock.release()       #release the lock
        if node is None:
            res = AddressError(msg.to_node)
            msg.set_result(res)
            return

        #pass the message object to the node
        try:
            node.process_msg(msg) 
        except:
            log.exception('Node recieve failed: '+str(msg.to_node))

    def _do_thread_send(self,msg):
        """
        Send the message in the main thread via an event.
        This should be overloaded to implement different event handlers
        msg is the message object to send and typically raise and event to call 
        the _do_send in the main thread.
        """
        pass

    #---Published messages------------------------------------------------------
    def publish_msg(self, subject, data):
        """
        Publish a message directly on the MessageBus
        """
        #construct the message object
        self._lock.acquire()
        msgid = str(self._name)+':'+str(self._count)
        self._count+=1
        self._lock.release()
        msg = mb_protocol.Message( msgid, msgtype=mb_protocol.MSG_PUBLISHED, 
                        from_node=self._name, to_node=None, 
                        subject=subject, data=data )
        self.process_published(msg)

    def process_published(self, msg):
        """
        Publish a message on the MessageBus
        msg         -   the mesage object to publish
        """
        #published message
        if (msg.msgtype!=mb_protocol.MSG_PUBLISHED):
            raise Exception('Not a published message type')
        #always publish via an event  
        self._do_thread_publish(msg)      

    def _do_publish(self,msg):
        """
        Publish the message in the main thread - this is only used in the 
        main thread.
        """
        #get the subject
        log.debug('Publishing message; subject= '+msg.subject)

        #pass message to each node (including the sending node) in turn for it 
        #to deal with i.e. ignore or call any subcribed handlers. Although it 
        #may seem odd to pass the message from the node>Bus>node it ensures that
        #the message is always published in the main thread and shouldn't add
        #too much overhead.
        self._lock.acquire()
        nodes = list(self._nodes.values()) #get the nodes into a new list to allow other nodes to be created in handlers
        self._lock.release()
        for node in nodes:
            try:
                node.process_published(msg)
            except:
                log.exception('Node recieve failed: '+node._name)

    def _do_thread_publish(self, msg):
        """
        Publish the message in the main thread via an event.
        This should be overloaded to implement different event handlers and 
        typically raise an event to call the _do_publish method in the main 
        thread
        """
        pass
    
